%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id: megaco_test_generator.erl,v 1.1.1.1 2003/10/27 16:22:25 richardc Exp $
%%
%%----------------------------------------------------------------------
%% Purpose: Message sequence generator for megaco
%%----------------------------------------------------------------------
-module(megaco_test_generator).

-export([start_link/1, start_link/2, 
	 stop/1, 
	 tcp/2, tcp_await_reply/1]).

%% Internal exports
-export([start/2, init/3, tcp_handler_main/3]).


%%----------------------------------------------------------------------

-record(state, {parent, handler, name}).

%% A TCP sequence generator data record
-record(tcp, 
	{
	  listen,     % Listen socket
	  connection, % Connection socket
	  encode,     % Encode fun
	  decode,     % Decode fun
	  result = [] % Accumulated results from verification
	 }).



%%----------------------------------------------------------------------

start_link(Name) when list(Name) ->
    start(Name, self()).

start_link(Name, Node) when list(Name), Node /= node() ->
    case rpc:call(Node, ?MODULE, start, [Name, self()]) of
	{ok, Pid} ->
	    link(Pid),
	    {ok, Pid};
	Error ->
	    Error
    end;
start_link(Name, Node) when list(Name), Node == node() ->
    case start(Name, self()) of
	{ok, Pid} ->
	    link(Pid),
	    {ok, Pid};
	Error ->
	    Error
    end.	    

start(Name, Pid) when pid(Pid) ->
    proc_lib:start(?MODULE, init, [Name, self(), Pid]).


stop(Pid) ->
    request(Pid, stop).


tcp(Pid, Instructions) when pid(Pid), list(Instructions) ->
    request(Pid, {tcp, Instructions}).

tcp_await_reply(Pid) ->
    await_reply(Pid, tcp_reply).


request(Pid, Req) ->
    Pid ! {Req, self()}.

await_reply(Pid, Tag) ->
    receive
	{Tag, Pid, Reply} ->
	    Reply
    end.


%%----------------------------------------------------------------------
    
init(Name, Starter, Parent) ->
    d("init -> entry with"
      "~n   Name:    ~s"
      "~n   Starter: ~p"
      "~n   Parent:  ~p", [Name, Starter, Parent]),
    put(name, "GEN"),
    proc_lib:init_ack(Starter, {ok, self()}),
    loop(#state{parent = Parent, name = Name}).

loop(#state{parent = Parent} = State) ->
    d("loop -> await request"),
    receive
	{stop, Parent} ->
	    abort(State),
	    exit(normal);
	
	{'EXIT', Parent, Reason} ->
	    exit({parent_exit, Reason});

	{abort, Parent} ->
	    abort(State),
	    loop(State);
	    
	{{tcp, Instructions}, Parent} ->
	    d("loop -> received tcp request"),
	    Handler = start_tcp_handler(State#state.name, Instructions),
	    d("loop -> tcp handler: ~p", [Handler]),
	    loop(State#state{handler = Handler});
	
	{tcp_result, Pid, Res} when State#state.handler == Pid ->
	    d("loop -> received tcp_reply"),
	    Parent ! {tcp_reply, self(), Res},
	    loop(State#state{handler = undefined});
	
	Any ->
	    p("init -> received "
	      "~n   Any: ~p", [Any]),
	    loop(State)
	    
    end.


%%% ----------------------------------------------------------------

abort(#state{handler = Pid}) when pid(Pid) ->    
    Pid ! {stop, self()};
abort(_) ->
    ok.


%%% ----------------------------------------------------------------

start_tcp_handler(Name, Instructions) ->
    spawn_link(?MODULE, tcp_handler_main, [Name, Instructions, self()]).

tcp_handler_main(Name, Instructions, Parent) ->
    put(name, Name),
    {Tcp, Reply} = handle_tcp(Instructions),
    Parent ! {tcp_result, self(), Reply},
    receive
	{stop, Parent} ->
	    tcp_cleanup(Tcp),
	    exit(normal)
    end.


%% Instructions: [instruction()]
%% instruction() = {encode, Encode} |
%%                 {decode, Decode} |
%%                 {send, Msg} |
%%                 {expect_receive, TransactionType} |
%%                 {expect_receive, {TransactionType, To}} |
%%                 {listen, Port} |
%%                 {expect_accept, {{Host, Port}, To}} |
%%                 {accept, Port} |
%%                 {accept, {Addr, Port}} |
%%                 {sleep, To} |
%%                 
handle_tcp(Instructions) when list(Instructions) ->
    (catch handle_tcp1(parse_tcp(Instructions, []))).
    
% handle_tcp1(Instructions) ->
%     F   = fun do_tcp/2, 
%     Tcp = lists:foldl(F, #tcp{}, Instructions),
%     Res = lists:reverse(Tcp#tcp.result),
%     Tcp#tcp{result = Res}.
handle_tcp1(Instructions) ->
    handle_tcp1(Instructions, #tcp{}).

handle_tcp1([], #tcp{result = Res} = Tcp) ->
    Reply = {ok, lists:reverse(Res)},
    {Tcp, Reply};
handle_tcp1([Instruction|Instructions], Tcp0) ->
    case (catch do_tcp(Instruction, Tcp0)) of
	Tcp when record(Tcp, tcp) ->
	    handle_tcp1(Instructions, Tcp);
	Error ->
	    {Tcp0, Error}
    end.

do_tcp({encode, Encode}, Tcp) ->
    p("encode: ~p", [Encode]),
    Tcp#tcp{encode = Encode};

do_tcp({decode, Decode}, Tcp) ->
    p("decode: ~p", [Decode]),
    Tcp#tcp{decode = Decode};

do_tcp({listen, Port}, Tcp) ->
    p("listen to ~p", [Port]),
    Opts = [binary, {packet, tpkt}, {active, false}, {reuseaddr, true}],
    case (catch gen_tcp:listen(Port, Opts)) of
	{ok, Listen} ->
	    d("listen -> listen socket created"),
	    Tcp#tcp{listen = Listen};
	Error ->
	    tcp_error(listen, Error)
    end;

do_tcp({expect_accept, {Addr, To}}, #tcp{listen = Listen} = Tcp) ->
    p("expect_accept from ~p (~p)", [Addr, To]),
    case (catch gen_tcp:accept(Listen, To)) of
	{ok, Sock} ->
	    d("expect_accept -> connection accepted"),
	    case (catch inet:peername(Sock)) of
		{ok, {Addr, _Port}} ->
		    d("expect_accept -> valid address"),
		    Tcp#tcp{connection = Sock};
		{ok, {_Addr, _Port}} when Addr == any ->
		    d("expect_accept -> valid"),
		    Tcp#tcp{connection = Sock};
		{ok, AddrPort} ->
		    tcp_error(expect_accept, {unknown_connect, AddrPort});
		Else ->
		    tcp_error(expect_accept, {peername, Else})
	    end;
	Error ->
	    close(Listen),
	    tcp_error(expect_accept, {accept_error, Error})
    end;

do_tcp({connect, {Addr, Port, To}}, Tcp) ->
    p("connect to ~p, ~p", [Addr, Port]),
    Opts = [binary, {packet, tpkt}, {active, once}],
    case (catch gen_tcp:connect(Addr, Port, Opts, To)) of
	{ok, Sock} ->
	    d("connect -> connected"),
	    Tcp#tcp{connection = Sock};
	Error ->
	    tcp_error(connect, Error)
    end;

do_tcp({send, Desc, Msg}, #tcp{connection = Sock,
			 encode     = Encode} = Tcp) ->
    p("send ~s message", [Desc]),
    case (catch Encode(Msg)) of
	{ok, Bin} ->
	    d("send -> message encoded, now add tpkt header"),
	    NewBin = add_tpkt_header(Bin),
	    d("send -> tpkt header added, now send"),
	    case (catch gen_tcp:send(Sock, NewBin)) of
		ok ->
		    d("send -> message sent"),
		    Tcp;
		Error ->
		    d("send -> send failed: ~n~p",[Error]),
		    close(Sock),
		    close(Tcp#tcp.listen),
		    tcp_error(send, {send_error, Error})
	    end;
	Error ->
	    d("send -> encode failed: ~n~p",[Error]),
	    close(Sock),
	    close(Tcp#tcp.listen),
	    tcp_error(send, {endode_error, Error})
    end;

do_tcp({expect_receive, Desc, {Verify, To}}, 
       #tcp{connection = Sock,
	    decode     = Decode} = Tcp) ->
    p("expect_receive ~s message", [Desc]),
    inet:setopts(Sock, [{active, once}]),
    receive
	{tcp, Sock, <<3:8, _X:8, Length:16, Msg/binary>>} ->
	    d("expect_receive -> received message: Length = ~p", [Length]),
	    case (catch Decode(Msg)) of
		{ok, MegaMsg} when tuple(MegaMsg) ->
		    d("expect_receive -> decode successfull, now verify"),
		    case (catch Verify(MegaMsg)) of
			{ok, Res} ->
			    d("expect_receive -> verify successfull"),
			    Acc = Tcp#tcp.result,
			    Tcp#tcp{result = [Res|Acc]};
			Else ->
			    d("expect_receive -> verify failed: ~n~p~n~p", 
			      [Else, MegaMsg]),
			    tcp_error(expect_receive, {verify_failed, Else})
		    end;
		Error ->
		    d("expect_receive -> decode failed: ~p", [Error]),
		    tcp_error(expect_receive, Error)
	    end;
	Else ->
	    d("expect_receive -> received unknown: ~p", [Else]),
	    tcp_error(expect_receive, {unexpected_message, Else})
    after To ->
	    tcp_error(expect_receive, timeout)
    end;

do_tcp({sleep, To}, Tcp) ->
    p("sleep ~p", [To]),
    sleep(To),
    Tcp.


tcp_error(Instr, Error) ->
    throw({error, {Instr, Error}}).


tcp_cleanup(#tcp{listen = Listen, connection = Conn}) ->
    close(Listen),
    close(Conn).


parse_tcp([], RevInstrs) ->
    lists:reverse(RevInstrs);

parse_tcp([{encode, Encode}|Instrs], RevInstrs) 
  when function(Encode) ->
    parse_tcp(Instrs, [{encode, Encode}|RevInstrs]);

parse_tcp([{decode, Decode}|Instrs], RevInstrs) 
  when function(Decode) ->
    parse_tcp(Instrs, [{decode, Decode}|RevInstrs]);

parse_tcp([{listen, Port}|Instrs], RevInstrs) 
  when integer(Port), Port > 0 ->
    parse_tcp(Instrs, [{listen, Port}|RevInstrs]);

parse_tcp([{expect_accept, any}|Instrs], RevInstrs) ->
    parse_tcp(Instrs, [{expect_accept, {any, infinity}}|RevInstrs]);

parse_tcp([{expect_accept, {any, To}}|Instrs], RevInstrs) 
  when integer(To), To >= 0 ->
    parse_tcp(Instrs, [{expect_accept, {any, To}}|RevInstrs]);

parse_tcp([{expect_accept, {Host, infinity}}|Instrs], RevInstrs) ->
    {ok, Addr} = inet:getaddr(Host, inet),
    parse_tcp(Instrs, [{expect_accept, {Addr, infinity}}|RevInstrs]);

parse_tcp([{expect_accept, {Host, To}}|Instrs], RevInstrs) 
  when integer(To), To >= 0 ->
    {ok, Addr} = inet:getaddr(Host, inet),
    parse_tcp(Instrs, [{expect_accept, {Addr, To}}|RevInstrs]);

parse_tcp([{expect_accept, Host}|Instrs], RevInstrs) ->
    {ok, Addr} = inet:getaddr(Host, inet),
    parse_tcp(Instrs, [{expect_accept, {Addr, infinity}}|RevInstrs]);

parse_tcp([{connect, Port}|Instrs], RevInstrs) 
  when integer(Port), Port > 0 ->
    {ok, Host} = inet:gethostname(),
    {ok, Addr} = inet:getaddr(Host, inet),
    parse_tcp(Instrs, [{connect, {Addr, Port, infinity}}|RevInstrs]);

parse_tcp([{connect, {Port, infinity}}|Instrs], RevInstrs) 
  when integer(Port), Port > 0 ->
    {ok, Host} = inet:gethostname(),
    {ok, Addr} = inet:getaddr(Host, inet),
    parse_tcp(Instrs, [{connect, {Addr, Port, infinity}}|RevInstrs]);

parse_tcp([{connect, {Port, To}}|Instrs], RevInstrs) 
  when integer(Port), Port > 0,
       integer(To), To > 0 ->
    {ok, Host} = inet:gethostname(),
    {ok, Addr} = inet:getaddr(Host, inet),
    parse_tcp(Instrs, [{connect, {Addr, Port, To}}|RevInstrs]);

parse_tcp([{connect, {Host, Port}}|Instrs], RevInstrs) 
  when integer(Port), Port > 0 ->
    {ok, Addr} = inet:getaddr(Host, inet),
    parse_tcp(Instrs, [{connect, {Addr, Port, infinity}}|RevInstrs]);

parse_tcp([{connect, {Host, Port, infinity}}|Instrs], RevInstrs) 
  when integer(Port), Port > 0 ->
    {ok, Addr} = inet:getaddr(Host, inet),
    parse_tcp(Instrs, [{connect, {Addr, Port, infinity}}|RevInstrs]);

parse_tcp([{connect, {Host, Port, To}}|Instrs], RevInstrs) 
  when integer(Port), Port > 0,
       integer(To), To > 0 ->
    {ok, Addr} = inet:getaddr(Host, inet),
    parse_tcp(Instrs, [{connect, {Addr, Port, To}}|RevInstrs]);

parse_tcp([{sleep, To}|Instrs], RevInstrs) 
  when integer(To), To > 0 ->
    parse_tcp(Instrs, [{sleep, To}|RevInstrs]);

parse_tcp([{send, Desc, Msg}|Instrs], RevInstrs) 
  when list(Desc),
       tuple(Msg) ->
    parse_tcp(Instrs, [{send, Desc, Msg}|RevInstrs]);

parse_tcp([{expect_receive, Desc, Verify}|Instrs], RevInstrs) 
  when list(Desc), 
       function(Verify) ->
    ExpRecv = {expect_receive, {Verify, infinity}},
    parse_tcp(Instrs, [ExpRecv|RevInstrs]);

parse_tcp([{expect_receive, Desc, {Verify, To}}|Instrs], RevInstrs) 
  when list(Desc), 
       function(Verify),
       integer(To), To > 0 ->
    ExpRecv = {expect_receive, Desc, {Verify, To}},
    parse_tcp(Instrs, [ExpRecv|RevInstrs]);

parse_tcp([Instr|_], _RevInstrs) ->
    throw({error, {invalid_instruction, Instr}}).



%%% ----------------------------------------------------------------

close(undefined) ->	   
    ok;
close(Sock) ->
    (catch gen_tcp:close(Sock)).

add_tpkt_header(Data) when binary(Data) ->
    L = size(Data) + 4,
    [3, 0, ((L) bsr 8) band 16#ff, (L) band 16#ff ,Data];
add_tpkt_header(IOList) when list(IOList) ->
    Binary = list_to_binary(IOList),
    L = size(Binary) + 4,
    [3, 0, ((L) bsr 8) band 16#ff, (L) band 16#ff , Binary].

sleep(X) -> receive after X -> ok end.


%%% ----------------------------------------------------------------

d(F) ->
    d(F, []).

d(F, A) ->
    d(true, F, A).

d(true, F, A) ->
    p(" DBG", F, A);
d(_, _, _) ->
    ok.

p(F, A) ->
    p("", F, A).

p(P, F, A) ->
    io:format("*** ~s~s *** " ++ F ++ "~n", [get(name),P|A]).

