/* ``The contents of this file are subject to the Erlang Public License,
 * Version 1.1, (the "License"); you may not use this file except in
 * compliance with the License. You should have received a copy of the
 * Erlang Public License along with this software. If not, it can be
 * retrieved via the world wide web at http://www.erlang.org/.
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Original Code is Ericsson Utvecklings AB.
 * Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
 * AB. All Rights Reserved.''
 * 
 *     $Id$
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "sys.h"
#include "erl_vm.h"
#include "global.h"
#include "erl_process.h"
#include "erl_db.h"
#include "beam_catches.h"
#include "erl_binary.h"
#if HIPE
#include "hipe_mode_switch.h"
#include "hipe_bif0.h"
#include "hipe_stack.h"
#endif

#ifndef HEAP_FRAG_ELIM_TEST

#if defined(SHARED_HEAP) || defined(HYBRID)
static char internal_gc_flags = 0;

/* Tell the GC to look at all processes. This flag i s set in the
 * beginning of a major collection, and cleared at the end of minor
 * collections. This makes the GC look at all processes in the major
 * collection and the following minor, which is needed since there is
 * no old generation after a major collection.
 */
#define GC_INCLUDE_ALL  0x01

/* To make a local GC know it's a part of a global */
#define GC_GLOBAL     0x02
#endif

BM_NEW_TIMER(gc);

static void remove_message_buffers(Process* p);

#define MY_IS_MOVED(x)	(!is_header((x)))

#define MOVE_CONS(PTR,CAR,HTOP,ORIG)					\
do {									\
    Eterm gval;								\
									\
    HTOP[0] = CAR;		/* copy car */				\
    HTOP[1] = PTR[1];		/* copy cdr */				\
    gval = make_list(HTOP);	/* new location */			\
    *ORIG = gval;		/* redirect original reference */	\
    PTR[0] = THE_NON_VALUE;	/* store forwarding indicator */	\
    PTR[1] = gval;		/* store forwarding address */		\
    HTOP += 2;			/* update tospace htop */		\
} while(0)

#define MOVE_BOXED(PTR,HDR,HTOP,ORIG)					\
do {									\
    Eterm gval;								\
    Sint nelts;								\
									\
    ASSERT(is_header(HDR));						\
    gval = make_boxed(HTOP);						\
    *ORIG = gval;							\
    *HTOP++ = HDR;							\
    *PTR++ = gval;							\
    nelts = header_arity(HDR);						\
    switch ((HDR) & _HEADER_SUBTAG_MASK) {				\
    case SUB_BINARY_SUBTAG: nelts++; break;				\
    case FUN_SUBTAG: nelts+=((ErlFunThing*)(PTR-1))->num_free+1; break;	\
    }									\
    while (nelts--)							\
	*HTOP++ = *PTR++;						\
} while(0)

#ifdef NOMOVE
#define BOXED_NEED(PTR,HDR)                                            \
  (((HDR) & _HEADER_SUBTAG_MASK) == SUB_BINARY_SUBTAG ?                \
    header_arity(HDR) + 2 :                                            \
   ((HDR) & _HEADER_SUBTAG_MASK) == FUN_SUBTAG ?                       \
    header_arity(HDR) + ((ErlFunThing*)(PTR))->num_free + 2 :          \
   header_arity(HDR) + 1)
#endif

/*
 * Returns number of elements in an array.
 */
#define ALENGTH(a) (sizeof(a)/sizeof(a[0]))

#ifdef DEBUG
#define HARDDEBUG  1
/* #define GC_HEAP_TRACE 1 */
/* #define GC_SWITCH_TRACE 1 */
/* #define OLD_HEAP_CREATION_TRACE 1 */
#endif

/*
 * This structure describes the rootset for the GC.
 */
typedef struct {
    Eterm* v[9];		/* Pointers to vectors with terms to GC
				 * (e.g. the stack).
				 */
    Uint sz[9];			/* Size of each vector. */
    Eterm* v_msg;		/* Pointer to messages to GC. */
    Eterm def_msg[32];		/* Default storage for messages (to avoid malloc). */
#if defined(SHARED_HEAP) || defined(HYBRID)
    Uint n;
#if defined(HIPE)
    Process *p;			/* For scanning the nstack. */
#endif
#endif
} Rootset;

/*
 * Used for printing beatiful stack dumps.
 */
extern Eterm beam_apply[];
extern Eterm beam_exit[];

#ifdef NOMOVE
/*
 * Heap pointers for the non moving generation.
 */
#define NM_NoPAGES  65536
#define NM_PAGESIZE 256
#define NM_ROOTSAVE 16384
#define NM_FULLPAGE (NM_PAGESIZE + offsetof(struct nm_page,start) / sizeof(void*))
#define NM_STORAGE_SIZE 1024

struct nm_mem_block
{
  int size;
  struct nm_mem_block *prev;
  struct nm_mem_block *next;
};

struct nm_object
{
  Eterm *this;
  int size;
  //struct nm_object *prev;
  //struct nm_object *next;
};

typedef struct nm_storage {
  struct nm_storage *next;
  struct nm_object data[NM_STORAGE_SIZE];
} nm_storage;

struct nm_page
{
  struct nm_page *prev;  /* Used only in the used lists, not in bibop */
  struct nm_page *next;
  Eterm start[1]; /* Has to be last in struct, this is where the data start */
};

Eterm *nm_heap;
Eterm *nm_hend;
static int nm_pages = NM_NoPAGES;
static Eterm* root_save[NM_ROOTSAVE];
static int roots_saved = 0;
static char *blackmap = NULL;
static char *fwdptrs = NULL;
static struct nm_page *nm_bibop = NULL;
static struct nm_page *nm_used_mem = NULL;
static struct nm_mem_block *nm_free_list = NULL;
//static struct nm_mem_block *nm_last_free = NULL;

#define MARK_FORWARD(ptr) fwdptrs[(ptr) - global_heap] = 1;

/* During the copying collection of the young generation we save all
 * live pointers to the old generation in a queue (the build
 * queue). If a collection of the old generation is needed, the queue
 * built during the previous collection of the young generation (the
 * root queue) is used as a root set together with the blue set
 * (objects allocated during the collection of the young
 * generation). At the end of a collection of the young generation,
 * the root queue is freed and the build queue becomes the new root
 * queue.
 */

/* These queues sould really be trees that do not store duplicates and
 * sort the stuff in address order (address = key) AAtree ?
 */

#define NM_STORAGE_DECLARATION(name)                                    \
    nm_storage *name##head;                                             \
    nm_storage *name##tail;                                             \
    struct nm_object *name##free;                                       \
    struct nm_object *name##last_free;                                  \
    int name##size;

#define NM_STORAGE_ITERATOR(name)                                       \
    nm_storage *name##iterator_head = name##tail;                       \
    struct nm_object *name##iterator_current = name##last_free;         \
    int name##iterator_left = name##size;

#define NM_STORAGE_INIT(name) {                                         \
    name##head = (nm_storage*)malloc(sizeof(nm_storage));               \
    name##head->next = name##head;                                      \
    name##tail = name##head;                                            \
    name##free = name##head->data;                                      \
    name##last_free = name##free + NM_STORAGE_SIZE - 1;                 \
    name##size = 0;                                                     \
}

#define NM_STORAGE_SWAP(s1,s2) {                                        \
    nm_storage *tmphead = s1##head;                                     \
    nm_storage *tmptail = s1##tail;                                     \
    struct nm_object *tmpfree = s1##free;                               \
    struct nm_object *tmplast = s1##last_free;                          \
    int tmpsize = s1##size;                                             \
    s1##head = s2##head;                                                \
    s1##tail = s2##tail;                                                \
    s1##free = s2##free;                                                \
    s1##last_free = s2##last_free;                                      \
    s1##size = s2##size;                                                \
    s2##head = tmphead;                                                 \
    s2##tail = tmptail;                                                 \
    s2##free = tmpfree;                                                 \
    s2##last_free = tmplast;                                            \
    s2##size = tmpsize;                                                 \
}

#define NM_STORAGE_NEXT(name) {                                         \
    if (name##free != name##last_free) {                                \
      name##free++;                                                     \
      if (name##free == name##head->data + NM_STORAGE_SIZE) {           \
        name##head = name##head->next;                                  \
        name##free = name##head->data;                                  \
      }                                                                 \
    } else {                                                            \
      name##free++;                                                     \
      name##tail = (nm_storage*)malloc(sizeof(nm_storage));             \
      memcpy(name##tail->data,name##head->data,                         \
             NM_STORAGE_SIZE * sizeof(struct nm_object));               \
      name##tail->next = name##head->next;                              \
      name##head->next = name##tail;                                    \
      name##last_free = ((void*)name##tail +                            \
                         ((void*)name##last_free - (void*)name##head)); \
    }                                                                   \
    name##size++;                                                       \
    if (name##free == name##head->data + NM_STORAGE_SIZE) {             \
      name##head = name##head->next;                                    \
      name##free = name##head->data;                                    \
    }                                                                   \
}

#define NM_STORAGE_GET(name) ((name##size == 0 ? 0 : (name##size--,     \
        (++name##last_free != name##tail->data + NM_STORAGE_SIZE) ?     \
        name##last_free : (name##tail = name##tail->next,               \
                           name##last_free = name##tail->data))))

#define NM_STORAGE_STEP(name) ((name##iterator_left == 0 ? 0 :          \
    (name##iterator_left--,                                             \
     (++name##iterator_current != name##iterator_head->data +           \
        NM_STORAGE_SIZE) ? name##iterator_current :                     \
          (name##iterator_head = name##iterator_head->next,             \
             name##iterator_current = name##iterator_head->data))))

#define NM_STORAGE_HEAD(name) (name##free)

#define NM_STORE(name,ptr,sz) {                                \
    NM_STORAGE_HEAD(name)->this = ptr;                         \
    NM_STORAGE_HEAD(name)->size = sz;                          \
    NM_STORAGE_NEXT(name);                                     \
}

NM_STORAGE_DECLARATION(gray);
NM_STORAGE_DECLARATION(blue);
NM_STORAGE_DECLARATION(root);
NM_STORAGE_DECLARATION(build);

static void print_heap(Eterm*,Eterm*);
static void print_nm_heap(void);
static Eterm *erts_nm_alloc(Process*, int, Eterm*, int);
static void erts_nm_copymark(Process*, Eterm*, int);

#else
#define NM_STORE(lst,ptr,sz)
#endif /* NOMOVE */

static int setup_rootset(Process*, Eterm*, int, Rootset*);
static void gen_gc(Process*, int, Eterm*, int);
static char* print_pid(Process* p);
static void sweep_proc_bins(Process *p, int fullsweep);
#ifndef SHARED_HEAP
#ifndef HYBRID /* FIND ME! */
static void sweep_proc_funs(Process *p, int fullsweep);
#endif
#endif
static void sweep_proc_externals(Process *p, int fullsweep);

#ifdef HARDDEBUG
static void check_stack(Process*, char*);
static int within(Eterm*, Process*);
void check_bins(Process *p);
int chk_sys(void);

#define CHECK(p) \
    check_stack((p), "check"); \
    check_bins(p);
#else

# define within(x,y) 1
# define CHECK(p) ((void) 1)

#endif

#define ptr_within(ptr, low, high) ((ptr) < (high) && (ptr) >= (low))

/* efficient range check */
#define in_area(ptr,start,nbytes) ((unsigned long)((char*)(ptr) - (char*)(start)) < (nbytes))

/*
 * Return the next heap size to use. Make sure we never return
 * a smaller heap size than the minimum heap size for the process.
 * (Use of the erlang:hibernate/3 BIF could have shrinked the
 * heap below the minimum heap size.)
 */
static Uint
next_heap_size(Process* p, Uint size, Uint offset)
{
    size = erts_next_heap_size(size, offset);
    return size < p->min_heap_size ? p->min_heap_size : size;
}

/*
 * Offset pointers to heap from stack.
 */

static void 
offset_heap_ptr(Eterm* hp, Uint sz, Sint offs, 
		Eterm* low, Eterm* high)
{
    while (sz--) {
	Eterm val = *hp;
	switch (primary_tag(val)) {
	  case TAG_PRIMARY_LIST:
	  case TAG_PRIMARY_BOXED: {
	      if (ptr_within(ptr_val(val), low, high)) {
		  *hp = offset_ptr(val, offs);
	      }
	      hp++;
	      continue;
	  }
	  default: {
	      hp++;
	      continue;
	  }
	}
    }
}

/*
 * Offset pointers into the heap (not stack).
 * Only offset pointers that point into the interval of low and high.
 */

static void 
offset_heap(Eterm* hp, Uint sz, Sint offs,
	    Eterm* low, Eterm* high)
{
    while (sz--) {
	Eterm val = *hp;
	switch (primary_tag(val)) {
	  case TAG_PRIMARY_LIST:
	  case TAG_PRIMARY_BOXED: {
	      if (ptr_within(ptr_val(val), low, high)) {
		  *hp = offset_ptr(val, offs);
	      }
	      hp++;
	      continue;
	  }
	  case TAG_PRIMARY_HEADER: {
	      Uint tari;

	      if (header_is_transparent(val)) {
		  hp++;
		  continue;
	      }
	      tari = thing_arityval(val);
	      switch (thing_subtag(val)) {
	      case REFC_BINARY_SUBTAG:
		  {
		      ProcBin* pb = (ProcBin*) hp;
		      Eterm** uptr = (Eterm **) &pb->next;

		      if (*uptr && ptr_within((Eterm *)pb->next, low, high)) {
			  *uptr += offs; /* Patch the mso chain */
		      }
		      sz -= tari;
		      hp += tari + 1;
		  }
		  break;
	      case FUN_SUBTAG:
		  {
#ifndef SHARED_HEAP
#ifndef HYBRID /* FIND ME! */
		      ErlFunThing* funp = (ErlFunThing *) hp;
		      Eterm** uptr = (Eterm **) &funp->next;

		      if (*uptr && ptr_within((Eterm *)funp->next, low, high)) {
			  *uptr += offs;
		      }
#endif
#endif
		      sz -= tari;
		      hp += tari + 1;
		  }
		  break;
	      case EXTERNAL_PID_SUBTAG:
	      case EXTERNAL_PORT_SUBTAG:
	      case EXTERNAL_REF_SUBTAG:
		  {
		      ExternalThing* etp = (ExternalThing *) hp;
		      Eterm** uptr = (Eterm **) &etp->next;

		      if (*uptr && ptr_within((Eterm *)etp->next, low, high)) {
			  *uptr += offs;
		      }
		      sz -= tari;
		      hp += tari + 1;
		  }
		  break;
	      default:
		  sz -= tari;
		  hp += tari + 1;
	      }
	      continue;
	  }
	  default: {
	      hp++;
	      continue;
	  }
	}
    }
}


/*
 * Offset pointers in message queue.
 */
static void
offset_mqueue(Process *p, Sint offs, Eterm* low, Eterm* high) 
{
    ErlMessage* mp = p->msg.first;

    while (mp != NULL) {
        Eterm mesg = ERL_MESSAGE_TERM(mp);
	switch (primary_tag(mesg)) {
	case TAG_PRIMARY_LIST:
	case TAG_PRIMARY_BOXED:
	    if (ptr_within(ptr_val(mesg), low, high)) {
		ERL_MESSAGE_TERM(mp) = offset_ptr(mesg, offs);
	    }
            break;
        }
	mesg = ERL_MESSAGE_TOKEN(mp);
	if (is_boxed(mesg) && ptr_within(ptr_val(mesg), low, high)) {
	    ERL_MESSAGE_TOKEN(mp) = offset_ptr(mesg, offs);
        }
        ASSERT((is_nil(ERL_MESSAGE_TOKEN(mp)) ||
		is_tuple(ERL_MESSAGE_TOKEN(mp)) ||
		is_atom(ERL_MESSAGE_TOKEN(mp))));
        mp = mp->next;
    }
}

/*
 * HiPE native code stack scanning procedures:
 * - fullsweep_nstack()
 * - gensweep_nstack()
 * - offset_nstack()
 */
#if defined(HIPE)

#if defined(__i386__)
#include "hipe_x86_asm.h"	/* for X86_NR_ARG_REGS */

#if !(defined(NOMOVE) && defined(SHARED_HEAP))
static Eterm *fullsweep_nstack(Process *p, Eterm *n_htop)
{
    Eterm *nsp;
    Eterm *nsp_end;
    const struct sdesc *sdesc;
    struct sdesc sdesc0;
    unsigned int sdesc_size;
    unsigned long ra;
    unsigned int i;
    unsigned int mask;
    char *const_start;
    unsigned long const_size;
    unsigned int nstkarity;

    nsp = p->hipe.nsp;
    nsp_end = p->hipe.nstgraylim;
    if( nsp_end ) {
	/* remove gray/white boundary trap */
	for(;;) {
	    --nsp_end;
	    if( nsp_end[0] == (unsigned long)nbif_stack_trap_ra ) {
		nsp_end[0] = (unsigned long)p->hipe.ngra;
		break;
	    }
	}
    }
    nsp_end = p->hipe.nstend;

    nstkarity = p->hipe.narity - X86_NR_ARG_REGS;
    if( (int)nstkarity < 0 )
	nstkarity = 0;
    sdesc0.summary = (0 << 9) | (0 << 8) | nstkarity;
    sdesc0.livebits[0] = ~1; /* all but RA, which is first since no locals */
    sdesc = &sdesc0;

    const_start = (char*)hipe_constants_start;
    const_size = (char*)hipe_constants_next - const_start;

    for(;;) {
	if( nsp >= nsp_end ) {
	    if( nsp == nsp_end ) {
		if( nsp ) {
		    /* see the HIGH_WATER update in fullsweep_heap() */
		    p->hipe.nstblacklim = p->hipe.nstend;
		    hipe_update_stack_trap(p, &sdesc0);
		}
		return n_htop;
	    }
	    fprintf(stderr, "%s: passed end of stack\r\n", __FUNCTION__);
	    break;
	}
	sdesc_size = sdesc_fsize(sdesc) + 1 + sdesc_arity(sdesc);
	i = 0;
	mask = sdesc->livebits[0];
	for(;;) {
	    if( mask & 1 ) {
		Eterm gval = nsp[i];
		if( is_boxed(gval) ) {
		    Eterm *ptr = boxed_val(gval);
		    Eterm val = *ptr;
		    if( MY_IS_MOVED(val) ) {
			nsp[i] = val;
		    } else if( in_area(ptr, const_start, const_size) ) {
			;
#ifdef HYBRID
                    } else if( ptr_within(ptr, global_heap, global_htop) ) {
                    } else if( ptr_within(ptr, OLD_M_DATA_START, OLD_M_DATA_END) ) {
#endif
		    } else {
			ASSERT(within(ptr, p));
			MOVE_BOXED(ptr,val,n_htop,&nsp[i]);
		    }
		} else if( is_list(gval) ) {
		    Eterm *ptr = list_val(gval);
		    Eterm val = *ptr;
		    if( is_non_value(val) ) {
			nsp[i] = ptr[1];
		    } else if( in_area(ptr, const_start, const_size) ) {
			;
#ifdef HYBRID
                    } else if( ptr_within(ptr, global_heap, global_htop) ) {
                    } else if( ptr_within(ptr, OLD_M_DATA_START, OLD_M_DATA_END) ) {
#endif
		    } else {
			ASSERT(within(ptr, p));
			MOVE_CONS(ptr,val,n_htop,&nsp[i]);
		    }
		}
	    }
	    if( ++i >= sdesc_size )
		break;
	    if( i & 31 )
		mask >>= 1;
	    else
		mask = sdesc->livebits[i >> 5];
	}
	ra = nsp[sdesc_fsize(sdesc)];
	sdesc = hipe_find_sdesc(ra);
	nsp += sdesc_size;
    }
    abort();
}
#endif /* !(NOMOVE && SHARED_HEAP) */

static void gensweep_nstack(Process *p, Eterm **ptr_old_htop, Eterm **ptr_n_htop, Eterm *objv, int nobj)
{
    Eterm *nsp;
    Eterm *nsp_end;
    const struct sdesc *sdesc;
    struct sdesc sdesc0;
    unsigned int sdesc_size;
    unsigned long ra;
    unsigned int i;
    unsigned int mask;
    Eterm *oh_start, *oh_end;
    Eterm *low_water, *high_water;
    Eterm *n_htop;
#if !(defined(NOMOVE) && defined(SHARED_HEAP))
    Eterm *old_htop;
#endif
    char *const_start;
    unsigned long const_size;
    unsigned int nstkarity;

    nsp = p->hipe.nsp;
    nsp_end = p->hipe.nstgraylim;
    if( nsp_end ) {
	/* if gray limit passed black limit, reset black limit */
	if( nsp_end > p->hipe.nstblacklim )
	    p->hipe.nstblacklim = nsp_end;
	/* remove gray/white boundary trap */
	for(;;) {
	    --nsp_end;
	    if( nsp_end[0] == (unsigned long)nbif_stack_trap_ra ) {
		nsp_end[0] = (unsigned long)p->hipe.ngra;
		break;
	    }
	}
	nsp_end = p->hipe.nstblacklim;
    } else
	nsp_end = p->hipe.nstend;

    nstkarity = p->hipe.narity - X86_NR_ARG_REGS;
    if( (int)nstkarity < 0 )
	nstkarity = 0;
    sdesc0.summary = (0 << 9) | (0 << 8) | nstkarity;
    sdesc0.livebits[0] = ~1; /* all but RA, which is first since no locals */
    sdesc = &sdesc0;

    const_start = (char*)hipe_constants_start;
    const_size = (char*)hipe_constants_next - const_start;
#if defined(NOMOVE) && defined(SHARED_HEAP)
    oh_start = nm_heap;
    oh_end = nm_hend;
#else
    oh_start = OLD_HEAP(p);
    oh_end = OLD_HEND(p);
    old_htop = *ptr_old_htop;
#endif
    low_water = HEAP_START(p);
    high_water = HIGH_WATER(p);
    n_htop = *ptr_n_htop;

    for(;;) {
	if( nsp >= nsp_end ) {
	    if( nsp == nsp_end ) {
#if !(defined(NOMOVE) && defined(SHARED_HEAP))
		*ptr_old_htop = old_htop;
#endif
		*ptr_n_htop = n_htop;
		if( nsp ) {
		    /* see the HIGH_WATER update in gen_gc() */
		    if( HEAP_START(p) != HIGH_WATER(p) ) {
			p->hipe.nstblacklim =
			    p->hipe.nstgraylim
			    ? p->hipe.nstgraylim
			    : p->hipe.nstend;
		    } else {
			/* blacklim = graylim ? blacklim : end */
			if( !p->hipe.nstgraylim )
			    p->hipe.nstblacklim = p->hipe.nstend;
		    }
		    hipe_update_stack_trap(p, &sdesc0);
		}
		return;
	    }
	    fprintf(stderr, "%s: passed end of stack\r\n", __FUNCTION__);
	    break;
	}
	sdesc_size = sdesc_fsize(sdesc) + 1 + sdesc_arity(sdesc);
	i = 0;
	mask = sdesc->livebits[0];
	for(;;) {
	    if( mask & 1 ) {
		Eterm gval = nsp[i];
		if( is_boxed(gval) ) {
		    Eterm *ptr = boxed_val(gval);
		    Eterm val = *ptr;
		    if( ptr_within(ptr, oh_start, oh_end) ) {
                        NM_STORE(build,ptr,BOXED_NEED(ptr,val));
		    } else if( in_area(ptr, const_start, const_size) ) {
#ifdef HYBRID
                    } else if( ptr_within(ptr, global_heap, global_htop) ) {
                    } else if( ptr_within(ptr, OLD_M_DATA_START, OLD_M_DATA_END) ) {
#endif
		    } else if( MY_IS_MOVED(val) ) {
			nsp[i] = val;
		    } else if( ptr_within(ptr, low_water, high_water) ) {
#if defined(NOMOVE) && defined(SHARED_HEAP)
                        Eterm *old_htop = erts_nm_alloc(p,BOXED_NEED(ptr,val),objv,nobj);
                        NM_STORE(build,old_htop,BOXED_NEED(ptr,val));
#endif
			MOVE_BOXED(ptr,val,old_htop,&nsp[i]);
		    } else {
			ASSERT(within(ptr, p));
			MOVE_BOXED(ptr,val,n_htop,&nsp[i]);
		    }
		} else if( is_list(gval) ) {
		    Eterm *ptr = list_val(gval);
		    Eterm val = *ptr;
		    if( ptr_within(ptr, oh_start, oh_end) ) {
                        NM_STORE(build,ptr,2);
		    } else if( in_area(ptr, const_start, const_size) ) {
#ifdef HYBRID
                    } else if( ptr_within(ptr, global_heap, global_htop) ) {
                    } else if( ptr_within(ptr, OLD_M_DATA_START, OLD_M_DATA_END) ) {
#endif
		    } else if( is_non_value(val) ) {
			nsp[i] = ptr[1];
		    } else if( ptr_within(ptr, low_water, high_water) ) {
#if defined(NOMOVE) && defined(SHARED_HEAP)
                        Eterm *old_htop = erts_nm_alloc(p,2,objv,nobj);
                        NM_STORE(build,old_htop,2);
#endif
			MOVE_CONS(ptr,val,old_htop,&nsp[i]);
		    } else {
			ASSERT(within(ptr, p));
			MOVE_CONS(ptr,val,n_htop,&nsp[i]);
		    }
		}
	    }
	    if( ++i >= sdesc_size )
		break;
	    if( i & 31 )
		mask >>= 1;
	    else
		mask = sdesc->livebits[i >> 5];
	}
	nsp += sdesc_size;
	ra = nsp[-(1+sdesc_arity(sdesc))];
	sdesc = hipe_find_sdesc(ra);
    }
    abort();
}

#endif	/* __i386__ */

#if defined(__sparc__)

extern void hipe_update_stack_trap(Process*, const struct sdesc*);
extern void nbif_stack_trap_ra(void);

#if !(defined(NOMOVE) && defined(SHARED_HEAP))
static Eterm *fullsweep_nstack(Process *p, Eterm *n_htop)
{
    Eterm *nsp;
    Eterm *nsp_end;
    const struct sdesc *sdesc0;
    const struct sdesc *sdesc;
    unsigned int sdesc_size;
    unsigned long ra;
    unsigned int i;
    unsigned int mask;
    char *const_start;
    unsigned long const_size;

    ra = (unsigned long)p->hipe.nra;
    if (!ra) return n_htop;

    nsp = p->hipe.nsp-1;
    nsp_end =  p->hipe.nstgraylim;

    if( nsp_end ) {
      /* Check if the trap frame is on the stack or in ra */
      if( ra == (unsigned long)nbif_stack_trap_ra ) {
	p->hipe.nra = p->hipe.ngra;
	ra = (unsigned long)p->hipe.nra;
      } else { /* The trap is on the stack, remove it. */
	/* remove gray/white boundary trap */
	for(;;) {
	  ++nsp_end;
	  /* Sanity check */
	  if(nsp_end > nsp) abort();
	  if( nsp_end[0] == (unsigned long)nbif_stack_trap_ra ) {
	    nsp_end[0] = (unsigned long)p->hipe.ngra;
	    break;
	  }
	}
      }
    }
    nsp_end = p->hipe.nstack;

    sdesc = hipe_find_sdesc(ra);
    sdesc0 = sdesc;

    const_start = (char*)hipe_constants_start;
    const_size = (char*)hipe_constants_next - const_start;

    for(;;) {
	if( nsp <= nsp_end ) {
	    if( nsp == nsp_end ) {
		if( nsp ) {
		    /* see the HIGH_WATER update in fullsweep_heap() */
		    p->hipe.nstblacklim = p->hipe.nstack;
		    hipe_update_stack_trap(p, sdesc0);
		}
		return n_htop;
	    }
	    fprintf(stderr, "%s: passed end of stack\r\n", __FUNCTION__);
	    break;
	}
	sdesc_size = sdesc_fsize(sdesc) + sdesc_arity(sdesc);
	i = 0;
	mask = sdesc->livebits[0];
	for(;;) {
	    if( mask & 1 ) {
		Eterm gval = nsp[-i];
		if( is_boxed(gval) ) {
		    Eterm *ptr = boxed_val(gval);
		    Eterm val = *ptr;
		    if( MY_IS_MOVED(val) ) {
			nsp[-i] = val;
		    } else if( in_area(ptr, const_start, const_size) ) {
			;
#ifdef HYBRID
                    } else if( ptr_within(ptr, global_heap, global_htop) ) {
                    } else if( ptr_within(ptr, OLD_M_DATA_START, OLD_M_DATA_END) ) {
#endif
		    } else {
			ASSERT(within(ptr, p));
			MOVE_BOXED(ptr,val,n_htop,&nsp[-i]);
		    }
		} else if( is_list(gval) ) {
		    Eterm *ptr = list_val(gval);
		    Eterm val = *ptr;
		    if( is_non_value(val) ) {
			nsp[-i] = ptr[1];
		    } else if( in_area(ptr, const_start, const_size) ) {
			;
#ifdef HYBRID
                    } else if( ptr_within(ptr, global_heap, global_htop) ) {
                    } else if( ptr_within(ptr, OLD_M_DATA_START, OLD_M_DATA_END) ) {
#endif
		    } else {
			ASSERT(within(ptr, p));
			MOVE_CONS(ptr,val,n_htop,&nsp[-i]);
		    }
		}
	    }
	    if( ++i >= sdesc_size )
		break;
	    if( i & 31 )
		mask >>= 1;
	    else
		mask = sdesc->livebits[i >> 5];
	}
	ra = nsp[1-sdesc_fsize(sdesc)];
	sdesc = hipe_find_sdesc(ra);
	nsp -= sdesc_size;
    }
    abort();
}
#endif /* !(NOMOVE && SHARED_HEAP) */

static void gensweep_nstack(Process *p, Eterm **ptr_old_htop, Eterm **ptr_n_htop, Eterm *objv, int nobj)
{
    Eterm *nsp;
    Eterm *nsp_end;
    const struct sdesc *sdesc0;
    const struct sdesc *sdesc;
    unsigned int sdesc_size;
    unsigned long ra;
    unsigned int i;
    unsigned int mask;
    Eterm *oh_start, *oh_end;
    Eterm *low_water, *high_water;
    Eterm *n_htop;
#if !(defined(NOMOVE) && defined(SHARED_HEAP))
    Eterm *old_htop;
#endif
    char *const_start;
    unsigned long const_size;

    nsp = p->hipe.nsp-1;
    nsp_end = p->hipe.nstgraylim;
    ra = (unsigned long)p->hipe.nra;
    if (!ra) return;


    if( nsp_end ) {
      /* if gray limit passed black limit, reset black limit */
      if( nsp_end < p->hipe.nstblacklim )
	p->hipe.nstblacklim = nsp_end;

      /* Check if the trap frame is on the stack or in ra */
      if( ra == (unsigned long)nbif_stack_trap_ra ) {
	p->hipe.nra = p->hipe.ngra;
	ra = (unsigned long)p->hipe.nra;
      } else {
	/* remove gray/white boundary trap */
	for(;;) {
	  ++nsp_end;
	  /* Sanity check */
	  if(nsp_end > nsp) abort();
	  if( nsp_end[0] == (unsigned long)nbif_stack_trap_ra ) {
	    nsp_end[0] = (unsigned long)p->hipe.ngra;
	    break;
	  }
	}
      }
      nsp_end = p->hipe.nstblacklim;
    } else
      nsp_end = p->hipe.nstack;

    sdesc = hipe_find_sdesc(ra);
    sdesc0 = sdesc;

    const_start = (char*)hipe_constants_start;
    const_size = (char*)hipe_constants_next - const_start;
#if defined(NOMOVE) && defined(SHARED_HEAP)
    oh_start = nm_heap;
    oh_end = nm_hend;
#else
    oh_start = OLD_HEAP(p);
    oh_end = OLD_HEND(p);
    old_htop = *ptr_old_htop;
#endif
    low_water = HEAP_START(p);
    high_water = HIGH_WATER(p);
    n_htop = *ptr_n_htop;

    for(;;) {
	if( nsp <= nsp_end ) {
	    if( nsp == nsp_end ) {
#if !(defined(NOMOVE) && defined(SHARED_HEAP))
		*ptr_old_htop = old_htop;
#endif
		*ptr_n_htop = n_htop;
		if( nsp ) {
		    /* see the HIGH_WATER update in gen_gc() */
		    if( HEAP_START(p) != HIGH_WATER(p) ) {
			p->hipe.nstblacklim =
			    p->hipe.nstgraylim
			    ? p->hipe.nstgraylim
			    : p->hipe.nstack;
		    } else {
			/* blacklim = graylim ? blacklim : end */
			if( !p->hipe.nstgraylim )
			    p->hipe.nstblacklim = p->hipe.nstack;
		    }
		    hipe_update_stack_trap(p, sdesc0);
		}
		return;
	    }
	    fprintf(stderr, "%s: passed end of stack\r\n", __FUNCTION__);

	    break;
	}
	sdesc_size = sdesc_fsize(sdesc) + sdesc_arity(sdesc);
	i = 0;
	mask = sdesc->livebits[0];
	for(;;) {
	    if( mask & 1 ) {
		Eterm gval = nsp[-i];
		if( is_boxed(gval) ) {
		    Eterm *ptr = boxed_val(gval);
		    Eterm val = *ptr;
		    if( ptr_within(ptr, oh_start, oh_end) ) {
                        NM_STORE(build,ptr,BOXED_NEED(ptr,val));
		    } else if( in_area(ptr, const_start, const_size) ) {
#ifdef HYBRID
                    } else if( ptr_within(ptr, global_heap, global_htop) ) {
                    } else if( ptr_within(ptr, OLD_M_DATA_START, OLD_M_DATA_END) ) {
#endif
		    } else if( MY_IS_MOVED(val) ) {
			nsp[-i] = val;
		    } else if( ptr_within(ptr, low_water, high_water) ) {
#if defined(NOMOVE) && defined(SHARED_HEAP)
                        Eterm *old_htop = erts_nm_alloc(p,BOXED_NEED(ptr,val),objv,nobj);
                        NM_STORE(build,old_htop,BOXED_NEED(ptr,val));
#endif
			MOVE_BOXED(ptr,val,old_htop,&nsp[-i]);
		    } else {
			ASSERT(within(ptr, p));
			MOVE_BOXED(ptr,val,n_htop,&nsp[-i]);
		    }
		} else if( is_list(gval) ) {
		    Eterm *ptr = list_val(gval);
		    Eterm val = *ptr;
		    if( ptr_within(ptr, oh_start, oh_end) ) {
                        NM_STORE(build,ptr,2);
		    } else if( in_area(ptr, const_start, const_size) ) {
#ifdef HYBRID
                    } else if( ptr_within(ptr, global_heap, global_htop) ) {
                    } else if( ptr_within(ptr, OLD_M_DATA_START, OLD_M_DATA_END) ) {
#endif
		    } else if( is_non_value(val) ) {
			nsp[-i] = ptr[1];
		    } else if( ptr_within(ptr, low_water, high_water) ) {
#if defined(NOMOVE) && defined(SHARED_HEAP)
                        Eterm *old_htop = erts_nm_alloc(p,2,objv,nobj);
                        NM_STORE(build,old_htop,2);
#endif
			MOVE_CONS(ptr,val,old_htop,&nsp[-i]);
		    } else {
			ASSERT(within(ptr, p));
			MOVE_CONS(ptr,val,n_htop,&nsp[-i]);
		    }
		}
	    }
	    if( ++i >= sdesc_size )
		break;
	    if( i & 31 )
		mask >>= 1;
	    else
		mask = sdesc->livebits[i >> 5];
	}
	nsp -= sdesc_size;
	ra = nsp[1+sdesc_arity(sdesc)];
	sdesc = hipe_find_sdesc(ra);
    }
    abort();
}

#endif	/* __sparc__ */

#define GENSWEEP_NSTACK(p,old_htop,n_htop,objv,nobj)             	\
	do {								\
		Eterm *tmp_old_htop = old_htop;				\
		Eterm *tmp_n_htop = n_htop;				\
		gensweep_nstack((p), &tmp_old_htop, &tmp_n_htop,objv,nobj); \
		old_htop = tmp_old_htop;				\
		n_htop = tmp_n_htop;					\
	} while(0)

/*
 * offset_nstack() can ignore the descriptor-based traversal the other
 * nstack procedures use and simply call offset_heap_ptr() instead.
 * This relies on two facts:
 * 1. The only live non-Erlang terms on an nstack are return addresses,
 *    and they will be skipped thanks to the low/high range check.
 * 2. Dead values, even if mistaken for pointers into the low/high area,
 *    can be offset safely since they won't be dereferenced.
 *
 * XXX: WARNING: If HiPE starts storing other non-Erlang values on the
 * nstack, such as floats, then this will have to be changed.
 */
#define offset_nstack(p,offs,low,high) offset_heap_ptr(hipe_nstack_start((p)),hipe_nstack_used((p)),(offs),(low),(high))

#else /* !HIPE */

#if !(defined(NOMOVE) && defined(SHARED_HEAP))
#define fullsweep_nstack(p,n_htop)		(n_htop)
#endif
#define GENSWEEP_NSTACK(p,old_htop,n_htop,objv,nobj)	do{}while(0)
#define offset_nstack(p,offs,low,high)		do{}while(0)

#endif /* HIPE */

static int
setup_rootset(Process *p, Eterm *objv, int nobj, Rootset *rootset)
{
    int n;
    ErlMessage* mp;
    Eterm* v_ptr;
    int v_msg_len;

    v_msg_len = 2 * p->msg.len;

    /*
     * Move pointers for all messages into an array pointed to by p->v_msg.
     */
    if (v_msg_len > ALENGTH(rootset->def_msg)) {
        rootset->v_msg = (Eterm *)
	    erts_alloc(ERTS_ALC_T_MSG_ROOTS, sizeof(Eterm) * v_msg_len);
    } else {
        rootset->v_msg = rootset->def_msg;
    }
    mp = p->msg.first;
    v_ptr = rootset->v_msg;
    while (mp != NULL) {
        *v_ptr++ = ERL_MESSAGE_TERM(mp);
        ASSERT((is_nil(ERL_MESSAGE_TOKEN(mp)) ||
		is_tuple(ERL_MESSAGE_TOKEN(mp)) ||
		is_atom(ERL_MESSAGE_TOKEN(mp))));
        *v_ptr++ = ERL_MESSAGE_TOKEN(mp);
        mp = mp->next;
    }

    n = 0;
    rootset->v[n]  = p->stop;
    rootset->sz[n] = STACK_START(p) - p->stop;
    ++n;

    if (p->dictionary != NULL) {
        rootset->v[n]  = p->dictionary->data;
        rootset->sz[n] = p->dictionary->used;
        ++n;
    }
    if (p->debug_dictionary != NULL) {
        rootset->v[n]  = p->debug_dictionary->data;
        rootset->sz[n] = p->debug_dictionary->used;
        ++n;
    }
    rootset->v[n]  = rootset->v_msg;
    rootset->sz[n] = v_msg_len;
    ++n;
    if (nobj > 0) {
        rootset->v[n]  = objv;
        rootset->sz[n] = nobj;
        ++n;
    }

    ASSERT((is_nil(p->seq_trace_token) 
	    || is_tuple(p->seq_trace_token)
	    || is_atom(p->seq_trace_token)));
    rootset->v[n] = &p->seq_trace_token;
    rootset->sz[n] = 1;
    n++;

    ASSERT(is_nil(p->tracer_proc)
	   || is_internal_pid(p->tracer_proc)
	   || is_internal_port(p->tracer_proc));

    ASSERT(is_pid(p->group_leader));
    rootset->v[n]  = &p->group_leader;
    rootset->sz[n] = 1;
    ++n;

    /*
     * The process may be garbage-collected while it is terminating.
     * (fvalue contains EXIT reason.)
     */
    rootset->v[n]  = &p->fvalue;
    rootset->sz[n] = 1;
    n++;

#if HIPE && (defined(SHARED_HEAP) || defined(HYBRID))
    rootset->p = p;
#endif

    ASSERT(n <= ALENGTH(rootset->v));
    return n;
}

#if defined(SHARED_HEAP) || defined(HYBRID)
static Uint
collect_roots(Process* current, Eterm *objv, int nobj, Rootset rootset[])
{
    Process* p;
    Uint i, j = 0;
    Uint n = erts_num_active_procs;

    for (i = 0; i < n; i++) {
        p = erts_active_procs[i];
	if (p->active || internal_gc_flags & GC_INCLUDE_ALL) {
	    if (p == current) {
		rootset[j].n = setup_rootset(p, objv, nobj, &rootset[j]);
	    } else {
		rootset[j].n = setup_rootset(p, p->arg_reg, p->arity, &rootset[j]);
	    }
	    j++;
	}
        else
          printf("Skipped rootset!!\n\r");
    }
    return j;
}
#endif

static ERTS_INLINE
void restore_this_rootset(Process *p, Rootset *rootset)
{
    Eterm* v_ptr;
    ErlMessage* mp;

    /*
     * Restore all message pointers.
     */
    mp = p->msg.first;
    v_ptr = rootset->v_msg;
    while (mp != NULL) {
	ERL_MESSAGE_TERM(mp) = *v_ptr++;
	ASSERT((is_nil(*v_ptr) || is_tuple(*v_ptr) || is_atom(*v_ptr)));
	ERL_MESSAGE_TOKEN(mp) = *v_ptr++;
	mp = mp->next;
    }
    
    if (rootset->v_msg != rootset->def_msg) {
        erts_free(ERTS_ALC_T_MSG_ROOTS, rootset->v_msg);
    }
}

#if defined(HYBRID)
static void
restore_rootset(Process *p, Rootset *rootset)
{
    if (internal_gc_flags & GC_GLOBAL)
    {
        Uint i, j = 0;
        Uint n = erts_num_active_procs;

        for (i = 0; i < n; i++) {
            p = erts_active_procs[i];
            if (p->active) {
                restore_this_rootset(p, &rootset[j++]);
            }
        }
    }
    else
        restore_this_rootset(p, rootset);
}
#else
static void
restore_rootset(Process *p, Rootset *rootset)
{
#ifdef SHARED_HEAP
    Uint i, j = 0;
    Uint n = erts_num_active_procs;

    for (i = 0; i < n; i++) {
        p = erts_active_procs[i];
	if (p->active) {
	    restore_this_rootset(p, &rootset[j++]);
	}
    }
#else
    restore_this_rootset(p, rootset);
#endif
}
#endif /* HYBRID */

/*
 * Remove all message buffers.
 */
static void
remove_message_buffers(Process* p)
{
    ErlHeapFragment* bp = MBUF(p);

    MBUF(p) = NULL;
#ifndef SHARED_HEAP
    p->halloc_mbuf = NULL;
#endif
    MBUF_SIZE(p) = 0;
    while (bp != NULL) {
	ErlHeapFragment* next_bp = bp->next;
#ifdef DEBUG
	sys_memset(bp->mem, 0xf9, bp->size*sizeof(Eterm));
#endif 
	free_message_buffer(bp);
	bp = next_bp;
    }
}

#if !(defined(NOMOVE) && defined(SHARED_HEAP))
/* 
** This function is used when using fullsweep gc.
** This replaces gen_gc for processes that use the
** fullsweep algorithm and don't have an old_heap.
** Parameters:
** p: The process who is being garbage collected.
** new_sz: The wanted size of the heap after collecting.
** objv: Vector of "extra" objects to be "saved".
** nobj: Number of objects in objv.
*/
static void fullsweep_heap(Process *p, int new_sz, Eterm* objv, int nobj)
{
#ifdef SHARED_HEAP
    Rootset *rootset = erts_alloc(ERTS_ALC_T_ROOTSET,
				  sizeof(Rootset)*erts_max_processes);
#else
    Rootset rootset;            /* Rootset for GC (stack, dictionary, etc). */
#endif
    Eterm* n_hstart;		/* Start of new heap */
    Eterm* n_htop;		/* Top of new heap */
    Eterm* n_heap;		/* The new heap */
    Eterm* n_hp;
    int n;
#if HIPE
    char *const_start = (char*)hipe_constants_start;
    unsigned long const_size = (char*)hipe_constants_next - const_start;
#endif
#ifdef HYBRID
    Eterm *g_heap  = global_heap;
    Eterm *g_htop  = global_htop;
    Eterm *go_heap = OLD_M_DATA_START;
    Eterm *go_htop = OLD_M_DATA_END;
#endif

    /* Create new, empty heap */
    n_heap = (Eterm *) ERTS_HEAP_ALLOC(ERTS_ALC_T_HEAP, sizeof(Eterm)*new_sz);

    n_hstart = n_htop = n_heap;
    FLAGS(p) &= ~F_NEED_FULLSWEEP;

#ifdef GC_HEAP_TRACE
    fprintf(stderr, "Fullsweep GC proc: %d\n", print_pid(p));
#endif

#ifdef SHARED_HEAP
    n = collect_roots(p, objv, nobj, rootset);
    while (n--) {
	n_htop = fullsweep_nstack(rootset[n].p, n_htop);
      while (rootset[n].n--) {
        Eterm* g_ptr = rootset[n].v[rootset[n].n];
        Uint  g_sz = rootset[n].sz[rootset[n].n];
#else
    n = setup_rootset(p, objv, nobj, &rootset);
    n_htop = fullsweep_nstack(p, n_htop);
    while (n--) {
        Eterm* g_ptr = rootset.v[n];
        Eterm g_sz = rootset.sz[n];
#endif
	while(g_sz--) {
	    Eterm* ptr;
	    Eterm val;
	    Eterm gval = *g_ptr;

	    switch (primary_tag(gval)) {

	      case TAG_PRIMARY_BOXED: {
		ptr = boxed_val(gval);
		val = *ptr;
		if (MY_IS_MOVED(val)) {
		    ASSERT(is_boxed(val));
		    *g_ptr++ = val;
                }
#ifdef HYBRID
                else if( ptr_within(ptr, g_heap, g_htop) )
                {
                    ++g_ptr;
                }
                else if( ptr_within(ptr, go_heap, go_htop) )
                {
                    ++g_ptr;
                }
#endif
#if HIPE
		else if( in_area(ptr, const_start, const_size) ) {
		    ++g_ptr;
                }
#endif
		else {
		    ASSERT(within(ptr, p));
		    MOVE_BOXED(ptr,val,n_htop,g_ptr++);
		}
		continue;
	      }

	      case TAG_PRIMARY_LIST: {
		ptr = list_val(gval);
		val = *ptr;
		if (is_non_value(val))
		    *g_ptr++ = ptr[1];
#ifdef HYBRID
                else if( ptr_within(ptr, g_heap, g_htop) )
                {
                    ++g_ptr;
                }
                else if( ptr_within(ptr, go_heap, go_htop) )
                {
                    ++g_ptr;
                }
#endif
#if HIPE
		else if( in_area(ptr, const_start, const_size) )
		    ++g_ptr;
#endif
		else {
		    ASSERT(within(ptr, p));
		    MOVE_CONS(ptr,val,n_htop,g_ptr++);
		}
		continue;
	      }

	      default: {
		g_ptr++;
		continue;
	      }
	    }
	}
    }
#ifdef SHARED_HEAP
    }
#endif

    /*
     * Now all references on the stack point to the new heap. However,
     * most references on the new heap point to the old heap so the next stage
     * is to scan through the new heap evacuating data from the old heap
     * until all is copied.
     */

	n_hp = n_heap;
    
	while (n_hp != n_htop) {
	Eterm* ptr;
	Eterm val;
	Eterm gval = *n_hp;

	switch (primary_tag(gval)) {

	  case TAG_PRIMARY_BOXED: {
	    ptr = boxed_val(gval);
	    val = *ptr;
	    if (MY_IS_MOVED(val)) {
		ASSERT(is_boxed(val));
		*n_hp++ = val;
            }
#ifdef HYBRID
            else if( ptr_within(ptr, g_heap, g_htop) )
            {
                ++n_hp;
            }
            else if( ptr_within(ptr, go_heap, go_htop) )
            {
                ++n_hp;
            }
#endif
#if HIPE
	    else if( in_area(ptr, const_start, const_size) ) {
		++n_hp;
            }
#endif
	    else {
		ASSERT(within(ptr, p));
		MOVE_BOXED(ptr,val,n_htop,n_hp++);
	    }
	    continue;
	  }

	  case TAG_PRIMARY_LIST: {
	    ptr = list_val(gval);
	    val = *ptr;
	    if (is_non_value(val)) {
		*n_hp++ = ptr[1];
            }
#ifdef HYBRID
            else if( ptr_within(ptr, g_heap, g_htop) )
            {
                ++n_hp;
            }
            else if( ptr_within(ptr, go_heap, go_htop) )
            {
                ++n_hp;
            }
#endif
#if HIPE
	    else if( in_area(ptr, const_start, const_size) ) {
		++n_hp;
            }
#endif
	    else {
		ASSERT(within(ptr, p));
		MOVE_CONS(ptr,val,n_htop,n_hp++);
	    }
	    continue;
	  }

	  case TAG_PRIMARY_HEADER: {
	      if (header_is_thing(gval))
		  n_hp += (thing_arityval(gval)+1);
	      else
		  n_hp++;
	      continue;
	  }

	  default: {
	    n_hp++;
	    continue;
	  }
	}
    }

#ifdef SHARED_HEAP
    restore_rootset(p, rootset);
    erts_free(ERTS_ALC_T_ROOTSET, (void *) rootset);
#else
    restore_rootset(p, &rootset);
#endif
    if (MSO(p).mso) {
	sweep_proc_bins(p, 1);
    }
#ifndef SHARED_HEAP
#ifndef HYBRID /* FIND ME! */
    if (MSO(p).funs) {
	sweep_proc_funs(p, 1);
    }
#endif
#endif
    if (MSO(p).externals) {
	sweep_proc_externals(p, 1);
    }

    remove_message_buffers(p);

    if (OLD_HEAP(p) != NULL) {
#ifdef DEBUG
        sys_memset(OLD_HEAP(p), 0xff,
                   (OLD_HEND(p) - OLD_HEAP(p)) * sizeof(Eterm));
#endif
        ERTS_HEAP_FREE(ERTS_ALC_T_OLD_HEAP,
		       OLD_HEAP(p),
		       (OLD_HEND(p)-OLD_HEAP(p))*sizeof(Eterm));
        OLD_HEAP(p) = OLD_HTOP(p) = OLD_HEND(p) = NULL;
    }

#ifndef SHARED_HEAP
    /* Move the stack, the beam stack is "in the heap" */
    n = HEAP_END(p) - p->stop;
    sys_memcpy(n_heap + new_sz - n, p->stop, n * sizeof(Eterm));
#endif
#ifdef DEBUG
    sys_memset(HEAP_START(p), 0xff,
               (HEAP_END(p) - HEAP_START(p)) * sizeof(Eterm));
#endif

    ERTS_HEAP_FREE(ERTS_ALC_T_HEAP,
		   (void *) HEAP_START(p),
		   HEAP_SIZE(p)*sizeof(Eterm));

    HEAP_START(p) = n_heap;
    HEAP_TOP(p) = n_htop;
    HEAP_SIZE(p) = new_sz;
    HEAP_END(p) = n_heap + new_sz;
#ifndef SHARED_HEAP
    p->stop = HEAP_END(p) - n;
#endif
    GEN_GCS(p) = 0;

    /*
     * Should we set p->high_water to p->heap or p->htop?
     *
     * Setting it to p->htop means that any surviving data will be
     * placed on the old heap in the next garbage collection.
     * This setting gives a better estone value, but one can assume
     * that more garbage is placed on the old heap.
     *
     * Setting it to p->heap means that two more garbage collections
     * are needed to move data to the old heap.
     */

    HIGH_WATER(p) = HEAP_TOP(p);
}
#endif /* !(NOMOVE && SHARED_HEAP) */

static void
offset_off_heap(Process* p, Eterm* low, Eterm* high, int offs)
{
    if (MSO(p).mso && ptr_within((Eterm *)MSO(p).mso, low, high)) {
        Eterm** uptr = (Eterm**) &MSO(p).mso;
        *uptr += offs;
    }

#ifndef SHARED_HEAP
#ifndef HYBRID /* FIND ME! */
    if (MSO(p).funs && ptr_within((Eterm *)MSO(p).funs, low, high)) {
        Eterm** uptr = (Eterm**) &MSO(p).funs;
        *uptr += offs;
    }
#endif
#endif

    if (MSO(p).externals && ptr_within((Eterm *)MSO(p).externals, low, high)) {
        Eterm** uptr = (Eterm**) &MSO(p).externals;
        *uptr += offs;
    }
}

static void
offset_rootset(Process *p, int offs, 
	       Eterm* low, Eterm* high, 
	       Eterm* objv, int nobj)
{
#ifdef SHARED_HEAP
    Uint i;
    Uint n = erts_num_active_procs;
    Process* current = p;

    for (i = 0; i < n; i++) {
        p = erts_active_procs[i];
        if (p->dictionary) {
            offset_heap(p->dictionary->data, 
                        p->dictionary->used, 
                        offs, low, high);
	}
        if (p->debug_dictionary) {
            offset_heap(p->debug_dictionary->data, 
                        p->debug_dictionary->used, 
                        offs, low, high);
	}
	offset_heap(&p->fvalue, 1, offs, low, high);
	offset_heap(&p->group_leader, 1, offs, low, high);
        offset_heap(&p->seq_trace_token, 1, offs, low, high);
        offset_mqueue(p, offs, low, high);
        offset_heap_ptr(p->stop, (STACK_START(p) - p->stop), offs, low, high);
        offset_nstack(p, offs, low, high);
	offset_off_heap(p, low, high, offs);
	if (p != current) {
	    offset_heap(p->arg_reg, p->arity, offs, low, high);
	} else if (nobj > 0) {
	    offset_heap(objv, nobj, offs, low, high);
	}
    }
#else
    if (p->dictionary) 
	offset_heap(p->dictionary->data, 
		    p->dictionary->used, 
		    offs, low, high);
    if (p->debug_dictionary) 
	offset_heap(p->debug_dictionary->data, 
		    p->debug_dictionary->used, 
		    offs, low, high);
    offset_heap(&p->fvalue, 1, offs, low, high);
    offset_heap(&p->group_leader, 1, offs, low, high);
    offset_heap(&p->seq_trace_token, 1, offs, low, high);
    offset_mqueue(p, offs, low, high);
    offset_heap_ptr(p->stop, (STACK_START(p) - p->stop), offs, low, high);
    offset_nstack(p, offs, low, high);
    if (nobj > 0) {
	offset_heap(objv, nobj, offs, low, high);
    }
    offset_off_heap(p, low, high, offs);
#endif
}

/*
 * Grow the new heap size to 'new_sz'.
 */
static void
grow_new_heap(Process *p, Uint new_sz, Eterm* objv, int nobj)
{
    Eterm* new_heap;
    int heap_size = HEAP_TOP(p) - HEAP_START(p);
#ifndef SHARED_HEAP
    int stack_size = p->hend - p->stop;
#endif
    Sint offs;

    ASSERT(HEAP_SIZE(p) < new_sz);
    new_heap = (Eterm *) ERTS_HEAP_REALLOC(ERTS_ALC_T_HEAP,
					   (void*)HEAP_START(p),
					   sizeof(Eterm)*(HEAP_SIZE(p)),
					   sizeof(Eterm)*new_sz);

#ifdef GC_HEAP_TRACE
    fprintf(stderr, "grow_new_heap: GREW %s FROM %d UPTO %d (used %d)\n",
            print_pid(p), HEAP_SIZE(p), new_sz, heap_size);
#endif

    if ((offs = new_heap - HEAP_START(p)) == 0) { /* No move. */
        HEAP_END(p) = new_heap + new_sz;
#ifndef SHARED_HEAP
        sys_memmove(p->hend - stack_size, p->stop, stack_size * sizeof(Eterm));
        p->stop = p->hend - stack_size;
#endif
    } else {
#ifndef SHARED_HEAP
        Eterm* prev_stop = p->stop;
#endif

        offset_heap(new_heap, heap_size, offs, HEAP_START(p), HEAP_TOP(p));
        HIGH_WATER(p) = new_heap + (HIGH_WATER(p) - HEAP_START(p));

        HEAP_END(p) = new_heap + new_sz;
#ifndef SHARED_HEAP
        prev_stop = new_heap + (p->stop - p->heap);
        p->stop = p->hend - stack_size;
        sys_memmove(p->stop, prev_stop, stack_size * sizeof(Eterm));
#endif
        offset_rootset(p, offs, HEAP_START(p), HEAP_TOP(p), objv, nobj);
        HEAP_TOP(p) = new_heap + heap_size;
        HEAP_START(p) = new_heap;
    }
    HEAP_SIZE(p) = new_sz;
}

void
erts_shrink_new_heap(Process *p, Uint new_sz, Eterm *objv, int nobj)
{
    Eterm* new_heap;
    int heap_size = HEAP_TOP(p) - HEAP_START(p);
    Sint offs;

#ifdef SHARED_HEAP
    ASSERT(new_sz < HEAP_SIZE(p));
    new_heap = (Eterm *) ERTS_HEAP_REALLOC(ERTS_ALC_T_HEAP,
					   (void*)HEAP_START(p),
					   sizeof(Eterm)*(HEAP_SIZE(p)),
					   sizeof(Eterm)*new_sz);
    HEAP_END(p) = new_heap + new_sz;
#else
    int stack_size = p->hend - p->stop;

    ASSERT(new_sz < p->heap_sz);
    sys_memmove(p->heap + new_sz - stack_size, p->stop, stack_size *
                                                        sizeof(Eterm));
    new_heap = (Eterm *) ERTS_HEAP_REALLOC(ERTS_ALC_T_HEAP,
					   (void*)p->heap,
					   sizeof(Eterm)*(HEAP_SIZE(p)),
					   sizeof(Eterm)*new_sz);
    p->hend = new_heap + new_sz;
    p->stop = p->hend - stack_size;
#endif

#ifdef GC_HEAP_TRACE
    fprintf(stderr,
	    "shrink_new_heap: SHRINKED %s FROM %d DOWNTO %d (used %d)\n",
            print_pid(p), HEAP_SIZE(p), new_sz, heap_size);
#endif

    if ((offs = new_heap - HEAP_START(p)) != 0) {

        /*
         * Normally, we don't expect a shrunk heap to move, but you never
         * know on some strange embedded systems...  Or when using purify.
         */

        offset_heap(new_heap, heap_size, offs, HEAP_START(p), HEAP_TOP(p));
        HIGH_WATER(p) = new_heap + (HIGH_WATER(p) - HEAP_START(p));
        offset_rootset(p, offs, HEAP_START(p), HEAP_TOP(p), objv, nobj);
        HEAP_TOP(p) = new_heap + heap_size;
        HEAP_START(p) = new_heap;
    }
    HEAP_SIZE(p) = new_sz;
}

#if !(defined(NOMOVE) && defined(SHARED_HEAP))
static void
adjust_after_fullsweep(Process *p, int size_before, int need, Eterm *objv, int nobj)
{
    int wanted, sz, size_after, need_after;
#ifdef SHARED_HEAP
    int stack_size = 0;        /* Size of stack ON HEAP */
#else
    int stack_size =  p->hend - p->stop;
#endif
    
    size_after = (HEAP_TOP(p) - HEAP_START(p));
    reclaimed += (size_before - size_after);
    
    /*
     * Resize the heap if needed.
     */

    need_after = size_after + need + stack_size;
    if (HEAP_SIZE(p) < need_after) {
        /* Too small - grow to match requested need */
        sz = next_heap_size(p, need_after, 0);
        grow_new_heap(p, sz, objv, nobj);
    } else if (3 * HEAP_SIZE(p) < 4 * need_after){
        /* Need more than 75% of current, postpone to next GC.*/
        FLAGS(p) |= F_HEAP_GROW;
    } else if (4 * need_after < HEAP_SIZE(p) && HEAP_SIZE(p) > H_MIN_SIZE){
        /* We need less than 25% of the current heap, shrink.*/
        /* XXX - This is how it was done in the old GC:
           wanted = 4 * need_after;
           I think this is better as fullsweep is used mainly on
           small memory systems, but I could be wrong... */
        wanted = 2 * need_after;
        if (wanted < p->min_heap_size) {
            sz = p->min_heap_size;
        } else {
            sz = next_heap_size(p, wanted, 0);
        }
        if (sz < HEAP_SIZE(p)) {
            erts_shrink_new_heap(p, sz, objv, nobj);
        }
    }
}
#endif /* !(NOMOVE && SHARED_HEAP) */

/*
** Garbage collect a process.
** Parameters:
** p: Pointer to the process structure.
** need: Number of (erlang) words needed on the heap.
** objv: Array of terms to add to rootset, that is to preserve.
** nobj: Number of objects in objv.
*/
int
erts_garbage_collect(Process* p, int need, Eterm* objv, int nobj)
{
    int size_before;
    int size_after;
    int need_after;
    Uint saved_status;
    int wanted;
    int stack_size;             /* Size of stack ON HEAP. */
#if !(defined(NOMOVE) && defined(SHARED_HEAP))
    int sz;
#ifdef __BENCHMARK__
    uint was_this_major = 0;
#endif
#endif
    Uint ms1, s1, us1;

    //BM_NEW_TIMER(gc);

#if defined(BM_TIMERS) && defined(HYBRID)
    if (!(internal_gc_flags & GC_GLOBAL))
        BM_STOP_TIMER(system);
#else
    BM_STOP_TIMER(system);
#endif

#ifdef BM_HEAP_SIZES
    {
        double total_used_heap = 0;
#ifdef SHARED_HEAP
        total_used_heap = (HEAP_TOP(p) - HEAP_START(p)) +
                          (OLD_HTOP(p) - OLD_HEAP(p)) + MBUF_SIZE(p);
        if (total_used_heap > max_used_global_heap)
            max_used_global_heap = total_used_heap;
#else
        int i;
        for (i = 0; i < erts_max_processes; i++)
        {
            Process *cp = process_tab[i];
            if (cp == NULL) continue;

            total_used_heap += (cp->htop - cp->heap) + cp->mbuf_sz +
                               (cp->old_htop - cp->old_heap);
        }
        if (total_used_heap > max_used_heap)
            max_used_heap = total_used_heap;
#endif
    }
#endif /* BM_HEAP_SIZES */

#ifdef SHARED_HEAP
    /* There are collections forced by the sheduler when the size of all
     * heap fragments and the overhead from binaries gets greater than the
     * heap size.
     * This is not the way to do it in the Shared heap...
     */
    if (need == 0 && nobj == 0 && !(FLAGS(p) & F_NEED_FULLSWEEP))
    {
        BM_SWAP_TIMER(gc,system);
        return 0;
    }
#endif

    if (SAVED_HEAP_TOP(p) != NULL) {
	HEAP_TOP(p) = SAVED_HEAP_TOP(p);
	SAVED_HEAP_TOP(p) = NULL;
    }

#ifdef SHARED_HEAP
#define OverRunCheck() \
    if (HEAP_END(p) < HEAP_TOP(p)) { \
        erl_exit(1, "%s: Overrun heap at line %d\n", print_pid(p),__LINE__); \
    }
#else
#define OverRunCheck() \
    if (p->stop < p->htop) { \
        erl_exit(1, "%s: Overrun stack and heap at line %d\n", print_pid(p),__LINE__); \
    }
#endif

#if defined(BM_TIMERS) && defined(HYBRID)
    if (!(internal_gc_flags & GC_GLOBAL))
    {
        gc_time = 0; /* FIND ME! */
    }
#endif

    if (IS_TRACED_FL(p, F_TRACE_GC)) {
        trace_gc(p, am_gc_start);
    }
    BM_START_TIMER(gc);

    if (erts_system_monitor_long_gc != 0) get_now(&ms1, &s1, &us1);
    saved_status = p->status;
    p->status = P_GARBING;
    CHECK(p);
    OverRunCheck();

#if !(defined(NOMOVE) && defined(SHARED_HEAP))
    if (GEN_GCS(p) >= MAX_GEN_GCS(p)) {
        FLAGS(p) |= F_NEED_FULLSWEEP;
    }
#endif

#ifndef SHARED_HEAP
    stack_size = p->hend - p->stop;
#else
    stack_size = 0;		/* No stack in the heap. */
#if 0
    n = erts_num_active_procs;
    for (i = 0; i < n; i++) {
        Process *p = erts_active_procs[i];
        ARITH_AVAIL(p) = 0;
        ARITH_HEAP(p) = NULL;
#ifdef DEBUG
        ARITH_CHECK_ME(p) = NULL;
#endif
    }
#endif
#endif

    MSO(p).overhead = 0;
    garbage_cols++;

    /* Size of heap before first GC */
    size_before = MBUF_SIZE(p) + (HEAP_TOP(p) - HEAP_START(p));

#ifdef SHARED_HEAP
    /* To prevent shrinking of the shared heap */
    p->min_heap_size = HEAP_SIZE(p);
#endif

#if !(defined(NOMOVE) && defined(SHARED_HEAP))
    /*
     * Generational GC from here on. We need an old heap.
     */

    if (OLD_HEAP(p) == NULL && HIGH_WATER(p) != HEAP_START(p) &&
        (FLAGS(p) & F_NEED_FULLSWEEP) == 0) {
        Eterm* n_old;
        /* Note: We choose a larger heap size than strictly needed,
         * which seems to reduce the number of fullsweeps.
         * This improved Estone by more than 1200 estones on my computer
         * (Ultra Sparc 10).
         */
        size_t new_sz = next_heap_size(p, HIGH_WATER(p) - HEAP_START(p), 1);

        /* Create new, empty old_heap */
        n_old = (Eterm *) ERTS_HEAP_ALLOC(ERTS_ALC_T_OLD_HEAP,
					  sizeof(Eterm)*new_sz);

        OLD_HEND(p) = n_old + new_sz;
        OLD_HEAP(p) = OLD_HTOP(p) = n_old;
#ifdef OLD_HEAP_CREATION_TRACE
#ifdef SHARED_HEAP
        fprintf(stderr, "Created a global old_heap because of %s.\r\n",
            print_pid(p));
#else
        fprintf(stderr, "Created old_heap for %s.\r\n", print_pid(p));
#endif
#endif
    }
#endif /* !(NOMOVE && SHARED_HEAP) */

#if !(defined(NOMOVE) && defined(SHARED_HEAP))
    /*
     * Try a generational GC if the old heap is large enough.
     */

    if ((FLAGS(p) & F_NEED_FULLSWEEP) == 0 &&
        HIGH_WATER(p) - HEAP_START(p) <= OLD_HEND(p) - OLD_HTOP(p)) {
#endif /* !(NOMOVE && SHARED_HEAP) */

#if defined(NOMOVE) && defined(SHARED_HEAP)
        /*
         * There is allways space left in the old generation since it
         * has its own GC!
         */

        gen_gc(p, next_heap_size(p, HEAP_SIZE(p) + MBUF_SIZE(p) + need, 0),
               objv, nobj);
#else
        /*
         * There is space enough in old_heap for everything
         * below the high water mark.  Do a generational GC.
         */

        gen_gc(p, next_heap_size(p, HEAP_SIZE(p) + MBUF_SIZE(p), 0),
               objv, nobj);
#endif

        GEN_GCS(p)++;
        size_after = HEAP_TOP(p) - HEAP_START(p);
        need_after = size_after + need + stack_size;
        reclaimed += (size_before - size_after);

        /*
         * Excessively large heaps should be shrunk, but
         * don't even bother on reasonable small heaps.
         *
         * The reason for this is that after tenuring, we often
         * use a really small portion of new heap, therefore, unless
         * the heap size is substantial, we don't want to shrink.
         */

        if ((HEAP_SIZE(p) > 300) && (4 * need_after < HEAP_SIZE(p)) &&
            ((HEAP_SIZE(p) > 8000)
#if !(defined(NOMOVE) && defined(SHARED_HEAP))
             || (HEAP_SIZE(p) > (OLD_HEND(p) - OLD_HEAP(p)))
#endif
             )) {
            wanted = 3 * need_after;
            if (wanted < p->min_heap_size) {
                wanted = p->min_heap_size;
            } else {
                wanted = next_heap_size(p, wanted, 0);
            }
            if (wanted < HEAP_SIZE(p)) {
                erts_shrink_new_heap(p, wanted, objv, nobj);
            }
            ASSERT(HEAP_SIZE(p) == p->min_heap_size
		   || HEAP_SIZE(p) == next_heap_size(p, HEAP_SIZE(p), 0));
            goto done;
        }

        /*
         * The heap size turned out to be just right. We are done.
         */

        if (HEAP_SIZE(p) >= need_after) {
            ASSERT(HEAP_SIZE(p) == next_heap_size(p, HEAP_SIZE(p), 0));
            goto done;
        }
#ifdef GC_HEAP_TRACE
        fprintf(stderr, "Did a gen_gc, still not enough room\n");
#endif

#if defined(NOMOVE) && defined(SHARED_HEAP)
        fprintf(stderr, "ggc.c: ERROR! Not enough space in young generation after GC!\n");
#endif

#if !(defined(NOMOVE) && defined(SHARED_HEAP))
    }

    /*
     * The previous generational GC did not leave enough free heap space.
     * We must do a fullsweep GC. First figure out the size of the heap
     * to receive all live data.
     */

    sz = HEAP_SIZE(p) + MBUF_SIZE(p) + (OLD_HTOP(p) - OLD_HEAP(p));
#ifndef SHARED_HEAP
    sz += p->hend - p->stop;
#endif
    sz = next_heap_size(p, sz, 0);

    /*
     * Should we grow although we don't actually need to?
     */

    if (sz == HEAP_SIZE(p) && FLAGS(p) & F_HEAP_GROW) {
        sz = next_heap_size(p, HEAP_SIZE(p), 1);
    }
    FLAGS(p) &= ~F_HEAP_GROW;


    fullsweep_heap(p, sz, objv, nobj);

    CHECK(p);
    adjust_after_fullsweep(p, size_before, need, objv, nobj);

#ifdef __BENCHMARK__
    was_this_major = 1;
#endif

#endif /* !(NOMOVE && SHARED_HEAP) */

 done:

    CHECK(p);
    OverRunCheck();
#ifdef SHARED_HEAP
    p->active = 1;
#endif
    p->status = saved_status;

    if (IS_TRACED_FL(p, F_TRACE_GC)) {
        trace_gc(p, am_gc_end);
    }
    BM_STOP_TIMER(gc);

    if (erts_system_monitor_long_gc != 0) {
	Uint ms2, s2, us2;
	Sint t;
	get_now(&ms2, &s2, &us2);
	t = ms2 - ms1;
	t = t*1000000 + s2 - s1;
	t = t*1000 + (us2 - us1)/1000;
	if (t > 0 && (Uint)t >= erts_system_monitor_long_gc) {
	    monitor_long_gc(p, t);
	}
    }
    if (erts_system_monitor_large_heap != 0
	&& HEAP_SIZE(p) >= erts_system_monitor_large_heap) {
	monitor_large_heap(p);
    }


#ifdef __BENCHMARK__
#if !(defined(NOMOVE) && defined(SHARED_HEAP))
#ifdef HYBRID
    if (!(internal_gc_flags & GC_GLOBAL))
#endif
    {
      if (was_this_major == 1)
      {
#ifdef SHARED_HEAP
          BM_COUNT(major_global_garbage_cols);
#else
          BM_COUNT(major_garbage_cols);
#endif
#ifdef BM_TIMERS
#ifdef SHARED_HEAP
          major_global_gc_time += gc_time;
          if (gc_time > max_global_major_time)
              max_global_major_time = gc_time;
#else
          major_gc_time += gc_time;
          if (gc_time > max_major_time)
              max_major_time = gc_time;
#endif
#endif
    }
    else
#endif /* !(NOMOVE && SHARED_HEAP) */
    {
#ifdef SHARED_HEAP
          BM_COUNT(minor_global_garbage_cols);
#else
          BM_COUNT(minor_garbage_cols);
#endif
#ifdef BM_TIMERS
#ifdef SHARED_HEAP
          minor_global_gc_time += gc_time;
          if (gc_time > max_global_minor_time)
              max_global_minor_time = gc_time;
#else
          minor_gc_time += gc_time;
          if (gc_time > max_minor_time)
              max_minor_time = gc_time;
#endif
#endif
      }
    }
#endif /* __BENCHMARK__ */

#ifdef BM_HEAP_SIZES
    {
        double total_used_heap = 0;
        double total_allocated_heap = 0;
#ifdef SHARED_HEAP
        total_used_heap = (HEAP_TOP(p) - HEAP_START(p)) +
                          (OLD_HTOP(p) - OLD_HEAP(p));
        total_allocated_heap = HEAP_SIZE(p) + (OLD_HEND(p) - OLD_HEAP(p));
        if (total_used_heap > max_used_global_heap)
            max_used_global_heap = total_used_heap;
        if (total_allocated_heap > max_allocated_global_heap)
            max_allocated_global_heap = total_allocated_heap;
#else
        int i;
        for (i = 0; i < erts_max_processes; i++)
        {
            Process *cp = process_tab[i];
            if (cp == NULL) continue;

            total_used_heap += (cp->htop - cp->heap) +
                               (cp->old_htop - cp->old_heap);
            total_allocated_heap += cp->heap_sz +
                                   (cp->old_hend - cp->old_heap);
        }
        if (total_used_heap > max_used_heap)
            max_used_heap = total_used_heap;
        if (total_allocated_heap > max_allocated_heap)
            max_allocated_heap = total_allocated_heap;
#endif
    }
#endif /* BM_HEAP_SIZES */

#if defined(BM_TIMERS) && defined(HYBRID)
    if (!(internal_gc_flags & GC_GLOBAL))
        BM_START_TIMER(system);
#else
    BM_START_TIMER(system);
#endif

    ARITH_LOWEST_HTOP(p) = (Eterm *) 0;
    ARITH_AVAIL(p) = 0;
    ARITH_HEAP(p) = NULL;
#ifdef DEBUG
    ARITH_CHECK_ME(p) = NULL;
#endif
    return ((int) (HEAP_TOP(p) - HEAP_START(p)) / 10);
#undef OverRunCheck
}

static Eterm*
gen_cheney(Process *p, Eterm* low, Eterm* high, Eterm* n_hp, Eterm* n_htop)
{
    Eterm* ptr;
    Eterm val;
    Eterm gval;

    while (n_hp != n_htop) {
        gval = *n_hp;

        switch (primary_tag(gval)) {
	case TAG_PRIMARY_BOXED: {
            ptr = boxed_val(gval);
            val = *ptr;
	    if (ptr_within(ptr, low, high)) {
		if (MY_IS_MOVED(val)) {
		    ASSERT(is_boxed(val));
		    *n_hp++ = val;
		} else {
		    MOVE_BOXED(ptr,val,n_htop,n_hp++);
		}
            } else {
                ASSERT(within(ptr, p));
                n_hp++;
            }
            continue;
	}
	case TAG_PRIMARY_LIST: {
            ptr = list_val(gval);
            val = *ptr;
	    if (ptr_within(ptr, low, high)) {
		if (is_non_value(val)) {
		    *n_hp++ = ptr[1];
		} else {
		    MOVE_CONS(ptr,val,n_htop,n_hp++);
		}
            } else {
                ASSERT(within(ptr, p));
                n_hp++;
            }
            continue;
          }
          case TAG_PRIMARY_HEADER: {
              if (header_is_thing(gval))
                  n_hp += (thing_arityval(gval)+1);
              else
                  n_hp++;
              continue;
          }
          default: {
            n_hp++;
            continue;
          }
        }
    }

    return n_htop;
}

/*
 * This function sweeps both the remainder of the new heap
 * as well as the remainder of the old heap after the first pass
 * of the generational collector gen_gc().
 */
static Eterm*
gen_cheney_old(Process *p, Eterm* from, Eterm** to,
           Eterm* low_water, Eterm* high_water, Eterm* old_htop,
           Eterm* objv, int nobj)
{
    Eterm* n_hp;
    Eterm* n_htop;
    Eterm* oh_start;
    Eterm* oh_end;

    Eterm* ptr;
    Eterm val;
    Eterm gval;
#if HIPE
    char *const_start = (char*)hipe_constants_start;
    unsigned long const_size = (char*)hipe_constants_next - const_start;
#endif
#ifdef HYBRID
    Eterm *g_heap  = global_heap;
    Eterm *g_htop  = global_htop;
    Eterm *go_heap = OLD_M_DATA_START;
    Eterm *go_htop = OLD_M_DATA_END;
#endif

    n_hp = from;
    n_htop = *to;
#if defined(NOMOVE) && defined(SHARED_HEAP)
    oh_start = nm_heap;
    oh_end = nm_hend;
#else
    oh_start = OLD_HEAP(p);
    oh_end = OLD_HEND(p);
#endif

    while (n_hp != n_htop) {
        gval = *n_hp;

        switch (primary_tag(gval)) {
          case TAG_PRIMARY_BOXED: {
            ptr = boxed_val(gval);
            val = *ptr;
            if (ptr_within(ptr, oh_start, oh_end))
            {
                n_hp++;
#if defined(NOMOVE) && defined(SHARED_HEAP)
                NM_STORE(build,ptr,BOXED_NEED(ptr,val));
#endif
            }
#ifdef HYBRID
            else if( ptr_within(ptr, g_heap, g_htop) )
            {
                ++n_hp;
            }
            else if( ptr_within(ptr, go_heap, go_htop) )
            {
                ++n_hp;
            }
#endif
#if HIPE
            else if( in_area(ptr, const_start, const_size) )
            {
                ++n_hp;
            }
#endif
            else if (MY_IS_MOVED(val)) {
		ASSERT(is_boxed(val));
                *n_hp++ = val;
            } else if (ptr_within(ptr, low_water, high_water)) {
                /* Make object old */
#if defined(NOMOVE) && defined(SHARED_HEAP)
                old_htop = erts_nm_alloc(p,BOXED_NEED(ptr,val),objv,nobj);
                NM_STORE(build,old_htop,BOXED_NEED(ptr,val));
#endif
                MOVE_BOXED(ptr,val,old_htop,n_hp++);
            } else {
                ASSERT(within(ptr, p));
                MOVE_BOXED(ptr,val,n_htop,n_hp++);
            }
            continue;
          }
          case TAG_PRIMARY_LIST: {
            ptr = list_val(gval);
            val = *ptr;
            if (ptr_within(ptr, oh_start, oh_end))
            {
                n_hp++;
#if defined(NOMOVE) && defined(SHARED_HEAP)
                NM_STORE(build,ptr,2);
#endif
            }
#ifdef HYBRID
            else if( ptr_within(ptr, g_heap, g_htop) )
            {
                ++n_hp;
            }
            else if( ptr_within(ptr, go_heap, go_htop) )
            {
                ++n_hp;
            }
#endif
#if HIPE
            else if( in_area(ptr, const_start, const_size) )
            {
                ++n_hp;
            }
#endif
            else if (is_non_value(val))
                *n_hp++ = ptr[1];
            else if (ptr_within(ptr, low_water, high_water)) {
                /* Make object old */
#if defined(NOMOVE) && defined(SHARED_HEAP)
                old_htop = erts_nm_alloc(p,2,objv,nobj);
                NM_STORE(build,old_htop,2);
#endif
                MOVE_CONS(ptr,val,old_htop,n_hp++);
            } else {
                ASSERT(within(ptr, p));
                MOVE_CONS(ptr,val,n_htop,n_hp++);
            }
            continue;
          }
          case TAG_PRIMARY_HEADER: {
              if (header_is_thing(gval))
                  n_hp += (thing_arityval(gval)+1);
              else
                  n_hp++;
              continue;
          }
          default: {
            n_hp++;
            continue;
          }
        }
    }

    /* Now set the parameter pointers for the caller */
    *to = n_htop;
    return old_htop;
}


/*
 * Garbage collect the heap. However, all objects pointing
 * to the old generation heap may be left as is.
 * Every other turn, we remember the position on the heap
 * that turned out to be the heap top, every other second turn
 * we tenure all live objects that reside below that water mark.
 * This means that we only tenure objects that have survived at
 * least two collections.
*/
static void
gen_gc(Process *p, int new_sz, Eterm* objv, int nobj)
{
#ifdef SHARED_HEAP
    Rootset *rootset = erts_alloc(ERTS_ALC_T_ROOTSET,
				  sizeof(Rootset)*erts_max_processes);
#else
    Rootset rootset;            /* Rootset for GC (stack, dictionary, etc). */
#endif
    Eterm* n_hstart;
    Eterm* n_htop;
    int n;
    Eterm* ptr;
    Eterm val;
    Eterm gval;
#if defined(NOMOVE) && defined(SHARED_HEAP)
    Eterm* oh_start = nm_heap;
    Eterm* oh_end = nm_hend;
    Eterm* old_htop = NULL;
#else
    Eterm* oh_start = OLD_HEAP(p);
    Eterm* oh_end = OLD_HEND(p);
    Eterm* old_htop = OLD_HTOP(p);
#endif
    Eterm* low_water = HEAP_START(p);
    Eterm* high_water = HIGH_WATER(p);
    Eterm* tmp;
#if HIPE
    char *const_start = (char*)hipe_constants_start;
    unsigned long const_size = (char*)hipe_constants_next - const_start;
#endif
#ifdef HYBRID
    Eterm *g_heap  = global_heap;
    Eterm *g_htop  = global_htop;
    Eterm *go_heap = OLD_M_DATA_START;
    Eterm *go_htop = OLD_M_DATA_END;
#endif

#ifdef GC_HEAP_TRACE
    fprintf(stderr, "Generational GC %s ", print_pid(p));
#endif
    /* If flip is true, we need to tenure all (live) objects */
    /* within the watermarks, if flip is 0, we need to alloc a */
    /* new new_heap and copy all live objects to the new new_heap */
    /* that is to not tenure any objects at all */

#if defined(NOMOVE) && defined(SHARED_HEAP)
    {
        struct nm_page *this = nm_used_mem;
        while (this)
        {
            memset(blackmap + ((void*)this - (void*)nm_heap) / sizeof(void*),
                   0,NM_FULLPAGE);
            this = this->next;
        }
    }
#endif

    n_hstart = (Eterm*) ERTS_HEAP_ALLOC(ERTS_ALC_T_HEAP, sizeof(Eterm)*new_sz);
    n_htop = n_hstart;
#ifdef SHARED_HEAP
    n = collect_roots(p, objv, nobj, rootset);

    while (n--) {
      GENSWEEP_NSTACK(rootset[n].p, old_htop, n_htop, objv, nobj);
      while (rootset[n].n--) {
        Eterm* g_ptr = rootset[n].v[rootset[n].n];
        Uint g_sz = rootset[n].sz[rootset[n].n];
#else
    n = setup_rootset(p, objv, nobj, &rootset);

    GENSWEEP_NSTACK(p, old_htop, n_htop, objv, nobj);
    while (n--) {
        Eterm* g_ptr = rootset.v[n];
        Uint g_sz = rootset.sz[n];
#endif

        while (g_sz--) {
            gval = *g_ptr;

            switch (primary_tag(gval)) {

              case TAG_PRIMARY_BOXED: {
                ptr = boxed_val(gval);
                val = *ptr;
                if (ptr_within(ptr, oh_start, oh_end))
                {
#if defined(NOMOVE) && defined(SHARED_HEAP)
                    NM_STORE(build,ptr,BOXED_NEED(ptr,val));
#endif
                    g_ptr++;
                }
#ifdef HYBRID
                else if( ptr_within(ptr, g_heap, g_htop) )
                {
                    ++g_ptr;
                }
                else if( ptr_within(ptr, go_heap, go_htop) )
                {
                    ++g_ptr;
                }
#endif
#if HIPE
                else if( in_area(ptr, const_start, const_size) )
                {
                    ++g_ptr;
                }
#endif
                else if (MY_IS_MOVED(val)) {
		    ASSERT(is_boxed(val));
                    *g_ptr++ = val;
                } else if (ptr_within(ptr, low_water, high_water)) {
#if defined(NOMOVE) && defined(SHARED_HEAP)
                    old_htop = erts_nm_alloc(p,BOXED_NEED(ptr,val),objv,nobj);
                    NM_STORE(build,old_htop,BOXED_NEED(ptr,val));
#endif
                    MOVE_BOXED(ptr,val,old_htop,g_ptr++);
                } else {
                    ASSERT(within(ptr, p));
                    MOVE_BOXED(ptr,val,n_htop,g_ptr++);
                }
                continue;
              }

              case TAG_PRIMARY_LIST: {
                ptr = list_val(gval);
                val = *ptr;
                if (ptr_within(ptr, oh_start, oh_end))
                {
#if defined(NOMOVE) && defined(SHARED_HEAP)
                    NM_STORE(build,ptr,2);
#endif
                    g_ptr++;
                }
#ifdef HYBRID
                else if( ptr_within(ptr, g_heap, g_htop) )
                {
                    ++g_ptr;
                }
                else if( ptr_within(ptr, go_heap, go_htop) )
                {
                    ++g_ptr;
                }
#endif
#if HIPE
                else if( in_area(ptr, const_start, const_size) )
                {
                    ++g_ptr;
                }
#endif
                else if (is_non_value(val)) {
                    *g_ptr++ = ptr[1];
                }
                else if (ptr_within(ptr, low_water, high_water)) {
#if defined(NOMOVE) && defined(SHARED_HEAP)
                    old_htop = erts_nm_alloc(p,2,objv,nobj);
                    NM_STORE(build,old_htop,2);
#endif
                    MOVE_CONS(ptr,val,old_htop,g_ptr++);
                } else {
                    ASSERT(within(ptr, p));
                    MOVE_CONS(ptr,val,n_htop,g_ptr++);
                }
                continue;
              }

              default: {
                g_ptr++;
                continue;
              }
            }
        }
    }
#ifdef SHARED_HEAP
    }
#endif

    /*
     * Now all references in the rootset point to the new heap. However,
     * most references on the new heap point to the old heap so the next stage
     * is to scan through the new heap evacuating data from the old heap
     * until all is changed.
     */

    tmp = n_htop;
#if defined(NOMOVE) && defined(SHARED_HEAP)
    (void)gen_cheney_old(p, n_hstart, &tmp, low_water, high_water, NULL, objv, nobj);
#else
    old_htop = gen_cheney_old(p, n_hstart, &tmp, low_water, high_water, old_htop, objv, nobj);
#endif
    n_htop = tmp;

    /*
     * And also if we have been tenuring, references on the second generation
     * may point to the old (soon to be deleted) heap.
     */

#if defined(NOMOVE) && defined(SHARED_HEAP)
    erts_nm_copymark(p,objv,nobj);
#else
    if (OLD_HTOP(p) < old_htop) {
      if (MBUF(p) == NULL) {
          old_htop = gen_cheney(p, HEAP_START(p), HEAP_END(p), OLD_HTOP(p), old_htop);
      } else {
          tmp = old_htop;
          (void) gen_cheney_old(p, OLD_HTOP(p), &tmp, OLD_HEAP(p),
                                OLD_HEAP(p), NULL, NULL, 0);
          old_htop = tmp;
      }
    }
    OLD_HTOP(p) = old_htop;
#endif

#ifdef SHARED_HEAP
    restore_rootset(p, rootset);
    erts_free(ERTS_ALC_T_ROOTSET, (void *) rootset);
#else
    restore_rootset(p, &rootset);
#endif

#if defined(NOMOVE) && defined(SHARED_HEAP)
    NM_STORAGE_SWAP(build,root);
#endif

    if (MSO(p).mso) {
        sweep_proc_bins(p, 0);
    }
#ifndef SHARED_HEAP
#ifndef HYBRID /* FIND ME! */
    if (MSO(p).funs) {
        sweep_proc_funs(p, 0);
    }
#endif
#endif
    if (MSO(p).externals) {
	sweep_proc_externals(p, 0);
    }

    remove_message_buffers(p);

    HIGH_WATER(p) = (HEAP_START(p) != HIGH_WATER(p)) ? n_hstart : n_htop;

#ifndef SHARED_HEAP
    /*
     * Now we got to move the stack to the top of the new heap...
     */
    n = HEAP_END(p) - p->stop;
    sys_memcpy(n_hstart + new_sz - n, p->stop, n * sizeof(Eterm));
#endif

#ifdef DEBUG
    sys_memset(HEAP_START(p), 0xff, HEAP_SIZE(p) * sizeof(Eterm));
#endif

    ERTS_HEAP_FREE(ERTS_ALC_T_HEAP,
		   (void *) HEAP_START(p),
		   HEAP_SIZE(p)*sizeof(Eterm));

    HEAP_START(p) = n_hstart;
    HEAP_TOP(p) = n_htop;
    HEAP_SIZE(p) = new_sz;
    HEAP_END(p) = n_hstart + new_sz;
#ifndef SHARED_HEAP
    p->stop = HEAP_END(p) - n;
#endif
}

static void
sweep_proc_externals(Process *p, int fullsweep)
{
    ExternalThing** prev;
    ExternalThing* ptr;
    Eterm* bot;
    Eterm* top;

#ifdef HYBRID
    if (internal_gc_flags && GC_GLOBAL)
    {
        bot = OLD_M_DATA_START;
        top = OLD_M_DATA_END;
        prev = &erts_global_offheap.externals;
        ptr = erts_global_offheap.externals;
    }
    else
#endif
    {
#if defined(NOMOVE) && defined(SHARED_HEAP)
        bot = nm_heap;
        top = nm_hend;
#else
        bot = OLD_HEAP(p);
        top = OLD_HEND(p);
#endif
        prev = &MSO(p).externals;
        ptr = MSO(p).externals;
    }

    while (ptr) {
        Eterm* ppt = (Eterm *) ptr;

        if (MY_IS_MOVED(*ppt)) {        /* Object is alive */
            ExternalThing* ro = external_thing_ptr(*ppt);

            *prev = ro;         /* Patch to moved pos */
            prev = &ro->next;
            ptr = ro->next;
        } else if (fullsweep == 0 && ptr_within(ppt, bot, top)) {

            /*
             * Object resides on old heap, and we just did a
             * generational collection - keep object in list.
             */
            prev = &ptr->next;
            ptr = ptr->next;
        } else {                /* Object has not been moved - deref it */
	    DEREF_ERL_NODE(ptr->node);
            *prev = ptr = ptr->next;
        }
    }
    ASSERT(*prev == NULL);
}

#ifndef SHARED_HEAP
#ifndef HYBRID /* FIND ME! */
static void
sweep_proc_funs(Process *p, int fullsweep)
{
    ErlFunThing** prev;
    ErlFunThing* ptr;
    Eterm* bot;
    Eterm* top;

#ifdef HYBRID
    if (internal_gc_flags && GC_GLOBAL)
    {
        bot = OLD_M_DATA_START;
        top = OLD_M_DATA_END;
        prev = &erts_global_offheap.funs;
        ptr = erts_global_offheap.funs;
    }
    else
#endif
    {
        bot = OLD_HEAP(p);
        top = OLD_HEND(p);
        prev = &MSO(p).funs;
        ptr = MSO(p).funs;
    }

    while (ptr) {
        Eterm* ppt = (Eterm *) ptr;

#ifdef HYBRID
        if (internal_gc_flags && GC_GLOBAL)
#endif

        if (MY_IS_MOVED(*ppt)) {        /* Object is alive */
            ErlFunThing* ro = (ErlFunThing *) fun_val(*ppt);

            *prev = ro;         /* Patch to moved pos */
            prev = &ro->next;
            ptr = ro->next;
        } else if (fullsweep == 0 && ptr_within(ppt, bot, top)) {

            /*
             * Object resides on old heap, and we just did a
             * generational collection - keep object in list.
             */
            prev = &ptr->next;
            ptr = ptr->next;
        } else {                /* Object has not been moved - deref it */
            ErlFunEntry* fe = ptr->fe;

            *prev = ptr = ptr->next;
            if (--(fe->refc) == 0) {
                erts_erase_fun_entry(fe);
            }
        }
    }
    ASSERT(*prev == NULL);
}
#endif
#endif

static void
sweep_proc_bins(Process *p, int fullsweep)
{
    ProcBin** prev;
    ProcBin* ptr;
    Binary* bptr;
    Eterm* bot;
    Eterm* top;

#ifdef HYBRID
    if (internal_gc_flags && GC_GLOBAL)
    {
        bot = OLD_M_DATA_START;
        top = OLD_M_DATA_END;
        prev = &erts_global_offheap.mso;
        ptr = erts_global_offheap.mso;
    }
    else
#endif
    {
#if defined(NOMOVE) && defined(SHARED_HEAP)
        bot = nm_heap;
        top = nm_hend;
#else
        bot = OLD_HEAP(p);
        top = OLD_HEND(p);
#endif
        prev = &MSO(p).mso;
        ptr = MSO(p).mso;
    }

    /*
     * Note: In R7 we no longer force a fullsweep when we find binaries
     * on the old heap. The reason is that with the introduction of the
     * bit syntax we can expect binaries to be used a lot more. Note that
     * in earlier releases a brand new binary (or any other term) could
     * be put on the old heap during a gen-gc fullsweep, but this is
     * no longer the case in R7.
     */
    while (ptr) {
        Eterm* ppt = (Eterm *) ptr;

        if (MY_IS_MOVED(*ppt)) {        /* Object is alive */
            ProcBin* ro = (ProcBin*) binary_val(*ppt);

            *prev = ro;         /* Patch to moved pos */
            prev = &ro->next;
            ptr = ro->next;
        } else if (fullsweep == 0 && ptr_within(ppt, bot, top)) {

            /*
             * Object resides on old heap, and we just did a
             * generational collection - keep object in list.
             */
            prev = &ptr->next;
            ptr = ptr->next;
        } else {                /* Object has not been moved - deref it */
            *prev = ptr->next;
            bptr = ptr->val;
            bptr->refc--;
            if (bptr->refc == 0) {
                if (bptr->flags & BIN_FLAG_MATCH_PROG) {
                    erts_match_set_free(bptr);
                } else {
		    erts_bin_free(bptr);
                }
            }
            ptr = *prev;
        }
    }
    ASSERT(*prev == NULL);
}


/************************************************************** */
/*  DEBUG routines                                              */
/****************************************************************/


#ifdef HARDDEBUG

static int
within(Eterm *ptr, Process *p)
{
    ErlHeapFragment* bp = MBUF(p);

#if !(defined(NOMOVE) && defined(SHARED_HEAP))
    if (OLD_HEAP(p) && (OLD_HEAP(p) <= ptr && ptr < OLD_HEND(p))) {
        return 1;
    }
#endif

    if (HEAP_START(p) <= ptr && ptr < HEAP_TOP(p)) {
        return 1;
    }

#ifdef HYBRID
    if (global_heap <= ptr && ptr < global_hend) {
        return 1;
    }
#ifndef NOMOVE
    if (global_old_heap <= ptr && ptr < global_old_hend) {
        return 1;
    }
#endif
#endif

#ifdef NOMOVE
    if (nm_heap <= ptr && ptr < nm_hend) {
        return 1;
    }
#endif
    while (bp != NULL) {
        if (bp->mem <= ptr && ptr < bp->mem + bp->size) {
            return 1;
        }
        bp = bp->next;
    }
    return 0;
}

static void
check_pointer(Process* p, Eterm obj, int back_pointers_allowed, char *msg)
{
    if (back_pointers_allowed && !within(ptr_val(obj), p)) {
        erl_exit(1, "%s: %s, line %d: %s: bad address %x\n",
                 print_pid(p), __FILE__, __LINE__, msg, obj);
    } else if (!back_pointers_allowed &&
#if defined(NOMOVE) && defined(SHARED_HEAP)
               !ptr_within(ptr_val(obj), nm_heap, nm_hend)
#else
               !ptr_within(ptr_val(obj), OLD_HEAP(p), OLD_HEND(p))
#endif
               ) {
        if (within(ptr_val(obj), p)) {
            erl_exit(1, "%s: %s, line %d: %s: back pointer %x\n",
                     print_pid(p), __FILE__, __LINE__, msg, obj);
        } else {
            erl_exit(1, "%s: %s, line %d: %s: bad address %x\n",
                     print_pid(p), __FILE__, __LINE__, msg, obj);
        }
    }
}


static void
check_binary(Process *p, Eterm obj, char* msg)
{
    Eterm* ptr = binary_val(obj);
    int btype = thing_subtag(*ptr);

    ASSERT(is_thing(*ptr));

    if (*ptr == 0xffffffff) {
        erl_exit(1, "%s: Clobbered binary left\n", print_pid(p));
    }
    check_pointer(p, obj, 1, msg);

    switch (btype) {
    case REFC_BINARY_SUBTAG:
        {
            ProcBin *bp = (ProcBin*)binary_val(obj);

            if (bp->size > 0xffffffff) {
                erl_exit(1, "%s: check_binary: %s: LARGE binary found %x\n",
                         print_pid(p), msg, obj);
            }
            if (bp->bytes == 0) {
                erl_exit(1, "%s: check_binary: %s: NULL binary found %x\n",
                         print_pid(p), msg, obj);
            }
            if (bp->val->refc <= 0) {
                erl_exit(1, "Bad refc in binary\n");
            }
        }
        break;
    case HEAP_BINARY_SUBTAG:
	/* TODO */
	break;
    case SUB_BINARY_SUBTAG:
	/* TODO */
	break;
    default:
        erl_exit(1, "Unknown subtag in thing binary \n");
    }
}

static void
check_stack(Process *p, char *msg)
{
    Eterm* sp;
#ifdef SHARED_HEAP
    if (p->stop < p->send) {
        erl_exit(1, "%s: Overrun stack\n", print_pid(p));
    }

    for (sp = p->stop; sp < p->stack; sp++) {
        if (is_not_catch(*sp)) {
            ; //stack_element_check(p, msg, *sp);
        }
    }
#else
    Eterm* stack_end;
    int stack_size;
    int sz;

    sp = p->stop;
    stack_end = p->hend;
    sz = p->hend - p->htop;

    stack_size = stack_end - sp;
    if (sp < p->htop) {
        erl_exit(1, "%s: Overrun stack and heap\n", print_pid(p));
    }

    for (sp = p->stop; sp < p->hend; sp++) {
        if (is_not_catch(*sp)) {
            ; //stack_element_check(p, msg, *sp);
        }
    }
#endif
}

int
chk_sys(void)
{
    Process *p = *process_tab;
    int i, res = 0;
    for (i = 0; i < erts_max_processes; i++) {
        if ((p = process_tab[i]) != NULL) {
            res++;
            check_stack(p, "chk");
        }
    }
    return res;
}

void
check_bins(Process *p)
{
    ProcBin *pb = MSO(p).mso;
    while (pb) {
        check_binary(p, make_binary((Eterm*) pb), "CHECK");
        pb = pb->next;
    }
    return;
}

#endif  /* HARDDEBUG  */

static char*
print_pid(Process *p)
{
    char static buf[64];

    Eterm obj = p->id;
    sprintf(buf,
	    "<%lu.%lu.%lu>",
	    internal_pid_channel_no(obj),
	    internal_pid_number(obj),
	    internal_pid_serial(obj));
    return buf;
}

#endif /* HEAP_FRAG_ELIM_TEST */


#ifdef HYBRID
/***************************************************************************
 *                                                                         *
 *                Garbage collection of the Message Area                   *
 *                                                                         *
 ***************************************************************************/

static void ma_gen_gc(Process*, int, Eterm*, int);

static void
ma_offset_rootset(Process *p, int offs, 
	       Eterm* low, Eterm* high, 
	       Eterm* objv, int nobj)
{
    Uint i;
    Uint n = erts_num_active_procs;
    Process *current = p;

    for (i = 0; i < n; i++) {
        p = erts_active_procs[i];
        if (p->dictionary) {
            offset_heap(p->dictionary->data, 
                        p->dictionary->used, 
                        offs, low, high);
	}
        if (p->debug_dictionary) {
            offset_heap(p->debug_dictionary->data, 
                        p->debug_dictionary->used, 
                        offs, low, high);
	}
	offset_heap(&p->fvalue, 1, offs, low, high);
	offset_heap(&p->group_leader, 1, offs, low, high);
        offset_heap(&p->seq_trace_token, 1, offs, low, high);
        offset_mqueue(p, offs, low, high);
        offset_heap_ptr(p->stop, (STACK_START(p) - p->stop), offs, low, high);
        offset_heap(p->heap, p->htop - p->heap, offs, low, high);
        offset_nstack(p, offs, low, high);
	offset_off_heap(p, low, high, offs);
	if (p->old_heap) printf("Someone has an old heap!!\n\r");
	if (p != current ) {
	    offset_heap(p->arg_reg, p->arity, offs, low, high);
	} else if (nobj > 0) {
	    offset_heap(objv, nobj, offs, low, high);
	}
    }
}

#if defined(HIPE)
#if defined(__i386__)

#ifndef NOMOVE
static Eterm *ma_fullsweep_nstack(Process *p, Eterm *n_htop)
{
    Eterm *nsp;
    Eterm *nsp_end;
    const struct sdesc *sdesc;
    struct sdesc sdesc0;
    unsigned int sdesc_size;
    unsigned long ra;
    unsigned int i;
    unsigned int mask;
    unsigned int nstkarity;
    Eterm *gheap = global_heap;
    Eterm *ghtop = global_htop;
    Eterm *goheap = global_old_heap;
    Eterm *gohtop = global_old_htop;

    nsp = p->hipe.nsp;
    nsp_end = p->hipe.nstend;

    nstkarity = p->hipe.narity - X86_NR_ARG_REGS;
    if( (int)nstkarity < 0 )
	nstkarity = 0;
    sdesc0.summary = (0 << 9) | (0 << 8) | nstkarity;
    sdesc0.livebits[0] = ~1; /* all but RA, which is first since no locals */
    sdesc = &sdesc0;

    for(;;) {
	if( nsp >= nsp_end ) {
	    if( nsp == nsp_end ) {
		return n_htop;
	    }
	    fprintf(stderr, "%s: passed end of stack\r\n", __FUNCTION__);
	    break;
	}
	sdesc_size = sdesc_fsize(sdesc) + 1 + sdesc_arity(sdesc);
	i = 0;
	mask = sdesc->livebits[0];
	for(;;) {
	    if( mask & 1 ) {
		Eterm gval = nsp[i];
		if( is_boxed(gval) ) {
		    Eterm *ptr = boxed_val(gval);
		    Eterm val = *ptr;
		    if( MY_IS_MOVED(val) ) {
			nsp[i] = val;
		    } else if (ptr_within(ptr, gheap, ghtop)) {
			MOVE_BOXED(ptr,val,n_htop,&nsp[i]);
		    } else if (ptr_within(ptr, goheap, gohtop)) {
			MOVE_BOXED(ptr,val,n_htop,&nsp[i]);
		    }
		} else if( is_list(gval) ) {
		    Eterm *ptr = list_val(gval);
		    Eterm val = *ptr;
		    if( is_non_value(val) ) {
			nsp[i] = ptr[1];
		    } else if (ptr_within(ptr, gheap, ghtop)) {
			MOVE_CONS(ptr,val,n_htop,&nsp[i]);
		    } else if (ptr_within(ptr, gheap, ghtop)) {
			MOVE_CONS(ptr,val,n_htop,&nsp[i]);
		    }
		}
	    }
	    if( ++i >= sdesc_size )
		break;
	    if( i & 31 )
		mask >>= 1;
	    else
		mask = sdesc->livebits[i >> 5];
	}

	ra = nsp[sdesc_fsize(sdesc)];
	if( ra == (unsigned long)nbif_stack_trap_ra )
	    sdesc = hipe_find_sdesc((unsigned long)p->hipe.ngra);
	else
	    sdesc = hipe_find_sdesc(ra);
	nsp += sdesc_size;
    }
    abort();
}
#endif /* !NOMOVE */

static void ma_gensweep_nstack(Process *p, Eterm **ptr_old_htop, Eterm **ptr_n_htop, Eterm *objv, int nobj)
{
    Eterm *nsp;
    Eterm *nsp_end;
    const struct sdesc *sdesc;
    struct sdesc sdesc0;
    unsigned int sdesc_size;
    unsigned long ra;
    unsigned int i;
    unsigned int mask;
    Eterm *low_water, *high_water, *surface;
    Eterm *n_htop;
#ifndef NOMOVE
    Eterm *old_htop;
#endif
    unsigned int nstkarity;

    nsp = p->hipe.nsp;
    nsp_end = p->hipe.nstend;

    low_water = global_heap;
    high_water = global_high_water;
    surface = global_htop;

#ifndef NOMOVE
    old_htop = *ptr_old_htop;
#endif
    n_htop = *ptr_n_htop;

    nstkarity = p->hipe.narity - X86_NR_ARG_REGS;
    if( (int)nstkarity < 0 )
	nstkarity = 0;
    sdesc0.summary = (0 << 9) | (0 << 8) | nstkarity;
    sdesc0.livebits[0] = ~1; /* all but RA, which is first since no locals */
    sdesc = &sdesc0;

    for(;;) {
	if( nsp >= nsp_end ) {
	    if( nsp == nsp_end ) {
#ifndef NOMOVE
		*ptr_old_htop = old_htop;
#endif
		*ptr_n_htop = n_htop;
		return;
	    }
	    fprintf(stderr, "%s: passed end of stack\r\n", __FUNCTION__);
	    break;
	}
	sdesc_size = sdesc_fsize(sdesc) + 1 + sdesc_arity(sdesc);
	i = 0;
	mask = sdesc->livebits[0];
	for(;;) {
	    if( mask & 1 ) {
		Eterm gval = nsp[i];
		if( is_boxed(gval) ) {
		    Eterm *ptr = boxed_val(gval);
		    Eterm val = *ptr;
		    if( MY_IS_MOVED(val) ) {
			nsp[i] = val;
		    } else if (ptr_within(ptr,low_water,high_water)) {
#ifdef NOMOVE
                        Eterm *old_htop = erts_nm_alloc(p,BOXED_NEED(ptr,val),objv,nobj);
                        NM_STORE(build,old_htop,BOXED_NEED(ptr,val));
#endif
			MOVE_BOXED(ptr,val,old_htop,&nsp[i]);
		    } else if (ptr_within(ptr,high_water,surface)) {
                        NM_STORE(build,ptr,BOXED_NEED(ptr,val));
			MOVE_BOXED(ptr,val,n_htop,&nsp[i]);
		    }
		} else if( is_list(gval) ) {
		    Eterm *ptr = list_val(gval);
		    Eterm val = *ptr;
		    if( is_non_value(val) ) {
			nsp[i] = ptr[1];
		    } else if (ptr_within(ptr,low_water,high_water)) {
#ifdef NOMOVE
                        Eterm *old_htop = erts_nm_alloc(p,2,objv,nobj);
                        NM_STORE(build,old_htop,2);
#endif
			MOVE_CONS(ptr,val,old_htop,&nsp[i]);
		    } else if (ptr_within(ptr,high_water,surface)) {
                        NM_STORE(build,ptr,2);
			MOVE_CONS(ptr,val,n_htop,&nsp[i]);
		    }
		}
	    }
	    if( ++i >= sdesc_size )
		break;
	    if( i & 31 )
		mask >>= 1;
	    else
		mask = sdesc->livebits[i >> 5];
	}
	ra = nsp[sdesc_fsize(sdesc)];
	if( ra == (unsigned long)nbif_stack_trap_ra )
	    sdesc = hipe_find_sdesc((unsigned long)p->hipe.ngra);
	else
	    sdesc = hipe_find_sdesc(ra);
	nsp += sdesc_size;
    }
    abort();
}

#endif	/* __i386__ */

#if defined(__sparc__)

#ifndef NOMOVE
static Eterm *ma_fullsweep_nstack(Process *p, Eterm *n_htop)
{
    Eterm *nsp;
    Eterm *nsp_end;
    const struct sdesc *sdesc0;
    const struct sdesc *sdesc;
    unsigned int sdesc_size;
    unsigned long ra;
    unsigned int i;
    unsigned int mask;
    char *const_start;
    unsigned long const_size;

    ra = (unsigned long)p->hipe.nra;
    if (!ra) return n_htop;

    nsp = p->hipe.nsp-1;
    nsp_end =  p->hipe.nstgraylim;

    if( nsp_end ) {
      /* Check if the trap frame is on the stack or in ra */
      if( ra == (unsigned long)nbif_stack_trap_ra ) {
	p->hipe.nra = p->hipe.ngra;
	ra = (unsigned long)p->hipe.nra;
      } else { /* The trap is on the stack, remove it. */
	/* remove gray/white boundary trap */
	for(;;) {
	  ++nsp_end;
	  /* Sanity check */
	  if(nsp_end > nsp) abort();
	  if( nsp_end[0] == (unsigned long)nbif_stack_trap_ra ) {
	    nsp_end[0] = (unsigned long)p->hipe.ngra;
	    break;
	  }
	}
      }
    }
    nsp_end = p->hipe.nstack;

    sdesc = hipe_find_sdesc(ra);
    sdesc0 = sdesc;

    const_start = (char*)hipe_constants_start;
    const_size = (char*)hipe_constants_next - const_start;

    for(;;) {
	if( nsp <= nsp_end ) {
	    if( nsp == nsp_end ) {
		if( nsp ) {
		    /* see the HIGH_WATER update in fullsweep_heap() */
		    p->hipe.nstblacklim = p->hipe.nstack;
		    hipe_update_stack_trap(p, sdesc0);
		}
		return n_htop;
	    }
	    fprintf(stderr, "%s: passed end of stack\r\n", __FUNCTION__);
	    break;
	}
	sdesc_size = sdesc_fsize(sdesc) + sdesc_arity(sdesc);
	i = 0;
	mask = sdesc->livebits[0];
	for(;;) {
	    if( mask & 1 ) {
		Eterm gval = nsp[-i];
		if( is_boxed(gval) ) {
		    Eterm *ptr = boxed_val(gval);
		    Eterm val = *ptr;
		    if( MY_IS_MOVED(val) ) {
			nsp[-i] = val;
		    } else if( in_area(ptr, const_start, const_size) ) {
			;
                    } else if( ptr_within(ptr, global_heap, global_htop) ) {
                    } else if( ptr_within(ptr, OLD_M_DATA_START, OLD_M_DATA_END) ) {
		    } else {
			ASSERT(within(ptr, p));
			MOVE_BOXED(ptr,val,n_htop,&nsp[-i]);
		    }
		} else if( is_list(gval) ) {
		    Eterm *ptr = list_val(gval);
		    Eterm val = *ptr;
		    if( is_non_value(val) ) {
			nsp[-i] = ptr[1];
		    } else if( in_area(ptr, const_start, const_size) ) {
			;
                    } else if( ptr_within(ptr, global_heap, global_htop) ) {
                    } else if( ptr_within(ptr, OLD_M_DATA_START, OLD_M_DATA_END) ) {
		    } else {
			ASSERT(within(ptr, p));
			MOVE_CONS(ptr,val,n_htop,&nsp[-i]);
		    }
		}
	    }
	    if( ++i >= sdesc_size )
		break;
	    if( i & 31 )
		mask >>= 1;
	    else
		mask = sdesc->livebits[i >> 5];
	}
	ra = nsp[1-sdesc_fsize(sdesc)];
	sdesc = hipe_find_sdesc(ra);
	nsp -= sdesc_size;
    }
    abort();
}
#endif /* !NOMOVE */

static void ma_gensweep_nstack(Process *p, Eterm **ptr_old_htop, Eterm **ptr_n_htop, Eterm *objv, int nobj)
{
    Eterm *nsp;
    Eterm *nsp_end;
    const struct sdesc *sdesc0;
    const struct sdesc *sdesc;
    unsigned int sdesc_size;
    unsigned long ra;
    unsigned int i;
    unsigned int mask;
    Eterm *oh_start, *oh_end;
    Eterm *low_water, *high_water;
    Eterm *n_htop, *old_htop;
    char *const_start;
    unsigned long const_size;

    nsp = p->hipe.nsp-1;
    nsp_end = p->hipe.nstgraylim;
    ra = (unsigned long)p->hipe.nra;
    if (!ra) return;


    if( nsp_end ) {
      /* if gray limit passed black limit, reset black limit */
      if( nsp_end < p->hipe.nstblacklim )
	p->hipe.nstblacklim = nsp_end;

      /* Check if the trap frame is on the stack or in ra */
      if( ra == (unsigned long)nbif_stack_trap_ra ) {
	p->hipe.nra = p->hipe.ngra;
	ra = (unsigned long)p->hipe.nra;
      } else {
	/* remove gray/white boundary trap */
	for(;;) {
	  ++nsp_end;
	  /* Sanity check */
	  if(nsp_end > nsp) abort();
	  if( nsp_end[0] == (unsigned long)nbif_stack_trap_ra ) {
	    nsp_end[0] = (unsigned long)p->hipe.ngra;
	    break;
	  }
	}
      }
      nsp_end = p->hipe.nstblacklim;
    } else
      nsp_end = p->hipe.nstack;

    sdesc = hipe_find_sdesc(ra);
    sdesc0 = sdesc;

    const_start = (char*)hipe_constants_start;
    const_size = (char*)hipe_constants_next - const_start;
    oh_start = OLD_HEAP(p);
    oh_end = OLD_HEND(p);
    low_water = HEAP_START(p);
    high_water = HIGH_WATER(p);
    old_htop = *ptr_old_htop;
    n_htop = *ptr_n_htop;

    for(;;) {
	if( nsp <= nsp_end ) {
	    if( nsp == nsp_end ) {
		*ptr_old_htop = old_htop;
		*ptr_n_htop = n_htop;
		if( nsp ) {
		    /* see the HIGH_WATER update in gen_gc() */
		    if( HEAP_START(p) != HIGH_WATER(p) ) {
			p->hipe.nstblacklim =
			    p->hipe.nstgraylim
			    ? p->hipe.nstgraylim
			    : p->hipe.nstack;
		    } else {
			/* blacklim = graylim ? blacklim : end */
			if( !p->hipe.nstgraylim )
			    p->hipe.nstblacklim = p->hipe.nstack;
		    }
		    hipe_update_stack_trap(p, sdesc0);
		}
		return;
	    }
	    fprintf(stderr, "%s: passed end of stack\r\n", __FUNCTION__);

	    break;
	}
	sdesc_size = sdesc_fsize(sdesc) + sdesc_arity(sdesc);
	i = 0;
	mask = sdesc->livebits[0];
	for(;;) {
	    if( mask & 1 ) {
		Eterm gval = nsp[-i];
		if( is_boxed(gval) ) {
		    Eterm *ptr = boxed_val(gval);
		    Eterm val = *ptr;
		    if( ptr_within(ptr, oh_start, oh_end) ) {
		    } else if( in_area(ptr, const_start, const_size) ) {
                    } else if( ptr_within(ptr, global_heap, global_htop) ) {
                    } else if( ptr_within(ptr, OLD_M_DATA_START, OLD_M_DATA_END) ) {
		    } else if( MY_IS_MOVED(val) ) {
			nsp[-i] = val;
		    } else if( ptr_within(ptr, low_water, high_water) ) {
			MOVE_BOXED(ptr,val,old_htop,&nsp[-i]);
		    } else {
			ASSERT(within(ptr, p));
			MOVE_BOXED(ptr,val,n_htop,&nsp[-i]);
		    }
		} else if( is_list(gval) ) {
		    Eterm *ptr = list_val(gval);
		    Eterm val = *ptr;
		    if( ptr_within(ptr, oh_start, oh_end) ) {
		    } else if( in_area(ptr, const_start, const_size) ) {
                    } else if( ptr_within(ptr, global_heap, global_htop) ) {
                    } else if( ptr_within(ptr, OLD_M_DATA_START, OLD_M_DATA_END) ) {
		    } else if( is_non_value(val) ) {
			nsp[-i] = ptr[1];
		    } else if( ptr_within(ptr, low_water, high_water) ) {
			MOVE_CONS(ptr,val,old_htop,&nsp[-i]);
		    } else {
			ASSERT(within(ptr, p));
			MOVE_CONS(ptr,val,n_htop,&nsp[-i]);
		    }
		}
	    }
	    if( ++i >= sdesc_size )
		break;
	    if( i & 31 )
		mask >>= 1;
	    else
		mask = sdesc->livebits[i >> 5];
	}
	nsp -= sdesc_size;
	ra = nsp[1+sdesc_arity(sdesc)];
	sdesc = hipe_find_sdesc(ra);
    }
    abort();
}

#endif	/* __sparc__ */

#define MA_GENSWEEP_NSTACK(p,old_htop,n_htop,objv,nobj)	                \
	do {								\
		Eterm *tmp_old_htop = old_htop;				\
		Eterm *tmp_n_htop = n_htop;				\
		ma_gensweep_nstack((p),&tmp_old_htop,&tmp_n_htop,objv,nobj); \
		old_htop = tmp_old_htop;				\
		n_htop = tmp_n_htop;					\
	} while(0)

/*
 * offset_nstack() can ignore the descriptor-based traversal the other
 * nstack procedures use and simply call offset_heap_ptr() instead.
 * This relies on two facts:
 * 1. The only live non-Erlang terms on an nstack are return addresses,
 *    and they will be skipped thanks to the low/high range check.
 * 2. Dead values, even if mistaken for pointers into the low/high area,
 *    can be offset safely since they won't be dereferenced.
 *
 * XXX: WARNING: If HiPE starts storing other non-Erlang values on the
 * nstack, such as floats, then this will have to be changed.
 */

#else /* !HIPE */

#ifndef NOMOVE
#define ma_fullsweep_nstack(p,n_htop)		(n_htop)
#endif
#define MA_GENSWEEP_NSTACK(p,old_htop,n_htop,objv,nobj)   do{}while(0)
#define ma_offset_nstack(p,offs,low,high)                 do{}while(0)

#endif /* HIPE */


/* A general Cheney-style sweep of given memory area. Copies objects
 * between low and high.
 */

static Eterm *cheneyScan(Eterm *from, Eterm *to, Eterm *low, Eterm *high)
{
    Eterm *n_hp;
    Eterm *n_htop;
    Eterm *ptr;
    Eterm val;
    Eterm gval;

    n_hp     = from;
    n_htop   = to;

    while (n_hp != n_htop) {
        gval = *n_hp;

        switch (primary_tag(gval)) {
          case TAG_PRIMARY_BOXED: {
            ptr = boxed_val(gval);
            val = *ptr;
            if (MY_IS_MOVED(val))
                *n_hp++ = val;
            else if (ptr_within(ptr,low,high))
                MOVE_BOXED(ptr,val,n_htop,n_hp++);
            else
                ++n_hp;
            continue;
          }
          case TAG_PRIMARY_LIST: {
            ptr = list_val(gval);
            val = *ptr;
            if (is_non_value(val))
                *n_hp++ = ptr[1];
            else if (ptr_within(ptr,low,high))
                MOVE_CONS(ptr,val,n_htop,n_hp++);
            else
                ++n_hp;
            continue;
          }
          case TAG_PRIMARY_HEADER: {
              if (header_is_thing(gval))
                  n_hp += (thing_arityval(gval)+1);
              else
                  n_hp++;
              continue;
          }
          default: {
            n_hp++;
            continue;
          }
        }
    }
    return n_htop;
}

#ifndef NOMOVE
/* 
** This function is used when using fullsweep gc.
** This replaces gen_gc for processes that use the
** fullsweep algorithm and don't have an old_heap.
** Parameters:
** p: The process who is being garbage collected.
** new_sz: The wanted size of the heap after collecting.
** objv: Vector of "extra" objects to be "saved".
** nobj: Number of objects in objv.
*/

static void ma_fullsweep_heap(Process *p, int new_sz, Eterm* objv, int nobj)
{
    Rootset rootset[erts_max_processes];
    Eterm *n_htop;		/* Top of new heap */
    Eterm *n_heap;		/* The new heap */
    Eterm *n_hp;
    Eterm *g_heap = global_heap;
    Eterm *g_hend = global_htop;
    Eterm *oh_start = global_old_heap;
    Eterm *oh_end   = global_old_htop;
    int n;

    /* Create new, empty heap */
    n_heap = (Eterm *) ERTS_HEAP_ALLOC(ERTS_ALC_T_HEAP, sizeof(Eterm)*new_sz);
    n_htop = n_heap;
    global_gc_flags &= ~F_NEED_FULLSWEEP;
#ifdef GC_HEAP_TRACE
    fprintf(stderr, "Global Fullsweep GC caused by: %d\n", print_pid(p));
#endif

    internal_gc_flags |= GC_INCLUDE_ALL;
    n = collect_roots(p, objv, nobj, rootset);
    while (n--)
    {
      n_htop = ma_fullsweep_nstack(rootset[n].p, n_htop);
      while (rootset[n].n--) {
        Eterm* g_ptr = rootset[n].v[rootset[n].n];
        Uint  g_sz = rootset[n].sz[rootset[n].n];
	while(g_sz--) {
	    Eterm gval = *g_ptr;

	    switch (primary_tag(gval)) {

	      case TAG_PRIMARY_BOXED: {
		Eterm *ptr = boxed_val(gval);
		Eterm val = *ptr;
		if (MY_IS_MOVED(val))
		{
		    ASSERT(is_boxed(val));
		    *g_ptr++ = val;
		}
                else if( ptr_within(ptr, g_heap, g_hend) )
                {
		    MOVE_BOXED(ptr,val,n_htop,g_ptr++);
                }
                else if( ptr_within(ptr, oh_start, oh_end) )
                {
		    MOVE_BOXED(ptr,val,n_htop,g_ptr++);
                }
		else {
		    g_ptr++;
		}
		continue;
	      }

	      case TAG_PRIMARY_LIST: {
		Eterm *ptr = list_val(gval);
		Eterm val = *ptr;
		if (is_non_value(val))
		    *g_ptr++ = ptr[1];
                else if( ptr_within(ptr, g_heap, g_hend) )
                {
		    MOVE_CONS(ptr,val,n_htop,g_ptr++);
                }
                else if( ptr_within(ptr, oh_start, oh_end) )
                {
		    MOVE_CONS(ptr,val,n_htop,g_ptr++);
                }
		else {
                    ++g_ptr;
		}
		continue;
	      }

	      default: {
		g_ptr++;
		continue;
	      }
	    }
	}
      }
    }

    /* Scan through all process heaps to find pointers to global heap */

    {
      Uint i;
      Uint n = erts_num_active_procs;

      for (i = 0; i < n; i++) {
        p = erts_active_procs[i];
        n_hp = p->heap;
        ASSERT(p->old_htop == p->old_heap);
        while (n_hp != p->htop) {
          Eterm gval = *n_hp;
          switch (primary_tag(gval)) {

	  case TAG_PRIMARY_BOXED: {
	    Eterm *ptr = boxed_val(gval);
	    Eterm val = *ptr;
	    if (MY_IS_MOVED(val))
	    {
		ASSERT(is_boxed(val));
 		*n_hp++ = val;
	    }
            else if( ptr_within(ptr, g_heap, g_hend) )
              {
		MOVE_BOXED(ptr,val,n_htop,n_hp++);
              }
            else if( ptr_within(ptr, oh_start, oh_end) )
              {
		MOVE_BOXED(ptr,val,n_htop,n_hp++);
              }
	    else {
              ++n_hp;
	    }
	    continue;
	  }

	  case TAG_PRIMARY_LIST: {
	    Eterm *ptr = list_val(gval);
	    Eterm val = *ptr;
	    if (is_non_value(val))
              *n_hp++ = ptr[1];
            else if( ptr_within(ptr, g_heap, g_hend) )
              {
		MOVE_CONS(ptr,val,n_htop,n_hp++);
              }
            else if( ptr_within(ptr, oh_start, oh_end) )
              {
		MOVE_CONS(ptr,val,n_htop,n_hp++);
              }
	    else {
              ++n_hp;
	    }
	    continue;
	  }
            
	  case TAG_PRIMARY_HEADER: {
            if (header_is_thing(gval))
              n_hp += (thing_arityval(gval)+1);
            else
              n_hp++;
            continue;
	  }
            
	  default: {
	    n_hp++;
	    continue;
	  }
          }
        }
      }
    }

    /*
     * Now all references on the stack point to the new heap. However,
     * most references on the new heap point to the old heap so the next stage
     * is to scan through the new heap evacuating data from the old heap
     * until all is copied.
     */

    n_hp = n_heap;
    while (n_hp != n_htop) {
	Eterm gval = *n_hp;

	switch (primary_tag(gval)) {

	  case TAG_PRIMARY_BOXED: {
	    Eterm *ptr = boxed_val(gval);
	    Eterm val = *ptr;
	    if (MY_IS_MOVED(val))
	    {
		ASSERT(is_boxed(val));
		*n_hp++ = val;
	    }
            else if( ptr_within(ptr, g_heap, g_hend) )
            {
		MOVE_BOXED(ptr,val,n_htop,n_hp++);
            }
            else if( ptr_within(ptr, oh_start, oh_end) )
            {
		MOVE_BOXED(ptr,val,n_htop,n_hp++);
            }
	    else {
                ++n_hp;
	    }
	    continue;
	  }

	  case TAG_PRIMARY_LIST: {
	    Eterm *ptr = list_val(gval);
	    Eterm val = *ptr;
	    if (is_non_value(val))
		*n_hp++ = ptr[1];
            else if( ptr_within(ptr, g_heap, g_hend) )
            {
		MOVE_CONS(ptr,val,n_htop,n_hp++);
            }
            else if( ptr_within(ptr, oh_start, oh_end) )
            {
		MOVE_CONS(ptr,val,n_htop,n_hp++);
            }
	    else {
                ++n_hp;
	    }
	    continue;
	  }

	  case TAG_PRIMARY_HEADER: {
            if (header_is_thing(gval)) {
		  n_hp += (thing_arityval(gval)+1);
            }
            else {
		  n_hp++;
            }
	      continue;
	  }

	  default: {
	    n_hp++;
	    continue;
	  }
	}
    }

    global_high_water = n_htop;

    /* Finally we have to rescue data from the copy_dst_stack. This
     * has to be done last since we want all this to be ABOVE the high
     * water mark on the new heap. (It has not been created yet, so
     * why should it have survived a GC?)
     */

    if (copy_dst_top > 1)
    {
        Uint top = copy_dst_top;
        Eterm *g_ptr = copy_dst_stack + 1;
        Eterm *start = n_htop;
        while (--top)
        {
            Eterm gval = *g_ptr;
            switch (primary_tag(gval))
            {
            case TAG_PRIMARY_LIST:
              {
	        Eterm *ptr = list_val(gval);
                Eterm val = *ptr;
                if (is_non_value(val))
                    *g_ptr++ = ptr[1];
                else
                    MOVE_CONS(ptr,val,n_htop,g_ptr++);
                break;
              }

            case TAG_PRIMARY_BOXED:
              {
	        Eterm *ptr = boxed_val(gval);
                Eterm val = *ptr;
                if (MY_IS_MOVED(val))
                    *g_ptr++ = val;
                else
                    MOVE_BOXED(ptr,val,n_htop,g_ptr++);
                break;
              }
            }
        }
        n_htop = cheneyScan(start,n_htop,g_heap,g_hend);
        *(Eterm*)copy_dst_stack[0] = copy_dst_stack[1];
    }

    restore_rootset(NULL,rootset);

    if (erts_global_offheap.mso) {
	sweep_proc_bins(NULL,1);
    }

#ifndef HYBRID /* FIND ME! */
    if (erts_global_offheap.funs) {
	sweep_proc_funs(NULL,1);
    }
#endif

    if (erts_global_offheap.externals) {
        sweep_proc_externals(NULL,1);
    }

    if (global_old_heap != NULL) {
#ifdef DEBUG
        sys_memset(global_old_heap, 0xff,
                   (global_old_hend - global_old_heap)*sizeof(Eterm));
#endif
        ERTS_HEAP_FREE(ERTS_ALC_T_HEAP,
                       (void*)global_old_heap,
                       (global_old_hend - global_old_heap) * sizeof(Eterm));
        global_old_heap = global_old_htop = global_old_hend = NULL;
    }

#ifdef DEBUG
    sys_memset(global_heap, 0xff, (global_hend - global_heap)*sizeof(Eterm));
#endif    
    global_hend = n_heap + new_sz;
    ERTS_HEAP_FREE(ERTS_ALC_T_HEAP,
                   (void*)global_heap,global_heap_sz * sizeof(Eterm));

    global_heap = n_heap;
    global_htop = n_htop;
    global_heap_sz = new_sz;
    global_gen_gcs = 0;
}
#endif /* NOMOVE */

/*
 * Grow the new heap size to 'new_sz'.
 */
static void
ma_grow_new_heap(Process *p, Uint new_sz, Eterm* objv, int nobj)
{
    Eterm* new_heap;
    int heap_size = global_htop - global_heap;
    Sint offs;

    ASSERT(global_heap_sz < new_sz);
    new_heap = (Eterm *) ERTS_HEAP_REALLOC(ERTS_ALC_T_HEAP,
                                           (void*)global_heap,
                                           sizeof(Eterm)*global_heap_sz,
                                           sizeof(Eterm)*new_sz);
#ifdef GC_HEAP_TRACE
    fprintf(stderr, "ma_grow_new_heap: GREW (%d) FROM %d UPTO %d (used %d)\n",
            pid_number(p->id), global_heap_sz, new_sz, heap_size);
#endif

    if ((offs = new_heap - global_heap) == 0) { /* No move. */
        global_hend = new_heap + new_sz;
    } else {
        offset_heap(new_heap, heap_size, offs, global_heap, global_htop);
        global_high_water = new_heap + (global_high_water - global_heap);
        global_hend = new_heap + new_sz;
        ma_offset_rootset(p, offs, global_heap, global_htop, objv, nobj);
        if (copy_dst_top > 1)
	{
            offset_heap_ptr(copy_dst_stack + 1,copy_dst_top - 1,offs,
                            global_heap,global_htop);
	    *(Eterm*)copy_dst_stack[0] = copy_dst_stack[1];
	}
        global_htop = new_heap + heap_size;
        global_heap = new_heap;
    }
    global_heap_sz = new_sz;
}

static void
ma_shrink_new_heap(Process *p, Uint new_sz, Eterm *objv, int nobj)
{
    Eterm* new_heap;
    int heap_size = global_htop - global_heap;
    Sint offs;

    ASSERT(new_sz < global_heap_sz);
    new_heap = (Eterm *) ERTS_HEAP_REALLOC(ERTS_ALC_T_HEAP,
                                           (void*)global_heap,
                                           sizeof(Eterm) * global_heap_sz,
                                           sizeof(Eterm) * new_sz);
    global_hend = new_heap + new_sz;

#ifdef GC_HEAP_TRACE
    fprintf(stderr, "ma_shrink_new_heap: SHRINKED (%d) FROM %d DOWNTO %d (used %d)\n",
            pid_number(p->id), global_heap_sz, new_sz, heap_size);
#endif

    if ((offs = new_heap - global_heap) != 0) {

        /*
         * Normally, we don't expect a shrunk heap to move, but you never
         * knows on some strange embedded systems...  Or when using purify.
         */
        offset_heap(new_heap, heap_size, offs, global_heap, global_htop);
        global_high_water = new_heap + (global_high_water - global_heap);
        ma_offset_rootset(p, offs, global_heap, global_htop, objv, nobj);
        if (copy_dst_top > 1)
	{
            offset_heap_ptr(copy_dst_stack + 1,copy_dst_top - 1,offs,
                            global_heap,global_htop);
	    *(Eterm*)copy_dst_stack[0] = copy_dst_stack[1];
	}

        global_htop = new_heap + heap_size;
        global_heap = new_heap;
    }
    global_heap_sz = new_sz;
}

#ifndef NOMOVE
static void
ma_adjust_after_fullsweep(Process *p, int size_before, int need, Eterm *objv, int nobj)
{
    int wanted, sz, size_after, need_after;

    size_after = (global_htop - global_heap);
    reclaimed += (size_before - size_after);

     /*
      * Resize the heap if needed.
      */

    need_after = size_after + need;
    if (global_heap_sz < need_after) {
        /* Too small - grow to match requested need */
        sz = next_heap_size(p,need_after, 0);
        ma_grow_new_heap(p, sz, objv, nobj);
    } else if (3 * global_heap_sz < 4 * need_after){
        /* Need more than 75% of current, postpone to next GC.*/
        global_gc_flags |= F_HEAP_GROW;
    } else if (2 * need_after < global_heap_sz &&
               global_heap_sz > SH_DEFAULT_SIZE) {
        /* We need less than 25% of the current heap, shrink.*/
        /* XXX - This is how it was done in the old GC:
           wanted = 4 * need_after;
           I think this is better as fullsweep is used mainly on
           small memory systems, but I could be wrong... */
        wanted = 2 * need_after;
        if (wanted < SH_DEFAULT_SIZE) {
            sz = SH_DEFAULT_SIZE;
        } else {
            sz = next_heap_size(p,wanted, 0);
        }
        if (sz < global_heap_sz) {
            ma_shrink_new_heap(p, sz, objv, nobj);
        }
    }
}
#endif /* NOMOVE */

/*
** Garbage collect the message area.
** Parameters:
** p: Pointer to the calling process structure.
** need: Number of (erlang) words needed on the heap.
** objv: Array of terms to add to rootset, that is to preserve.
** nobj: Number of objects in objv.
*/
int
erts_global_garbage_collect(Process* p, int need, Eterm* objv, int nobj)
{
    int size_before;
    int size_after;
    int need_after;
    Uint saved_status;
    int wanted;
#ifndef NOMOVE
    int sz;
#ifdef __BENCHMARK__
    uint was_this_major = 0;
#endif
#endif
    //BM_NEW_TIMER(gc);

    BM_STOP_TIMER(system);

#ifdef BM_HEAP_SIZES
    {
        unsigned long total_used_heap = (global_htop - global_heap) +
          (global_old_htop - global_old_heap);
        unsigned long total_allocated_heap = global_heap_sz +
          (global_old_hend - global_old_heap);

        if (total_used_heap > max_used_global_heap)
            max_used_global_heap = total_used_heap;
        if (total_allocated_heap > max_allocated_global_heap)
            max_allocated_global_heap = total_allocated_heap;
    }
#endif /* BM_HEAP_SIZES */

#ifdef BM_TIMERS
    gc_time = 0;
#endif
    BM_START_TIMER(gc);

#define OverRunCheck() \
    if (global_hend < global_htop) { \
        erl_exit(1, "%s: Overrun global heap at line %d\n", print_pid(p),__LINE__); \
    }

    if (IS_TRACED_FL(p, F_TRACE_GC)) {
        trace_gc(p, am_gc_start);
    }
    saved_status = p->status;
    p->status = P_GARBING;
    CHECK(p);
    OverRunCheck();

#ifndef NOMOVE
    if (global_gen_gcs >= global_max_gen_gcs) {
        global_gc_flags |= F_NEED_FULLSWEEP;
    }
#endif

    garbage_cols++;

    /* Size of heap before first GC */
    size_before = global_htop - global_heap;

#ifndef NOMOVE
    /*
     * Generational GC from here on. We need an old heap.
     */

    if (global_old_heap == NULL && global_high_water != global_heap &&
        (global_gc_flags & F_NEED_FULLSWEEP) == 0) {
        Eterm* n_old;
        /* Note: We choose a larger heap size than strictly needed,
         * which seems to reduce the number of fullsweeps.
         * This improved Estone by more than 1200 estones on my computer
         * (Ultra Sparc 10).
         */
        size_t new_sz = next_heap_size(p,global_high_water -
                                            global_heap, 1);

        /* Create new, empty old_heap */
        n_old = (Eterm *) ERTS_HEAP_ALLOC(ERTS_ALC_T_HEAP,
                                          sizeof(Eterm)*new_sz);

        global_old_hend = n_old + new_sz;
        global_old_heap = global_old_htop = n_old;
#ifdef OLD_HEAP_CREATION_TRACE
        fprintf(stderr, "Created a global old_heap because of %s.\r\n",
            print_pid(p));
#endif
    }
#endif /* !NOMOVE */

    /* Before we can do a global colection, the local heaps of all
     * processes that have been active since the last major collection
     * has to be cleaned up.
     */
    {
        Uint i;
	Uint n = erts_num_active_procs;

        BM_SWAP_TIMER(gc,local_in_global);

	for (i = 0; i < n; i++)
	{
	    Process *cp = erts_active_procs[i];
	    if (!cp->active)
	    {
              printf("Skipped Local GC\n\r");
                ASSERT(cp->high_water == cp->htop);
	        continue;
	    }

	    cp->flags |= F_NEED_FULLSWEEP;
	    if (p == cp)
	        erts_garbage_collect(cp,0,objv,nobj);
	    else
	        erts_garbage_collect(cp,0,cp->arg_reg,cp->arity);
	}

        BM_SWAP_TIMER(local_in_global,gc);
    }

    /*
     * Make sure functions shared between local and global collection
     * knows this is a global collection.
     */
    internal_gc_flags |= GC_GLOBAL;

#ifndef NOMOVE
    /*
     * Try a generational GC if the old heap is big enough.
     */

    if ((global_gc_flags & F_NEED_FULLSWEEP) == 0 &&
        global_high_water - global_heap <= global_old_hend - global_old_htop)
    {
#endif /* !NOMOVE */

        /*
         * There is space enough in old_heap for everything
         * below the high water mark.  Do a generational GC.
         */
#ifdef NOMOVE
        /*
         * There is allways space left in the old generation since it
         * has its own GC!
         */
#endif

        ma_gen_gc(p, next_heap_size(p, global_heap_sz, 0), objv, nobj);

        global_gen_gcs++;
        size_after = global_htop - global_heap;
        need_after = size_after + need;
        reclaimed += (size_before - size_after);

        /*
         * Excessively large heaps should be shrunk, but
         * don't even bother on reasonable small heaps.
         *
         * The reason for this is that after tenuring, we often
         * use a really small portion of new heap, therefore, unless
         * the heap size is substantial, we don't want to shrink.
         */

        if ((global_heap_sz > 300) && (4 * need_after < global_heap_sz) &&
            ((global_heap_sz > 8000)
#ifndef NOMOVE
             || (global_heap_sz > (global_old_hend - global_old_heap))
#endif
             )) {
            wanted = 3 * need_after;
            if (wanted < SH_DEFAULT_SIZE) {
                wanted = SH_DEFAULT_SIZE;
            } else {
                wanted = next_heap_size(p, wanted, 0);
            }
            if (wanted < global_heap_sz) {
	        ma_shrink_new_heap(p, wanted, NULL, 0);
            }
            ASSERT(global_heap_sz == next_heap_size(p,global_heap_sz, 0));
            goto ma_done;
        }

        /*
         * The heap size turned out to be just right. We are done.
         */

        if (global_heap_sz >= need_after) {
            ASSERT(global_heap_sz == next_heap_size(p, global_heap_sz, 0));
            goto ma_done;
        }
#ifdef GC_HEAP_TRACE
        fprintf(stderr, "Did a global gen_gc, still not enough room\n");
#endif

#ifdef NOMOVE
        fprintf(stderr, "ggc.c: ERROR! Not enough space in young generation after GC!\n");
#endif

#ifndef NOMOVE
    }

    /*
     * The previous generational GC did not leave enough free heap space.
     * We must do a fullsweep GC. First figure out the size of the heap
     * to receive all live data.
     */

    sz = global_heap_sz + (global_old_htop - global_old_heap);
    sz = next_heap_size(p, sz, 0);

    /*
     * Should we grow although we don't actually need to?
     */

    if (sz == global_heap_sz && global_gc_flags & F_HEAP_GROW) {
        sz = next_heap_size(p, global_heap_sz, 1);
    }

    global_gc_flags &= ~F_HEAP_GROW;
    ma_fullsweep_heap(p, sz, objv, nobj);
    CHECK(p);
    ma_adjust_after_fullsweep(p, size_before, need, NULL, 0);

#ifdef __BENCHMARK__
    was_this_major = 1;
#endif

#endif /* NOMOVE */

 ma_done:

    CHECK(p);
    OverRunCheck();
    p->active = 1;
    p->status = saved_status;
    if (IS_TRACED_FL(p, F_TRACE_GC)) {
        trace_gc(p, am_gc_end);
    }
    internal_gc_flags &= ~GC_GLOBAL;

    BM_STOP_TIMER(gc);

#ifdef __BENCHMARK__
#ifndef NOMOVE
    if (was_this_major == 1)
    {
        BM_COUNT(major_global_garbage_cols);
#ifdef BM_TIMERS
        major_global_gc_time += gc_time;
        if (gc_time > max_global_major_time)
            max_global_major_time = gc_time;
#endif
    }
    else
#endif /* NOMOVE */
    {
        BM_COUNT(minor_global_garbage_cols);
#ifdef BM_TIMERS
        minor_global_gc_time += gc_time;
        if (gc_time > max_global_minor_time)
            max_global_minor_time = gc_time;
#endif
    }
#endif /* __BENCHMARK__ */

#ifdef BM_HEAP_SIZES
    {
        unsigned long total_used_heap = (global_htop - global_heap) +
          (global_old_htop - global_old_heap);
        unsigned long total_allocated_heap = global_heap_sz +
          (global_old_hend - global_old_heap);

        if (total_used_heap > max_used_global_heap)
            max_used_global_heap = total_used_heap;
        if (total_allocated_heap > max_allocated_global_heap)
            max_allocated_global_heap = total_allocated_heap;
    }
#endif /* BM_HEAP_SIZES */

    BM_START_TIMER(system);
    return ((int) (global_htop - global_heap) / 10);
#undef OverRunCheck
}

/*
 * This function sweeps both the remainder of the new heap
 * as well as the remainder of the old heap after the first pass
 * of the generational collector ma_gen_gc().
 */
static Eterm *ma_gen_cheney(Eterm *from, Eterm **to, Eterm *low_water,
                            Eterm *high_water, Eterm *surface, Eterm *old_htop,
                            Eterm *objv, int nobj)
{
    Eterm *n_hp;
    Eterm *n_htop;
    Eterm *ptr;
    Eterm val;
    Eterm gval;

    n_hp     = from;
    n_htop   = *to;

    while (n_hp != n_htop) {
        gval = *n_hp;

        switch (primary_tag(gval)) {
          case TAG_PRIMARY_BOXED: {
            ptr = boxed_val(gval);
            val = *ptr;
            if (MY_IS_MOVED(val))
                *n_hp++ = val;
            else if (ptr_within(ptr,low_water,high_water)) {
#ifdef NOMOVE
                old_htop = erts_nm_alloc(NULL,BOXED_NEED(ptr,val),objv,nobj);
                //NM_STORE(build,old_htop,BOXED_NEED(ptr,val));
                MARK_FORWARD(ptr);
#endif
                MOVE_BOXED(ptr,val,old_htop,n_hp++);
            }
            else if (ptr_within(ptr,high_water,surface))
                MOVE_BOXED(ptr,val,n_htop,n_hp++);
            else
                ++n_hp;
            continue;
          }
          case TAG_PRIMARY_LIST: {
            ptr = list_val(gval);
            val = *ptr;
            if (is_non_value(val))
                *n_hp++ = ptr[1];
            else if (ptr_within(ptr,low_water,high_water)) {
#ifdef NOMOVE
                old_htop = erts_nm_alloc(NULL,2,objv,nobj);
                //NM_STORE(build,old_htop,2);
#endif
                MOVE_CONS(ptr,val,old_htop,n_hp++);
            }
            else if (ptr_within(ptr,high_water,surface))
                MOVE_CONS(ptr,val,n_htop,n_hp++);
            else
                ++n_hp;
            continue;
          }
          case TAG_PRIMARY_HEADER: {
              if (header_is_thing(gval))
                  n_hp += (thing_arityval(gval)+1);
              else
                  n_hp++;
              continue;
          }
          default: {
            n_hp++;
            continue;
          }
        }
    }

    /* Now set the parameter pointers for the caller */
    *to = n_htop;
    return old_htop;
}

/*
 * Garbage collect the heap. However, all objects pointing
 * to the old generation heap may be left as is.
 * Every other turn, we remember the position on the heap
 * that turned out to be the heap top, every other second turn
 * we tenure all live objects that reside below that water mark.
 * This means that we only tenure objects that have survived at
 * least two collections.
 */
static void
ma_gen_gc(Process *p, int new_sz, Eterm* objv, int nobj)
{
    Rootset rootset[erts_max_processes];
    Eterm *n_hstart;
    Eterm *n_htop;
    int n;
    Eterm *low_water  = global_heap;
    Eterm *high_water = global_high_water;
#ifdef NOMOVE
    Eterm* oh_start   = nm_heap;
    Eterm* oh_end     = nm_hend;
#else
    Eterm *old_htop   = global_old_htop;
#endif

    /* We accually only want this to be up to where the message starts,
     * not all the way up to htop. But there shouldn't be any pointers to
     * the message we are copying reachable from the rootset. As long as
     * the first entry in the dst-stack is a local c variable, this will hold.
     */
    Eterm *surface    = global_htop;
    Eterm *tmp;

#ifdef GC_HEAP_TRACE
    fprintf(stderr, "Global Generational GC (caused by %d) ", (int)pid_number(p->id));
#endif

#ifdef NOMOVE
    {
        struct nm_page *this = nm_used_mem;
        while (this)
        {
            memset(blackmap + ((void*)this - (void*)nm_heap) / sizeof(void*),
                   0,NM_FULLPAGE);
            this = this->next;
        }
    }

    fwdptrs = erts_alloc(ERTS_ALC_T_ROOTSET,global_heap_sz);
    memset(fwdptrs,0,global_heap_sz);
#endif

    n_hstart = (Eterm*) ERTS_HEAP_ALLOC(ERTS_ALC_T_HEAP, sizeof(Eterm)*new_sz);
    n_htop = n_hstart;
    n = collect_roots(p, objv, nobj, rootset);
    while (n--) {
#ifdef NOMOVE
      {
          Eterm *old_htop = NULL;
          MA_GENSWEEP_NSTACK(rootset[n].p, old_htop, n_htop, objv, nobj);
      }
#else
      MA_GENSWEEP_NSTACK(rootset[n].p, old_htop, n_htop, objv, nobj);
#endif
      while (rootset[n].n--) {
        Eterm *g_ptr = rootset[n].v[rootset[n].n];
        Uint g_sz = rootset[n].sz[rootset[n].n];

        while (g_sz--) {
            Eterm gval = *g_ptr;
            switch (primary_tag(gval))
            {
            case TAG_PRIMARY_LIST:
              {
	        Eterm *ptr = list_val(gval);
		if (ptr_within(ptr,low_water,high_water))
                {
                    Eterm val = *ptr;
                    if (is_non_value(val))
                        *g_ptr++ = ptr[1];
                    else {
#ifdef NOMOVE
                        Eterm *old_htop = erts_nm_alloc(p,2,objv,nobj);
                        //NM_STORE(build,old_htop,2);
#endif
                        MOVE_CONS(ptr,val,old_htop,g_ptr++);
                    }
                }
                else if (ptr_within(ptr,high_water,surface))
                {
                    Eterm val = *ptr;
                    if (is_non_value(val))
                        *g_ptr++ = ptr[1];
                    else
                        MOVE_CONS(ptr,val,n_htop,g_ptr++);
                }
#ifdef NOMOVE
                else if (ptr_within(ptr, oh_start, oh_end))
                {
                    g_ptr++;
                    //NM_STORE(build,ptr,2);
                }
#endif
                else
                    ++g_ptr;
                break;
              }

            case TAG_PRIMARY_BOXED:
              {
	        Eterm *ptr = boxed_val(gval);
                if (ptr_within(ptr,low_water,high_water))
                {
                    Eterm val = *ptr;
                    if (MY_IS_MOVED(val))
                        *g_ptr++ = val;
                    else {
#ifdef NOMOVE
                        Eterm *old_htop = erts_nm_alloc(p,BOXED_NEED(ptr,val),
                                                        objv,nobj);
                        //NM_STORE(build,old_htop,BOXED_NEED(ptr,val));
                        MARK_FORWARD(ptr);
#endif
                        MOVE_BOXED(ptr,val,old_htop,g_ptr++);
                    }
                }
                else if (ptr_within(ptr,high_water,surface))
                {
                    Eterm val = *ptr;
                    if (MY_IS_MOVED(val))
                        *g_ptr++ = val;
                    else
                        MOVE_BOXED(ptr,val,n_htop,g_ptr++);
                }
#ifdef NOMOVE
                else if (ptr_within(ptr, oh_start, oh_end))
                {
                    g_ptr++;
                    //NM_STORE(build,ptr,BOXED_NEED(ptr,*ptr));
                }
#endif
                else
                    ++g_ptr;
                break;
              }

            default:
              {
                g_ptr++;
                break;
              }
            }
         }
      }
    }

    /* Scan the heaps of processes that has been active since last major GC */
    {
      Uint i;
      Uint n = erts_num_active_procs;

      for (i = 0; i < n; i++)
      {
        Eterm *g_ptr;
        Eterm *htop;
        Process *cp = erts_active_procs[i];
        if (!(cp->active || global_gc_flags & GC_INCLUDE_ALL))
          continue;

        ASSERT(cp->old_htop == cp->old_heap);
        g_ptr = cp->heap;
        htop = cp->htop;
        while (g_ptr != htop) {
            Eterm gval = *g_ptr;
            switch (primary_tag(gval))
            {
            case TAG_PRIMARY_LIST:
              {
	        Eterm *ptr = list_val(gval);
                if (ptr_within(ptr,low_water,high_water))
                {
                    Eterm val = *ptr;
                    if (is_non_value(val))
                        *g_ptr++ = ptr[1];
                    else {
#ifdef NOMOVE
                        Eterm *old_htop = erts_nm_alloc(p,2,objv,nobj);
                        //NM_STORE(build,old_htop,2);
#endif
                        MOVE_CONS(ptr,val,old_htop,g_ptr++);
                    }
                }
                else if (ptr_within(ptr,high_water,surface))
                {
                    Eterm val = *ptr;
                    if (is_non_value(val))
                        *g_ptr++ = ptr[1];
                    else
                        MOVE_CONS(ptr,val,n_htop,g_ptr++);
                }
#ifdef NOMOVE
                else if (ptr_within(ptr, oh_start, oh_end))
                {
                    g_ptr++;
                    //NM_STORE(build,ptr,2);
                }
#endif
                else
                    ++g_ptr;
                break;
              }

            case TAG_PRIMARY_BOXED:
              {
	        Eterm *ptr = boxed_val(gval);
                if (ptr_within(ptr,low_water,high_water))
                {
                    Eterm val = *ptr;
                    if (MY_IS_MOVED(val))
                        *g_ptr++ = val;
                    else {
#ifdef NOMOVE
                        Eterm *old_htop = erts_nm_alloc(p,BOXED_NEED(ptr,val),
                                                        objv,nobj);
                        //NM_STORE(build,old_htop,BOXED_NEED(ptr,val));
                        MARK_FORWARD(ptr);
#endif
                        MOVE_BOXED(ptr,val,old_htop,g_ptr++);
                    }
                }
                else if (ptr_within(ptr,high_water,surface))
                {
                    Eterm val = *ptr;
                    if (MY_IS_MOVED(val))
                        *g_ptr++ = val;
                    else
                        MOVE_BOXED(ptr,val,n_htop,g_ptr++);
                }
#ifdef NOMOVE
                else if (ptr_within(ptr, oh_start, oh_end))
                {
                    g_ptr++;
                    //NM_STORE(build,ptr,BOXED_NEED(ptr,*ptr));
                }
#endif
                else
                    ++g_ptr;
                break;
              }

            case TAG_PRIMARY_HEADER:
              {
                if (header_is_thing(gval))
                    g_ptr += (thing_arityval(gval) + 1);
                else
                    g_ptr++;
                break;
              }

            default:
              {
                g_ptr++;
                break;
              }
            }
        }
      }
    }

    /*
     * Now all references in the rootset point to the new heap. However,
     * most references on the new heap point to the old heap so the next stage
     * is to scan through the new heap evacuating data from the old heap
     * until all is changed.
     */

    tmp = n_htop;
#ifdef NOMOVE
    (void)ma_gen_cheney(n_hstart, &tmp, low_water, high_water, surface,
                        NULL, objv, nobj);
#else
    old_htop = ma_gen_cheney(n_hstart, &tmp, low_water, high_water, surface,
                             old_htop, objv, nobj);
#endif
    n_htop = tmp;

    /*
     * And also if we have been tenuring, references on the second generation
     * may point to the old (soon to be deleted) heap.
     */

#ifdef NOMOVE
    erts_nm_copymark(p,objv,nobj);
#else
    if (global_old_htop < old_htop) {
        old_htop = cheneyScan(global_old_htop, old_htop, low_water, high_water);
    }
    global_old_htop = old_htop;
#endif

    //global_high_water = (global_heap != global_high_water) ? n_hstart : n_htop;
    global_high_water = n_htop;

    /* Finally we have to rescue data from the copy_dst_stack. This
     * has to be done last since we want all this to be ABOVE the high
     * water mark on the new heap. (It has not been created yet, so
     * why should it have survived a GC?) If we did not have to worry
     * about the destructive updates that copy_struct_lazy does, we
     * could add this stack to the normal root set.. Well, sorry for
     * the code bloat. Ahem, lets call it loop unrolling.. ;-)
     */

    if (copy_dst_top > 1)
    {
        Uint top = copy_dst_top;
        Eterm *g_ptr = copy_dst_stack + 1;
        Eterm *start = n_htop;
        while (--top)
        {
            Eterm gval = *g_ptr;
            switch (primary_tag(gval))
            {
            case TAG_PRIMARY_LIST:
              {
	        Eterm *ptr = list_val(gval);
                Eterm val = *ptr;
                if (is_non_value(val))
                    *g_ptr++ = ptr[1];
                else
                    MOVE_CONS(ptr,val,n_htop,g_ptr++);
                break;
              }

            case TAG_PRIMARY_BOXED:
              {
	        Eterm *ptr = boxed_val(gval);
                Eterm val = *ptr;
                if (MY_IS_MOVED(val))
                    *g_ptr++ = val;
                else
                    MOVE_BOXED(ptr,val,n_htop,g_ptr++);
                break;
              }
            }
        }
        n_htop = cheneyScan(start, n_htop, global_heap, global_htop);
        *(Eterm*)copy_dst_stack[0] = copy_dst_stack[1];
    }

#ifdef NOMOVE
    NM_STORAGE_SWAP(build,root);
    erts_free(ERTS_ALC_T_ROOTSET,fwdptrs);
#endif

    if (erts_global_offheap.mso) {
        sweep_proc_bins(NULL,0);
    }

#ifndef HYBRID /* FIND ME! */
    if (erts_global_offheap.funs) {
        sweep_proc_funs(NULL,0);
    }
#endif

    if (erts_global_offheap.externals) {
        sweep_proc_externals(NULL,0);
    }

#ifdef DEBUG
    sys_memset(global_heap, 0xff, global_heap_sz*sizeof(Eterm));
#endif

    ERTS_HEAP_FREE(ERTS_ALC_T_HEAP,
                   (void*)global_heap,global_heap_sz * sizeof(Eterm));
    restore_rootset(NULL,rootset);
    internal_gc_flags &= ~GC_INCLUDE_ALL;

    global_heap = n_hstart;
    global_hend = n_hstart + new_sz;
    global_htop = n_htop;
    global_heap_sz = new_sz;
}
#endif /* HYBRID */


#ifdef NOMOVE
/***************************************************************************
 *                                                                         *
 *         Non-Moving Garbage Collector for the Old Generation             *
 *                                                                         *
 ***************************************************************************/

void erts_nm_init(void)
{
    int i;
    int size = nm_pages * NM_FULLPAGE;
    nm_heap = (Eterm *)ERTS_HEAP_ALLOC(ERTS_ALC_T_OLD_HEAP,size * sizeof(Eterm));
    nm_hend = nm_heap + size;

    blackmap = (char*)erts_alloc(ERTS_ALC_T_ROOTSET, NM_FULLPAGE * nm_pages);

    for (i = 0; i < nm_pages; i++)
    {
        struct nm_page *this = (struct nm_page*)(nm_heap + i * NM_FULLPAGE);
        this->prev = (struct nm_page*)((Eterm*)this - NM_FULLPAGE);
        this->next = (struct nm_page*)((Eterm*)this + NM_FULLPAGE);
    }
    nm_bibop = (struct nm_page*)nm_heap;
    nm_bibop->prev = NULL;
    ((struct nm_page*)(nm_heap + (nm_pages - 1) * NM_FULLPAGE))->next = NULL;

    //print_page_list(nm_bibop);

    nm_used_mem = nm_bibop;
    nm_bibop = nm_bibop->next;
    if (nm_bibop)
        nm_bibop->prev = NULL;
    nm_used_mem->next = nm_used_mem->prev = NULL;

    nm_free_list = (struct nm_mem_block*)nm_used_mem->start;
    nm_free_list->size = NM_PAGESIZE;
    nm_free_list->prev = NULL;
    nm_free_list->next = NULL;

    NM_STORAGE_INIT(gray);
    NM_STORAGE_INIT(blue);
    NM_STORAGE_INIT(root);
    NM_STORAGE_INIT(build);

    //print_page_list(nm_bibop);
    //print_page_list(nm_used_mem);
    //print_free_list();
}

void erts_nm_cleanup(void)
{
    erts_free(ERTS_ALC_T_OLD_HEAP,(void*)nm_heap);
    erts_free(ERTS_ALC_T_ROOTSET,(void*)blackmap);
}

/* XXX: Fix to not copy things:
 * if (ptr_within(min_nm_address,max_nm_adress))
 *   while (page)
 *     if (in this one)
 */

#ifdef HYBRID
/*
 * Scan a root-area for pointers to the nm_heap. (Stack, PCB etc.)
 */
static void root_area_scan(Eterm *ptr, int size)
{
    while (size--)
    {
        Eterm val = *ptr;
        switch (primary_tag(val))
        {
          case TAG_PRIMARY_LIST:
          {
              if (ptr_within(ptr_val(val),nm_heap,nm_hend))
              {
                  NM_STORE(gray,ptr_val(val),2);
              }
              break;
          }

          case TAG_PRIMARY_BOXED:
          {
              if (ptr_within(ptr_val(val),nm_heap,nm_hend))
              {
                  NM_STORE(gray,ptr_val(val),BOXED_NEED(ptr_val(val),*ptr_val(val)));
              }
              break;
          }
        }
        ptr++;
    }
}

/*
 * Scan a heap area for pointers to the nm_heap. (p->heap, p->old_heap)
 */
static void private_heap_scan(Eterm *ptr, int size)
{
    while (size--)
    {
        Eterm val = *ptr;
        switch (primary_tag(val))
        {
          case TAG_PRIMARY_LIST:
              if (ptr_within(ptr_val(val),nm_heap,nm_hend))
              {
                  NM_STORE(gray,ptr_val(val),2);
              }
              break;

          case TAG_PRIMARY_BOXED:
              if (ptr_within(ptr_val(val),nm_heap,nm_hend))
              {
                  NM_STORE(gray,ptr_val(val),BOXED_NEED(ptr_val(val),*ptr_val(val)));
              }
              break;

          case TAG_PRIMARY_IMMED1:
              break;

          case TAG_PRIMARY_HEADER:
          {
              switch (val & _TAG_HEADER_MASK)
              {
                case ARITYVAL_SUBTAG:
                    break;

                default:
                    ptr += thing_arityval(val);
                    size -= thing_arityval(val);
                    break;
              }
              break;
          }

          default: printf("You see, this can't really happen...\n\r");
        }
        ptr++;
    }
}

/*
 * Scan the young generation of the message area for pointers to the nm_heap.
 */
static void message_area_scan(Eterm *ptr, int size)
{
    Eterm *stop = ptr + size;
    while (ptr < stop)
    {
        Eterm val = *ptr;
        switch (primary_tag(val))
        {
          case TAG_PRIMARY_LIST:
              if (ptr_within(ptr_val(val),nm_heap,nm_hend))
              {
                  NM_STORE(gray,ptr_val(val),2);
              }
              ptr++;
              break;

          case TAG_PRIMARY_BOXED:
              if (ptr_within(ptr_val(val),nm_heap,nm_hend))
              {
                  if (fwdptrs[ptr - global_heap] != 0)
                  {
                      ptr += BOXED_NEED(ptr_val(val),*ptr_val(val));
                  }
                  else
                  {
                      NM_STORE(gray,ptr_val(val),BOXED_NEED(ptr_val(val),*ptr_val(val)));
                      ptr++;
                  }
              }
              else
                  ptr++;
              break;

          case TAG_PRIMARY_IMMED1:
              ptr++;
              break;

          case TAG_PRIMARY_HEADER:
          {
              if (is_non_value(val))
              {
                  ptr += 2;
              }
              else
              {
                  switch (val & _TAG_HEADER_MASK)
                  {
                      case ARITYVAL_SUBTAG:
                          break;

                      default:
                          ptr += thing_arityval(val);
                          break;
                  }
                  ptr++;
              }
              break;
          }

          default: printf("You see, this can't really happen...\n\r");
        }
    }
}
#endif /* HYBRID */

/*
 * Scans an area within the nm_heap for pointers to the nm_heap
 */
static void area_scan(Eterm *area, int size)
{
    while (size--)
    {
        Eterm val = *area;
        switch (primary_tag(val))
        {
          case TAG_PRIMARY_LIST:
              if (ptr_within(ptr_val(val),nm_heap,nm_hend))
              {
                  NM_STORE(gray,ptr_val(val),2);
              }
              break;

          case TAG_PRIMARY_BOXED:
              if (ptr_within(ptr_val(val),nm_heap,nm_hend))
              {
                  NM_STORE(gray,ptr_val(val),BOXED_NEED(ptr_val(val),*ptr_val(val)));
              }
              break;

          case TAG_PRIMARY_IMMED1:
              break;

          case TAG_PRIMARY_HEADER:
          {
              switch (val & _TAG_HEADER_MASK)
              {
                case ARITYVAL_SUBTAG:
                    break;
              /*
                Vectors may cause a problem some day...
                case VECTOR_SUBTAG:
                  DPRINT("*************** vector ***************\n\r");
                    break;
               */
                default:
                    area += thing_arityval(val);
                    size -= thing_arityval(val);
                    /* There is a bug somewhere... */
                    //if (size < 0) size = 0;
                    break;
              }
              break;
          }

          default: printf("You are dreaming!\n\r");
        }
        area++;
    }
}

#ifdef SHARED_HEAP
static struct nm_mem_block *erts_nm_collect(Process *p, int need, Eterm* objv, int nobj)
{
  //Rootset *rootset = erts_safe_sl_alloc(sizeof(Rootset)*erts_max_processes);
    Eterm *free_start = NULL;
    struct nm_mem_block *found = NULL;
    struct nm_page *page = NULL;
    BM_NEW_TIMER(old_gc);

    //BM_SWAP_TIMER(gc,old_gc);
    BM_COUNT(major_global_garbage_cols);

    /*
    roots_saved = 0;
    {
        int n = collect_roots(p, objv, nobj, rootset);
        while (n--)
        {
            while (rootset[n].n--)
            {
                Eterm* ptr = rootset[n].v[rootset[n].n];
                Uint  sz = rootset[n].sz[rootset[n].n];
                root_area_scan(ptr,sz);
            }
        }
    }
    */
    //DPRINT("Mark infants\n\r");
    //root_area_scan(HEAP_START(p),HEAP_TOP(p) - HEAP_START(p));  
    //BM_SWAP_TIMER(old_gc,misc0);
    {
        NM_STORAGE_ITERATOR(root);
        struct nm_object *obj = NM_STORAGE_STEP(root);
        //DPRINT("Mark root list\n\r");
        while (obj)
        {
            if (blackmap[obj->this - nm_heap] == 0)
            {
                blackmap[obj->this - nm_heap] = obj->size;
                //DPRINT("Rescue #%d (0x%08x) size: %d\n\r",(int)(obj->this - nm_heap),(int)obj->this,obj->size);
                area_scan(obj->this,obj->size);
            }
            //else
            //BM_COUNT(messages_ego);
            //DPRINT("Already scanned #%d\n\r",(int)(ptr_val(obj->this) - nm_heap));
            obj = NM_STORAGE_STEP(root);
        }
    }

    //DPRINT("Mark gray list\n\r");
    {
        struct nm_object *obj = NM_STORAGE_GET(gray);
        while (obj)
        {
            if (blackmap[obj->this - nm_heap] == 0)
            {
                blackmap[obj->this - nm_heap] = obj->size;
                //DPRINT("Rescue #%d (0x%08x) size: %d\n\r",(int)(obj->this - nm_heap),(int)obj->this,obj->size);
                //BM_SWAP_TIMER(old_gc,misc2);
                area_scan(obj->this,obj->size);
                //BM_SWAP_TIMER(misc2,old_gc);
            }
            //else
            //BM_COUNT(messages_ego);
            //DPRINT("Already scanned #%d\n\r",(int)(ptr_val(obj->this) - nm_heap));
            obj = NM_STORAGE_GET(gray);
        }
    }

    /* XXX: If we remember old free list, can we gain in any way..? */
    nm_free_list = NULL;
    page = nm_used_mem;

    //DPRINT("Sweep phase\n\r");
    while (page)
    {
        int scavenging = 0;
        Eterm *ptr = page->start;
        Eterm *page_end = ptr + NM_PAGESIZE;
        int n = page->start - nm_heap;
        while (ptr < page_end)
        {
            if (blackmap[n] != 0)
            {
                if (scavenging)
                {
                    scavenging = 0;
                    if ((ptr - free_start) * sizeof(Eterm) >=
                        sizeof(struct nm_mem_block))
                    {
                        struct nm_mem_block *new =
                          (struct nm_mem_block*)(free_start);
                        new->size = ptr - free_start;
                        new->prev = NULL;
                        new->next = nm_free_list;
                        if (nm_free_list)
                            nm_free_list->prev = new;
                        nm_free_list = new;
                        //DPRINT("Created free block (1): 0x%08x of size %d\n\r",(int)new,new->size);
                        if ((new->size >= need) &&
                            (!found || (found->size > new->size)))
                        {
                            found = new;
                            //DPRINT("We've got one!\n\r");
                        }
                    }
                }
                //DPRINT("Found a live one #%d @ 0x%08x!\n\r",n,(int)ptr);
                ptr += blackmap[n];
                n += blackmap[n];
            }
            else if (!scavenging)
            {
              //DPRINT("Free block starts @ 0x%08x\n\r",(int)ptr);
                free_start = ptr;
                scavenging = 1;
                n++; ptr++;
            }
            else
            {
                n++; ptr++;
            }
        }

        if (scavenging)
        {
            if (free_start == page->start)
            {
                struct nm_page *next = page->next;
                //DPRINT("Returning page 0x%08x to BIBOP\n\r",(int)page);
                if (page->next)
                    page->next->prev = page->prev;

                if (page->prev)
                    page->prev->next = page->next;
                else
                    nm_used_mem = page->next;

                page->next = nm_bibop;
                nm_bibop = page;
                page = next;
                //print_page_list(nm_bibop);
                //print_page_list(nm_used_mem);
                continue;
            }
            else if ((ptr - free_start) * sizeof(Eterm) >
                     sizeof(struct nm_mem_block))
            {
                struct nm_mem_block *new = (struct nm_mem_block*)(free_start);
                new->size =  ptr - free_start;
                new->prev = NULL;
                new->next = nm_free_list;
                if (nm_free_list)
                    nm_free_list->prev = new;
                nm_free_list = new;
                //DPRINT("Created free block (2): 0x%08x of size %d\n\r",(int)new,new->size);
                if ((new->size >= need) &&
                    (!found || (found->size > new->size)))
                {
                    found = new;
                    //DPRINT("We've got one!\n\r");
                }
            }
        }
        page = page->next;
    }
    //print_free_list();
    //BM_SWAP_TIMER(misc2,old_gc);

    if (!found && !nm_bibop)
    {
        int i;
        int new_pages = nm_pages * 2;
        int size = sizeof(Eterm) * new_pages * NM_FULLPAGE;
        Eterm *new_heap = ERTS_HEAP_ALLOC(ERTS_ALC_T_OLD_HEAP,size);
        Eterm *new_hend = new_heap + size;
        Eterm *new_htop;
        Eterm *last_page_end;
        struct nm_page *new_used_mem;

        printf("We need to copy things!\n\r");

        /* Create new, bigger bag of pages */
        for (i = 0; i < new_pages; i++)
        {
            struct nm_page *this =
              (struct nm_page*)(new_heap + i * NM_FULLPAGE);
            this->prev = (struct nm_page*)((Eterm*)this - NM_FULLPAGE);
            this->next = (struct nm_page*)((Eterm*)this + NM_FULLPAGE);
        }
        nm_bibop = (struct nm_page*)new_heap;
        nm_bibop->prev = NULL;
        ((struct nm_page*)(new_heap + (new_pages - 1) *
                           NM_FULLPAGE))->next = NULL;

        new_used_mem = nm_bibop;
        nm_bibop = nm_bibop->next;
        nm_bibop->prev = NULL;
        new_used_mem->next = new_used_mem->prev = NULL;

        //print_page_list(nm_bibop);
        //print_page_list(new_used_mem);

        /* Move stuff from old bag to new */
        nm_free_list = NULL;
        new_htop = new_used_mem->start;
        last_page_end = new_htop + NM_PAGESIZE;
        page = nm_used_mem;
        while (page)
        {
            Eterm *ptr = page->start;
            Eterm *page_end = ptr + NM_PAGESIZE;
            int n = offsetof(struct nm_page,start) / sizeof(void*) +
              ((Eterm*)page - nm_heap);
            while (ptr < page_end)
            {
                if (blackmap[n] > 0)
                {
                  //DPRINT("Found a live one #%d @ 0x%08x! Left: %d\n\r",n,(int)ptr,last_page_end-new_htop);
                    if (last_page_end - new_htop < blackmap[n])
                    {
                        struct nm_page *new_page = nm_bibop;

                        //DPRINT("Need to fetch a new page!\n\r");
                        nm_bibop = nm_bibop->next;
                        if (nm_bibop)
                            nm_bibop->prev = NULL;

                        new_page->next = new_used_mem;
                        new_used_mem->prev = new_page;
                        new_used_mem = new_page;
                        //DPRINT("Snatched a new page @ 0x%08x\n\r",(int)new_page);
                        new_htop = new_page->start;
                        last_page_end = new_htop + NM_PAGESIZE;
                    }

                    memcpy(new_htop,ptr,blackmap[n] * sizeof(Eterm));
                    for (i = 0; i < blackmap[n]; i++)
                      {
                        //DPRINT("cp 0x%08x = 0x%08x",(int)ptr,(int)new_htop);
                        *ptr++ = (Eterm)new_htop++;
                        //DPRINT("    done: 0x%08x\n\r",(int)*ptr++); OBS OBS!!
                      }
                  //new_htop += blackmap[n];
                  //ptr += blackmap[n];
                    n += blackmap[n];
                }
                else
                {
                    n++; ptr++;
                }
            }
            page = page->next;
        }

        page = nm_used_mem;
        while (page)
        {
            Eterm *ptr = page->start;
            Eterm *page_end = ptr + NM_PAGESIZE;

            /* XXX: If nm_used_mem is sorted in address order, this
             * pass can be done at the same time as copying. */
            while (ptr < page_end)
            {
                //DPRINT("? 0x%08x: 0x%08x\n\r",(int)ptr,(int)*ptr);
                if (ptr_within(ptr_val(*ptr),nm_heap,nm_hend))
                {
                    //DPRINT("- 0x%08x: 0x%08x\n\r",(int)ptr_val(*ptr),(int)*((Eterm*)ptr_val(*ptr)));
                    *ptr = *((Eterm*)ptr_val(*ptr));
                    //DPRINT("! 0x%08x: 0x%08x\n\r",(int)ptr,(int)*ptr);
                }
                ptr++;
            }
            page = page->next;
        }

        while (roots_saved--)
        {
            Eterm *ptr = root_save[roots_saved];
            //DPRINT("? 0x%08x: 0x%08x\n\r",(int)ptr,(int)*ptr);
            //DPRINT("- 0x%08x: 0x%08x\n\r",(int)ptr_val(*ptr),(int)*((Eterm*)ptr_val(*ptr)));
            *ptr = *((Eterm*)ptr_val(*ptr));
            //DPRINT("! 0x%08x: 0x%08x\n\r",(int)ptr,(int)*ptr);
        }

        erts_free(ERTS_ALC_T_OLD_HEAP,(void*)nm_heap);

        nm_heap = new_heap;
        nm_hend = new_hend;
        nm_used_mem = new_used_mem;
        nm_pages = new_pages;

        if ((last_page_end - new_htop) * sizeof(Eterm) >=
            sizeof(struct nm_mem_block))
        {
            nm_free_list = (struct nm_mem_block*)(new_htop);
            nm_free_list->size = last_page_end - new_htop;
            nm_free_list->prev = NULL;
            nm_free_list->next = NULL;
            //DPRINT("Created free block (3): 0x%08x of size %d\n\r",(int)nm_free_list,nm_free_list->size);
            if (nm_free_list->size >= need)
                found = nm_free_list;
        }
        //print_nm_heap();
        //print_free_list();
    }

    if (!found)
    {
        struct nm_page *new_page = nm_bibop;
        struct nm_mem_block *new_free =
          (struct nm_mem_block*)new_page->start;

        nm_bibop = nm_bibop->next;

        new_page->next = nm_used_mem;
        if (nm_used_mem)
            nm_used_mem->prev = new_page;
        nm_used_mem = new_page;

        //BM_SWAP_TIMER(gc,misc1);
        memset(blackmap + ((void*)new_page - (void*)nm_heap) / sizeof(void*),
               0,NM_FULLPAGE);
        //BM_SWAP_TIMER(misc1,gc);

        new_free->prev = NULL;
        new_free->next = nm_free_list;
        new_free->size = NM_PAGESIZE;
        if (nm_free_list)
            nm_free_list->prev = new_free;
        nm_free_list = new_free;
        //DPRINT("Snatched a new page @ 0x%08x\n\r",(int)new_page);
        //print_free_list();
        found = new_free;
    }

    //restore_rootset(NULL, rootset);
    //erts_sl_free((void *)rootset);
    //print_nm_heap();
    //print_page_list(nm_used_mem);
    //print_free_list();

    //BM_STOP_TIMER(old_gc);
#ifdef BM_TIMER
    major_global_gc_time += old_gc_time;
    if (old_gc_time > max_global_major_time)
      max_global_major_time = old_gc_time;
#endif
    //BM_START_TIMER(gc);

    return found;
}

#else /* !SHARED_HEAP */

static struct nm_mem_block *erts_nm_collect(Process *p, int need, Eterm* objv, int nobj)
{
    Rootset *rootset = erts_alloc(ERTS_ALC_T_ROOTSET,
                                  sizeof(Rootset)*erts_max_processes);
    Eterm *free_start = NULL;
    struct nm_mem_block *found = NULL;
    struct nm_page *page = NULL;
    BM_NEW_TIMER(old_gc);

    //BM_SWAP_TIMER(gc,old_gc);
    BM_COUNT(major_global_garbage_cols);

    //print_nm_heap();
    //print_heap(HEAP_START(p),HEAP_END(p));


    roots_saved = 0;
    {
        int n = collect_roots(p, objv, nobj, rootset);
        while (n--)
        {
            while (rootset[n].n--)
            {
                Eterm* ptr = rootset[n].v[rootset[n].n];
                Uint  sz = rootset[n].sz[rootset[n].n];
                root_area_scan(ptr,sz);
            }
        }
    }

    {
        Process* p;
        Uint i;
        Uint n = erts_num_active_procs;

        for (i = 0; i < n; i++) {
            p = erts_active_procs[i];
            private_heap_scan(p->heap,p->htop - p->heap);
        }
    }

    message_area_scan(global_heap,global_htop - global_heap);  

    //BM_SWAP_TIMER(old_gc,misc0);
    /*
    {
        NM_STORAGE_ITERATOR(root);
        struct nm_object *obj = NM_STORAGE_STEP(root);
        while (obj)
        {
            if (blackmap[obj->this - nm_heap] == 0)
            {
                blackmap[obj->this - nm_heap] = obj->size;
                //DPRINT("Rescue #%d (0x%08x) size: %d\n\r",(int)(obj->this - nm_heap),(int)obj->this,obj->size);
                area_scan(obj->this,obj->size);
            }
            //else
            //BM_COUNT(messages_ego);
            //DPRINT("Already scanned #%d\n\r",(int)(ptr_val(obj->this) - nm_heap));
            obj = NM_STORAGE_STEP(root);
        }
    }
    */
    {
        struct nm_object *obj = NM_STORAGE_GET(gray);
        while (obj)
        {
            if (blackmap[obj->this - nm_heap] == 0)
            {
                blackmap[obj->this - nm_heap] = obj->size;
                //DPRINT("Rescue #%d (0x%08x) size: %d\n\r",(int)(obj->this - nm_heap),(int)obj->this,obj->size);
                //BM_SWAP_TIMER(old_gc,misc2);
                area_scan(obj->this,obj->size);
                //BM_SWAP_TIMER(misc2,old_gc);
            }
            //else
            //BM_COUNT(messages_ego);
            //DPRINT("Already scanned #%d\n\r",(int)(ptr_val(obj->this) - nm_heap));
            obj = NM_STORAGE_GET(gray);
        }
    }

    /* XXX: If we remember old free list, can we gain in any way..? */
    nm_free_list = NULL;
    page = nm_used_mem;

    while (page)
    {
        int scavenging = 0;
        Eterm *ptr = page->start;
        Eterm *page_end = ptr + NM_PAGESIZE;
        int n = page->start - nm_heap;
        while (ptr < page_end)
        {
            if (blackmap[n] != 0)
            {
                if (scavenging)
                {
                    scavenging = 0;
                    if ((ptr - free_start) * sizeof(Eterm) >=
                        sizeof(struct nm_mem_block))
                    {
                        struct nm_mem_block *new =
                          (struct nm_mem_block*)(free_start);
                        new->size = ptr - free_start;
                        new->prev = NULL;
                        new->next = nm_free_list;
                        if (nm_free_list)
                            nm_free_list->prev = new;
                        nm_free_list = new;
                        //DPRINT("Created free block (1): 0x%08x of size %d\n\r",(int)new,new->size);
                        if ((new->size >= need) &&
                            (!found || (found->size > new->size)))
                        {
                            found = new;
                            //DPRINT("We've got one!\n\r");
                        }
                    }
                }
                //DPRINT("Found a live one #%d @ 0x%08x!\n\r",n,(int)ptr);
                ptr += blackmap[n];
                n += blackmap[n];
            }
            else if (!scavenging)
            {
              //DPRINT("Free block starts @ 0x%08x\n\r",(int)ptr);
                free_start = ptr;
                scavenging = 1;
                n++; ptr++;
            }
            else
            {
                n++; ptr++;
            }
        }

        if (scavenging)
        {
            if (free_start == page->start)
            {
                struct nm_page *next = page->next;
                //DPRINT("Returning page 0x%08x to BIBOP\n\r",(int)page);
                if (page->next)
                    page->next->prev = page->prev;

                if (page->prev)
                    page->prev->next = page->next;
                else
                    nm_used_mem = page->next;

                page->next = nm_bibop;
                nm_bibop = page;
                page = next;
                //print_page_list(nm_bibop);
                //print_page_list(nm_used_mem);
                continue;
            }
            else if ((ptr - free_start) * sizeof(Eterm) >
                     sizeof(struct nm_mem_block))
            {
                struct nm_mem_block *new = (struct nm_mem_block*)(free_start);
                new->size =  ptr - free_start;
                new->prev = NULL;
                new->next = nm_free_list;
                if (nm_free_list)
                    nm_free_list->prev = new;
                nm_free_list = new;
                //DPRINT("Created free block (2): 0x%08x of size %d\n\r",(int)new,new->size);
                if ((new->size >= need) &&
                    (!found || (found->size > new->size)))
                {
                    found = new;
                    //DPRINT("We've got one!\n\r");
                }
            }
        }
        page = page->next;
    }
    //print_free_list();
    //BM_SWAP_TIMER(misc2,old_gc);

    if (!found && !nm_bibop)
    {
        int i;
        int new_pages = nm_pages * 2;
        int size = sizeof(Eterm) * new_pages * NM_FULLPAGE;
        Eterm *new_heap = ERTS_HEAP_ALLOC(ERTS_ALC_T_OLD_HEAP,size);
        Eterm *new_hend = new_heap + size;
        Eterm *new_htop;
        Eterm *last_page_end;
        struct nm_page *new_used_mem;

        printf("We need to copy things!\n\r");

        /* Create new, bigger bag of pages */
        for (i = 0; i < new_pages; i++)
        {
            struct nm_page *this =
              (struct nm_page*)(new_heap + i * NM_FULLPAGE);
            this->prev = (struct nm_page*)((Eterm*)this - NM_FULLPAGE);
            this->next = (struct nm_page*)((Eterm*)this + NM_FULLPAGE);
        }
        nm_bibop = (struct nm_page*)new_heap;
        nm_bibop->prev = NULL;
        ((struct nm_page*)(new_heap + (new_pages - 1) *
                           NM_FULLPAGE))->next = NULL;

        new_used_mem = nm_bibop;
        nm_bibop = nm_bibop->next;
        nm_bibop->prev = NULL;
        new_used_mem->next = new_used_mem->prev = NULL;

        //print_page_list(nm_bibop);
        //print_page_list(new_used_mem);

        /* Move stuff from old bag to new */
        nm_free_list = NULL;
        new_htop = new_used_mem->start;
        last_page_end = new_htop + NM_PAGESIZE;
        page = nm_used_mem;
        while (page)
        {
            Eterm *ptr = page->start;
            Eterm *page_end = ptr + NM_PAGESIZE;
            int n = offsetof(struct nm_page,start) / sizeof(void*) +
              ((Eterm*)page - nm_heap);
            while (ptr < page_end)
            {
                if (blackmap[n] > 0)
                {
                  //DPRINT("Found a live one #%d @ 0x%08x! Left: %d\n\r",n,(int)ptr,last_page_end-new_htop);
                    if (last_page_end - new_htop < blackmap[n])
                    {
                        struct nm_page *new_page = nm_bibop;

                        //DPRINT("Need to fetch a new page!\n\r");
                        nm_bibop = nm_bibop->next;
                        if (nm_bibop)
                            nm_bibop->prev = NULL;

                        new_page->next = new_used_mem;
                        new_used_mem->prev = new_page;
                        new_used_mem = new_page;
                        //DPRINT("Snatched a new page @ 0x%08x\n\r",(int)new_page);
                        new_htop = new_page->start;
                        last_page_end = new_htop + NM_PAGESIZE;
                    }

                    memcpy(new_htop,ptr,blackmap[n] * sizeof(Eterm));
                    for (i = 0; i < blackmap[n]; i++)
                      {
                        //DPRINT("cp 0x%08x = 0x%08x",(int)ptr,(int)new_htop);
                        *ptr++ = (Eterm)new_htop++;
                        //DPRINT("    done: 0x%08x\n\r",(int)*ptr++); OBS OBS!!
                      }
                  //new_htop += blackmap[n];
                  //ptr += blackmap[n];
                    n += blackmap[n];
                }
                else
                {
                    n++; ptr++;
                }
            }
            page = page->next;
        }

        page = nm_used_mem;
        while (page)
        {
            Eterm *ptr = page->start;
            Eterm *page_end = ptr + NM_PAGESIZE;

            /* XXX: If nm_used_mem is sorted in address order, this
             * pass can be done at the same time as copying. */
            while (ptr < page_end)
            {
                //DPRINT("? 0x%08x: 0x%08x\n\r",(int)ptr,(int)*ptr);
                if (ptr_within(ptr_val(*ptr),nm_heap,nm_hend))
                {
                    //DPRINT("- 0x%08x: 0x%08x\n\r",(int)ptr_val(*ptr),(int)*((Eterm*)ptr_val(*ptr)));
                    *ptr = *((Eterm*)ptr_val(*ptr));
                    //DPRINT("! 0x%08x: 0x%08x\n\r",(int)ptr,(int)*ptr);
                }
                ptr++;
            }
            page = page->next;
        }

        while (roots_saved--)
        {
            Eterm *ptr = root_save[roots_saved];
            //DPRINT("? 0x%08x: 0x%08x\n\r",(int)ptr,(int)*ptr);
            //DPRINT("- 0x%08x: 0x%08x\n\r",(int)ptr_val(*ptr),(int)*((Eterm*)ptr_val(*ptr)));
            *ptr = *((Eterm*)ptr_val(*ptr));
            //DPRINT("! 0x%08x: 0x%08x\n\r",(int)ptr,(int)*ptr);
        }

        erts_free(ERTS_ALC_T_OLD_HEAP,(void*)nm_heap);

        nm_heap = new_heap;
        nm_hend = new_hend;
        nm_used_mem = new_used_mem;
        nm_pages = new_pages;

        if ((last_page_end - new_htop) * sizeof(Eterm) >=
            sizeof(struct nm_mem_block))
        {
            nm_free_list = (struct nm_mem_block*)(new_htop);
            nm_free_list->size = last_page_end - new_htop;
            nm_free_list->prev = NULL;
            nm_free_list->next = NULL;
            //DPRINT("Created free block (3): 0x%08x of size %d\n\r",(int)nm_free_list,nm_free_list->size);
            if (nm_free_list->size >= need)
                found = nm_free_list;
        }
        //print_nm_heap();
        //print_free_list();
    }

    if (!found)
    {
        struct nm_page *new_page = nm_bibop;
        struct nm_mem_block *new_free =
          (struct nm_mem_block*)new_page->start;

        nm_bibop = nm_bibop->next;

        new_page->next = nm_used_mem;
        if (nm_used_mem)
            nm_used_mem->prev = new_page;
        nm_used_mem = new_page;

        //BM_SWAP_TIMER(gc,misc1);
        memset(blackmap + ((void*)new_page - (void*)nm_heap) / sizeof(void*),
               0,NM_FULLPAGE);
        //BM_SWAP_TIMER(misc1,gc);

        new_free->prev = NULL;
        new_free->next = nm_free_list;
        new_free->size = NM_PAGESIZE;
        if (nm_free_list)
            nm_free_list->prev = new_free;
        nm_free_list = new_free;
        //DPRINT("Snatched a new page @ 0x%08x\n\r",(int)new_page);
        //print_free_list();
        found = new_free;
    }

    restore_rootset(NULL, rootset);
    erts_free(ERTS_ALC_T_ROOTSET,(void *)rootset);
    //print_nm_heap();
    //print_page_list(nm_used_mem);
    //print_free_list();

    //BM_STOP_TIMER(old_gc);
#ifdef BM_TIMER
    major_global_gc_time += old_gc_time;
    if (old_gc_time > max_global_major_time)
      max_global_major_time = old_gc_time;
#endif
    //BM_START_TIMER(gc);

    return found;
}
#endif /* SHARED_HEAP */

static void erts_nm_copymark(Process *p, Eterm* objv, int nobj)
{
#if HIPE
    char *const_start = (char*)hipe_constants_start;
    unsigned long const_size = (char*)hipe_constants_next - const_start;
#endif
    struct nm_object *obj = NM_STORAGE_GET(blue);
    while (obj)
    {
        Eterm *this = obj->this;
        int sz = obj->size;
        //DPRINT("Blue object: 0x%08x of size: %d\n\r",(int)this,obj->size);
        while(sz--)
        {
          //DPRINT("%d: 0x%08x  ",obj->size,(int)*this);
            switch (primary_tag(*this))
            {
              case TAG_PRIMARY_BOXED:
              {
                  Eterm *ptr = boxed_val(*this);
                  Eterm val = *ptr;
                  //DPRINT("(boxed)\n\r");
                  if (ptr_within(ptr, nm_heap, nm_hend))
                  {
                      this++;
                  }
#if HIPE
                  else if( in_area(ptr, const_start, const_size) )
                  {
                      this++;
                  }
#endif
                  else if (MY_IS_MOVED(val))
                  {
                      ASSERT(is_boxed(val));
                      *this++ = val;
                  }
                  else
                  {
                    //DPRINT("ptr: 0x%08x  val: 0x%08x  BN: %d\n\r",(int)ptr,(int)val,BOXED_NEED(ptr,val));
                    //DPRINT("fun-free: %d\n\r",((ErlFunThing*)(ptr - 1))->num_free);
                      Eterm *old_htop = erts_nm_alloc(p,BOXED_NEED(ptr,val),objv,nobj);
                      MOVE_BOXED(ptr,val,old_htop,this++);
                  }
                  continue;
              }

              case TAG_PRIMARY_LIST:
              {
                  Eterm *ptr = list_val(*this);
                  Eterm val = *ptr;
                  //DPRINT("(list)\n\r");
                  if (ptr_within(ptr, nm_heap, nm_hend))
                  {
                      this++;
                  }
#if HIPE
                  else if( in_area(ptr, const_start, const_size) )
                  {
                      this++;
                  }
#endif
                  else if (is_non_value(val))
                  {
                      *this++ = ptr[1];
                  }
                  else
                  {
                      Eterm *old_htop = erts_nm_alloc(p,2,objv,nobj);
                      MOVE_CONS(ptr,val,old_htop,this++);
                  }
                  continue;
              }

              case TAG_PRIMARY_HEADER:
              {
                //DPRINT("(header");
                  if (header_is_thing(*this))
                  {
                    //DPRINT("arity: %d)\n\r",thing_arityval(*this));
                      sz -= thing_arityval(*this);
                      this += (thing_arityval(*this) + 1);
                  }
                  else
                  {
                    //DPRINT(")\n\r");
                      this++;
                  }
                  continue;
              }

              default:
              {
                //DPRINT("(other)\n\r");
                  this++;
                  continue;
              }
            }
        }
        obj = NM_STORAGE_GET(blue);
    }
}

Eterm *erts_nm_alloc(Process *p, int need, Eterm* objv, int nobj)
{
    struct nm_mem_block *this = nm_free_list;

    //DPRINT("Allocate %d\n\r",need);
    if (!this)
    {
        this = erts_nm_collect(p,need,objv,nobj);
        if (!this)
          printf("We may be low on memory...\n\r");
    }

    if (need > NM_PAGESIZE)
        printf("nm_alloc: ERROR! Your need is larger than the page size!\n\r");

    while (this->size < need)
    {
      //DPRINT("Too small: %d\n\r",this->size);
        this = this->next;
        if (!this)
        {
            this = erts_nm_collect(p,need,objv,nobj);
        }
    }

    if ((this->size - need) * sizeof(Eterm) >= sizeof(struct nm_mem_block))
    {
        struct nm_mem_block *rest =
          (struct nm_mem_block*)((Eterm*)this + need);
        if (this->size > NM_PAGESIZE) print_nm_heap();
        //DPRINT("NF: need: %d  rest: 0x%08x  this: 0x%08x\n\r",need,(int)rest,(int)this);
        /* The order here IS important! */
        rest->next = this->next;
        //DPRINT("this  sz: %d  prev: 0x%08x  next: 0x%08x\n\r",this->size,(int)this->prev,(int)this->next);
        //DPRINT("rest  sz: %d  prev: 0x%08x  next: 0x%08x\n\r",rest->size,(int)rest->prev,(int)rest->next);
        if (rest->next)
            rest->next->prev = rest;

        rest->prev = this->prev;

        if (rest->prev)
            rest->prev->next = rest;
        else
            nm_free_list = rest;

        rest->size = this->size - need;
        //DPRINT("New free @ 0x%08x with size %d\n\r",(int)rest,rest->size);
    }
    else
    {
      //DPRINT("Blocksize left: %d, min size: %d\n\r",this->size - need,sizeof(struct nm_mem_block) / sizeof(void*));
        if (this->prev)
            this->prev->next = this->next;
        else
            nm_free_list = this->next;

        if (this->next)
          this->next->prev = this->prev;
        //print_free_list();
    }

    //DPRINT("Allocated %d @ 0x%08x\n\r",need,(int)this);
    //print_nm_heap();

#ifdef SHARED_HEAP
    /* We do not add everything we allocate to the build list, since
     * it would only incur a lot of duplets in the gray list. If all
     * the allocated objects are in the build list they are part of
     * the root set, but they will also be found by the mark scan and
     * added to the gray list.
    */
    NM_STORE(blue,(Eterm*)this,need);

    if (need > 255)
      printf("Implement 32-bit support in blackmap!!\n\r");

    blackmap[((Eterm*)this) - nm_heap] = need;
#endif

    return (Eterm*)this;
}
#endif /* NOMOVE */


#ifdef INCREMENTAL_GC
/***************************************************************************
 *                                                                         *
 *                    Incremental Garbage Collector                        *
 *                                                                         *
 ***************************************************************************/


static double   inc_scan_unit       = 1.0;
static Process *inc_current_process = NULL;
static Rootset *inc_current_rootset = NULL;

void inc_gc(Process* p, int need) //, Eterm* objv, int nobj)
{
    int scanlimit;

    BM_STOP_TIMER(system);

    //DPRINT("GC START\n\r");

    gc_time = 0;
    BM_START_TIMER(gc);
    //print_heap(HEAP_START(p),HEAP_TOP(p));



    if (!inc_current_process)
    {
        inc_current_process = next_p_in_process_tab;
        if (!inc_current_process)
        {
            /* We have now processed all roots in this GC cycle,
             * let's prepare for the next cycle.
             */
            inc_scan_unit = HEAP_SIZE(p) / (HEAP_SIZE(p) - HEAP_TOP(p));

            /* Swap to-space and from-space. All allocations from now on
             * will end up in the new to-space */
            HEAP_TOP(p) = inc_fromspace;
            inc_fromspace = HEAP_START(p);
            HEAP_START(p) = HEAP_TOP(p);
        }
        else
            inc_current_rootset = get_roots(inc_current_process);
    }

    scanlimit = need * inc_scan_unit + 1;

    if (inc_current_rootset)
    {
        /* There are roots left to scan */
        while (scanlimit && !finished)
        {
            do your thing;
            scanlimit--;
        }

        if (finished)
        {
            inc_current_rootset = NULL;
        }
    }

    while (scanlimit)
    {
        /* Cheney scan the to-space */
        scanlimit--;
    }

    BM_STOP_TIMER(gc);

#ifdef __BENCHMARK__
#ifdef SHARED_HEAP
        BM_COUNT(minor_global_garbage_cols);
#else
        BM_COUNT(minor_garbage_cols);
#endif
#ifdef BM_TIMERS
#ifdef SHARED_HEAP
        minor_global_gc_time += gc_time;
        if (gc_time > max_global_minor_time)
            max_global_minor_time = gc_time;
#else
        minor_gc_time += gc_time;
        if (gc_time > max_minor_time)
            max_minor_time = gc_time;
#endif
#endif
#endif /* __BENCHMARK__ */

    BM_START_TIMER(system);

    //DPRINT("GC END\n\r");
    //print_heap(HEAP_START(p),HEAP_TOP(p));
    //print_nm_heap();
}



Eterm *erts_alloc_with_gc(Process *p, int sz)
{
    inc_gc(p,sz);
    HEAP_TOP(p) = HEAP_TOP(p) + sz;
    return HEAP_TOP(p) - sz;
}

#endif /* INCREMENTAL_GC */
